@@ -1,7 +1,7 @@
 /*
  *	apply.c
  *
- *	Apply Function to Files
+ *	apply function to files
  */
 
 #include <fcntl.h>
@@ -10,8 +10,10 @@
 #include "kanji.h"
 
 #define	MAXTIMESTR	8
+#define	ATTRWIDTH	10
 
 typedef struct _attrib_t {
+	short nlink;
 	u_int mode;
 #ifdef	HAVEFLAGS
 	u_long flags;
@@ -24,8 +26,6 @@ extern int subwindow;
 extern int win_x;
 extern int win_y;
 extern int lcmdline;
-extern int tradlayout;
-extern int sizeinfo;
 extern int mark;
 #ifdef	HAVEFLAGS
 extern u_long fflaglist[];
@@ -47,7 +47,7 @@ extern int inheritcopy;
 #endif
 
 static int NEAR issamedir __P_((char *, char *));
-static int NEAR islowerdir __P_((char *, char *));
+static int NEAR islowerdir __P_((VOID_A));
 static char *NEAR getdestdir __P_((char *, char *));
 static int NEAR getdestpath __P_((char *, char *, struct stat *));
 static int NEAR checkdupl __P_((char *, char *, struct stat *, struct stat *));
@@ -59,7 +59,7 @@ static int touchdir __P_((char *));
 #ifndef	_NOEXTRACOPY
 static int mvdir __P_((char *));
 #endif
-static VOID NEAR showattr __P_((namelist *, attrib_t *, int y));
+static VOID NEAR showattr __P_((namelist *, attrib_t *, int));
 static char **NEAR getdirtree __P_((char *, char **, int *, int));
 static int NEAR _applydir __P_((char *, int (*)(char *),
 		int (*)(char *), int (*)(char *), int, char *, int));
@@ -71,12 +71,14 @@ int copypolicy = 0;
 int removepolicy = 0;
 char *destpath = NULL;
 
-static u_int attrmode = 0;
+static short attrnlink = 0;
+static u_short attrmode = 0;
 #ifdef	HAVEFLAGS
 static u_long attrflags = 0;
 #endif
 static time_t attrtime = 0;
 static char *destdir = NULL;
+static short destnlink = 0;
 static time_t destmtime = (time_t)-1;
 static time_t destatime = (time_t)-1;
 #ifndef	_NOEXTRACOPY
@@ -117,19 +119,21 @@ char *path, *org;
 	if (org != cwd) free(org);
 	free(cwd);
 	if (path) free(path);
+
 	return(i);
 }
 
-static int NEAR islowerdir(path, org)
-char *path, *org;
+static int NEAR islowerdir(VOID_A)
 {
 #ifdef	_USEDOSEMU
 	char orgpath[MAXPATHLEN];
 #endif
-	char *cp, *top, *cwd, tmp[MAXPATHLEN];
+	char *cp, *top, *cwd, *path, *org, tmp[MAXPATHLEN];
 	int i;
 
 	cwd = getwd2();
+	path = destpath;
+	org = filelist[filepos].name;
 	if (!org) org = cwd;
 #ifdef	_USEDOSEMU
 	else org = nodospath(orgpath, org);
@@ -165,6 +169,7 @@ char *path, *org;
 		free(path);
 	}
 	free(cwd);
+
 	return(i);
 }
 
@@ -177,8 +182,8 @@ char *mes, *arg;
 #endif
 
 	if (arg && *arg) dir = strdup2(arg);
-	else if (!(dir = inputstr(mes, 1, -1, NULL, 1))) return(NULL);
-	else if (!*(dir = evalpath(dir, 1))) {
+	else if (!(dir = inputstr(mes, 1, -1, NULL, HST_PATH))) return(NULL);
+	else if (!*(dir = evalpath(dir, 0))) {
 		dir = realloc2(dir, 2);
 		dir[0] = '.';
 		dir[1] = '\0';
@@ -208,6 +213,7 @@ char *mes, *arg;
 #endif
 		return(NULL);
 	}
+
 	return(dir);
 }
 
@@ -233,6 +239,7 @@ struct stat *stp;
 		warning(-1, file);
 		return(-1);
 	}
+
 	return(0);
 }
 
@@ -241,7 +248,7 @@ char *file, *dest;
 struct stat *stp1, *stp2;
 {
 	char *cp, *tmp, *str[MAXCOPYITEM];
-	int i, n, val[MAXCOPYITEM];
+	int i, n, ch, val[MAXCOPYITEM];
 #ifndef	_NOEXTRACOPY
 	char path[MAXPATHLEN];
 #endif
@@ -256,9 +263,8 @@ struct stat *stp1, *stp2;
 		warning(-1, dest);
 		return(-1);
 	}
-	if ((stp2 -> st_mode & S_IFMT) != S_IFDIR)
-		n = (copypolicy & ~FLAG_SAMEDIR);
-	else if ((stp1 -> st_mode & S_IFMT) != S_IFDIR) n = 2;
+	if (!s_isdir(stp2)) n = (copypolicy & ~FLAG_SAMEDIR);
+	else if (!s_isdir(stp1)) n = 2;
 #ifdef	_NOEXTRACOPY
 	else return(1);
 #else
@@ -271,12 +277,12 @@ struct stat *stp1, *stp2;
 
 	for (;;) {
 		if (!n || n == 2) {
-			locate(0, L_CMDLINE);
-			putterm(l_clear);
+			Xlocate(0, L_CMDLINE);
+			Xputterm(L_CLEAR);
 			cp = SAMEF_K;
 			i = strlen2(cp) - (sizeof("%.*s") - 1);
 			cp = asprintf3(cp, n_column - i, dest);
-			kanjiputs(cp);
+			Xkanjiputs(cp);
 			free(cp);
 		}
 		if (!n) {
@@ -293,15 +299,16 @@ struct stat *stp1, *stp2;
 			val[4] = 5;
 #endif
 			i = 1;
-			if (selectstr(&i, MAXCOPYITEM, 0, str, val) == K_ESC)
-				return(-1);
+			ch = selectstr(&i, MAXCOPYITEM, 0, str, val);
+			if (ch == K_ESC) return(-1);
+			else if (ch != K_CR) return(-2);
 #ifndef	_NOEXTRACOPY
 			if (i == 5) {
 				str[0] = UPDAT_K;
 				str[1] = OVERW_K;
 				val[0] = 5;
 				val[1] = 6;
-				if (selectstr(&i, 2, 0, str, val) == K_ESC)
+				if (selectstr(&i, 2, 0, str, val) != K_CR)
 					continue;
 # ifndef	_NODOSDRIVE
 				dupdestdrive = destdrive;
@@ -333,8 +340,8 @@ struct stat *stp1, *stp2;
 				break;
 			case 2:
 				lcmdline = L_INFO;
-				if (!(tmp = inputstr(NEWNM_K, 1,
-					-1, NULL, -1))) return(-1);
+				tmp = inputstr(NEWNM_K, 1, -1, NULL, -1);
+				if (!tmp) return(-1);
 				strcpy(getbasename(dest), tmp);
 				free(tmp);
 				if (Xlstat(dest, stp2) < 0) {
@@ -343,10 +350,8 @@ struct stat *stp1, *stp2;
 					warning(-1, dest);
 					return(-1);
 				}
-				if ((stp1 -> st_mode & S_IFMT) == S_IFDIR
-				&& (stp2 -> st_mode & S_IFMT) == S_IFDIR)
-					return(1);
-				putterm(t_bell);
+				if (s_isdir(stp1) && s_isdir(stp2)) return(1);
+				Xputterm(T_BELL);
 				break;
 			case 3:
 				return(2);
@@ -391,7 +396,7 @@ char *path;
 int mode;
 {
 	char *cp, *str[4];
-	int len, val[4];
+	int ch, len, val[4];
 #if	MSDOS
 
 	if (Xaccess(path, mode) >= 0) return(0);
@@ -411,7 +416,7 @@ int mode;
 			warning(-1, path);
 			return(-1);
 		}
-		if ((st.st_mode & S_IFMT) == S_IFLNK) return(0);
+		if (s_islnk(&st)) return(0);
 
 		if (!(tmp = strrdelim(path, 0))) strcpy(dir, ".");
 		else if (tmp == path) strcpy(dir, _SS_);
@@ -421,8 +426,8 @@ int mode;
 			warning(-1, dir);
 			return(-1);
 		}
-		stp = &st;
 		if (Xaccess(path, mode) >= 0) return(0);
+		stp = &st;
 	}
 #endif	/* !MSDOS */
 	if (errno == ENOENT) return(0);
@@ -449,12 +454,12 @@ int mode;
 	}
 #endif	/* !MSDOS */
 	if (removepolicy > 0) return(removepolicy - 2);
-	locate(0, L_CMDLINE);
-	putterm(l_clear);
+	Xlocate(0, L_CMDLINE);
+	Xputterm(L_CLEAR);
 	cp = DELPM_K;
 	len = strlen2(cp) - (sizeof("%.*s") - 1);
 	cp = asprintf3(cp, n_column - len, path);
-	kanjiputs(cp);
+	Xkanjiputs(cp);
 	free(cp);
 	str[0] = ANYES_K;
 	str[1] = ANNO_K;
@@ -464,8 +469,10 @@ int mode;
 	val[1] = -1;
 	val[2] = 2;
 	val[3] = 1;
-	if (selectstr(&removepolicy, 4, 0, str, val) == K_ESC)
-		removepolicy = -1;
+	ch = selectstr(&removepolicy, 4, 0, str, val);
+	if (ch == K_ESC) removepolicy = -1;
+	else if (ch != K_CR) removepolicy = -2;
+
 	return((removepolicy > 0) ? removepolicy - 2 : removepolicy);
 }
 
@@ -474,9 +481,12 @@ char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
+	int n;
 
-	if (checkdupl(path, dest, &st1, &st2) < 0) return(0);
+	if ((n = checkdupl(path, dest, &st1, &st2)) < 0)
+		return((n < -1) ? -2 : 1);
 	if (safecpfile(path, dest, &st1, &st2) < 0) return(-1);
+
 	return(0);
 }
 
@@ -485,10 +495,13 @@ char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
+	int n;
 
-	if (checkdupl(path, dest, &st1, &st2) < 0) return(0);
-	if (checkrmv(path, W_OK) < 0) return(1);
+	if ((n = checkdupl(path, dest, &st1, &st2)) < 0
+	|| (n = checkrmv(path, W_OK)) < 0)
+		return((n < -1) ? -2 : 1);
 	if (safemvfile(path, dest, &st1, &st2) < 0) return(-1);
+
 	return(0);
 }
 
@@ -498,6 +511,7 @@ char *path;
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
 
+	destnlink = (TCH_MODE | TCH_UID | TCH_GID);
 	switch (checkdupl(path, dest, &st1, &st2)) {
 		case 2:
 		/* Already exist, but not directory */
@@ -511,22 +525,29 @@ char *path;
 			if (Xmkdir(dest, st1.st_mode & 0777) < 0)
 #endif
 				return(-1);
+			destnlink |= (TCH_ATIME | TCH_MTIME);
 			destmtime = st1.st_mtime;
 			destatime = st1.st_atime;
 			break;
 		case -1:
+		case -2:
 		/* Abandon copy */
 			return(-2);
 /*NOTREACHED*/
 			break;
 		default:
 		/* Already exist */
+			destnlink |= (TCH_ATIME | TCH_MTIME);
 			destmtime = st2.st_mtime;
 			destatime = st2.st_atime;
 			break;
 	}
+
 	if (destdir) free(destdir);
-	destdir = strdup2(&(dest[strlen(destpath) + 1]));
+	destdir = &(dest[strlen(destpath)]);
+	while (*destdir == _SC_) destdir++;
+	destdir = strdup2(destdir);
+
 	return(0);
 }
 
@@ -542,15 +563,12 @@ char *path;
 	if (!inheritcopy) return(0);
 # endif
 	if (getdestpath(path, dest, &st) < 0) return(-2);
-	if (destmtime != (time_t)-1 || destatime != (time_t)-1) {
-		st.st_mtime = destmtime;
-		st.st_atime = destatime;
-	}
-# ifdef	HAVEFLAGS
-	st.st_flags = 0xffffffff;
-# endif
+	st.st_nlink = destnlink;
+	st.st_mtime = destmtime;
+	st.st_atime = destatime;
 	if (touchfile(dest, &st) < 0) return(-1);
 #endif	/* _USEDOSCOPY || !_NOEXTRACOPY */
+
 	return(0);
 }
 
@@ -560,17 +578,16 @@ char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
+	int n;
 
 	if (getdestpath(path, dest, &st) < 0) return(-2);
-	if (destmtime != (time_t)-1 || destatime != (time_t)-1) {
-		st.st_mtime = destmtime;
-		st.st_atime = destatime;
-	}
-# ifdef	HAVEFLAGS
-	st.st_flags = 0xffffffff;
-# endif
+	st.st_nlink = destnlink;
+	st.st_mtime = destmtime;
+	st.st_atime = destatime;
 	if (touchfile(dest, &st) < 0) return(-1);
-	if (checkrmv(path, R_OK | W_OK | X_OK) < 0) return(1);
+	if ((n = checkrmv(path, R_OK | W_OK | X_OK)) < 0)
+		return((n < -1) ? -2 : 1);
+
 	return(Xrmdir(path));
 }
 #endif	/* !_NOEXTRACOPY */
@@ -578,14 +595,21 @@ char *path;
 int rmvfile(path)
 char *path;
 {
-	if (checkrmv(path, W_OK) < 0) return(1);
+	int n;
+
+	if ((n = checkrmv(path, W_OK)) < 0) return((n < -1) ? -2 : 1);
+
 	return(Xunlink(path));
 }
 
 int rmvdir(path)
 char *path;
 {
-	if (checkrmv(path, R_OK | W_OK | X_OK) < 0) return(1);
+	int n;
+
+	if ((n = checkrmv(path, R_OK | W_OK | X_OK)) < 0)
+		return((n < -1) ? -2 : 1);
+
 	return(Xrmdir(path));
 }
 
@@ -594,14 +618,15 @@ char *path;
 {
 	if (regexp_exec(findregexp, getbasename(path), 1)) {
 		if (path[0] == '.' && path[1] == _SC_) path += 2;
-		locate(0, L_CMDLINE);
-		putterm(l_clear);
-		cprintf2("[%.*k]", n_column - 2, path);
+		Xlocate(0, L_CMDLINE);
+		Xputterm(L_CLEAR);
+		Xcprintf2("[%.*k]", n_column - 2, path);
 		if (yesno(FOUND_K)) {
 			destpath = strdup2(path);
 			return(-2);
 		}
 	}
+
 	return(0);
 }
 
@@ -614,17 +639,18 @@ char *path;
 			return(-2);
 		}
 	}
+
 	return(0);
 }
 
-static VOID NEAR showattr(namep, attr, y)
+static VOID NEAR showattr(namep, attr, yy)
 namelist *namep;
 attrib_t *attr;
-int y;
+int yy;
 {
 	struct tm *tm;
 	char buf[WMODE + 1];
-	int x1, x2, w;
+	int x1, x2, y, w;
 
 	tm = localtime(&(namep -> st_mtim));
 	if (isbestomit()) {
@@ -637,87 +663,94 @@ int y;
 		x2 = n_column / 2;
 		w = 16;
 	}
+	y = yy;
 
-	locate(0, y);
-	putterm(l_clear);
+	Xlocate(0, y);
+	Xputterm(L_CLEAR);
 
-	locate(0, ++y);
-	putterm(l_clear);
-	locate(x1, y);
-	cprintf2("[%-*.*k]", w, w, namep -> name);
-	locate(x2 + 3, y);
-	kanjiputs(TOLD_K);
-	locate(x2 + 13, y);
-	kanjiputs(TNEW_K);
+	Xlocate(0, ++y);
+	Xputterm(L_CLEAR);
+	Xlocate(x1, y);
+	Xcprintf2("[%-*.*k]", w, w, namep -> name);
+	Xlocate(x2 + 3, y);
+	Xkanjiputs(TOLD_K);
+	Xlocate(x2 + 13, y);
+	Xkanjiputs(TNEW_K);
 
-	locate(0, ++y);
-	putterm(l_clear);
-	locate(x1, y);
-	kanjiputs(TMODE_K);
-	locate(x2, y);
+	Xlocate(0, ++y);
+	Xputterm(L_CLEAR);
+	Xlocate(x1, y);
+	Xkanjiputs(TMODE_K);
+	Xlocate(x2, y);
 	putmode(buf, namep -> st_mode, 1);
-	cputs2(buf);
-	locate(x2 + 10, y);
+	Xcputs2(buf);
+	Xlocate(x2 + ATTRWIDTH, y);
 	putmode(buf, attr -> mode, 1);
-	cputs2(buf);
+	Xcputs2(buf);
 
 #ifdef	HAVEFLAGS
-	locate(0, ++y);
-	putterm(l_clear);
-	locate(x1, y);
-	kanjiputs(TFLAG_K);
-	locate(x2, y);
+	Xlocate(0, ++y);
+	Xputterm(L_CLEAR);
+	Xlocate(x1, y);
+	Xkanjiputs(TFLAG_K);
+	Xlocate(x2, y);
 	putflags(buf, namep -> st_flags);
-	cputs2(buf);
-	locate(x2 + 10, y);
+	Xcputs2(buf);
+	Xlocate(x2 + ATTRWIDTH, y);
 	putflags(buf, attr -> flags);
-	cputs2(buf);
+	Xcputs2(buf);
 #endif
 
-	locate(0, ++y);
-	putterm(l_clear);
-	locate(x1, y);
-	kanjiputs(TDATE_K);
-	locate(x2, y);
-	cprintf2("%02d-%02d-%02d",
+	Xlocate(0, ++y);
+	Xputterm(L_CLEAR);
+	Xlocate(x1, y);
+	Xkanjiputs(TDATE_K);
+	Xlocate(x2, y);
+	Xcprintf2("%02d-%02d-%02d",
 		tm -> tm_year % 100, tm -> tm_mon + 1, tm -> tm_mday);
-	locate(x2 + 10, y);
-	cputs2(attr -> timestr[0]);
+	Xlocate(x2 + ATTRWIDTH, y);
+	Xcputs2(attr -> timestr[0]);
 
-	locate(0, ++y);
-	putterm(l_clear);
-	locate(x1, y);
-	kanjiputs(TTIME_K);
-	locate(x2, y);
-	cprintf2("%02d:%02d:%02d",
+	Xlocate(0, ++y);
+	Xputterm(L_CLEAR);
+	Xlocate(x1, y);
+	Xkanjiputs(TTIME_K);
+	Xlocate(x2, y);
+	Xcprintf2("%02d:%02d:%02d",
 		tm -> tm_hour, tm -> tm_min, tm -> tm_sec);
-	locate(x2 + 10, y);
-	cputs2(attr -> timestr[1]);
+	Xlocate(x2 + ATTRWIDTH, y);
+	Xcputs2(attr -> timestr[1]);
 
-	locate(0, ++y);
-	putterm(l_clear);
+	Xlocate(0, ++y);
+	Xputterm(L_CLEAR);
 }
 
 int inputattr(namep, flag)
 namelist *namep;
 int flag;
 {
+#if	!MSDOS
+	u_int tmp;
+#endif
 	struct tm *tm;
 	attrib_t attr;
 	char buf[WMODE + 1];
-	u_int tmp;
-	int i, ch, x, y, xx, yy, ymin, ymax, dupwin_x, dupwin_y;
+	u_int mask;
+	int ch, x, y, xx, yy, ymin, ymax, dupwin_x, dupwin_y, excl;
 
 	dupwin_x = win_x;
 	dupwin_y = win_y;
 	subwindow = 1;
 	Xgetkey(-1, 0);
 
-	yy = LFILETOP;
+	yy = filetop(win);
 	while (yy + WMODELINE + 5 > n_line - 1) yy--;
 	if (yy <= L_TITLE) yy = L_TITLE + 1;
-	xx = n_column / 2 + ((isbestomit()) ? 7 : 10);
+	xx = n_column / 2 + ((isbestomit()) ? 7 : ATTRWIDTH);
 
+	excl = (flag & ATR_EXCLUSIVE);
+	attr.nlink = TCH_CHANGE;
+	if (flag & ATR_MULTIPLE) attr.nlink |= (TCH_MODE | TCH_MTIME);
 	attr.mode = namep -> st_mode;
 #ifdef	HAVEFLAGS
 	attr.flags = namep -> st_flags;
@@ -728,17 +761,22 @@ int flag;
 	snprintf2(attr.timestr[1], sizeof(attr.timestr[1]), "%02d:%02d:%02d",
 		tm -> tm_hour, tm -> tm_min, tm -> tm_sec);
 	showattr(namep, &attr, yy);
-	y = ymin = (flag & 1) ? 0 : WMODELINE;
-	ymax = (flag & 2) ? WMODELINE + 1 : 0;
+	y = ymin = (excl == ATR_TIMEONLY) ? WMODELINE : 0;
+	ymax = (excl == ATR_MODEONLY) ? 0 : WMODELINE + 1;
 	x = 0;
 
 	do {
 		win_x = xx + x;
 		win_y = yy + y + 2;
-		locate(win_x, win_y);
-		tflush();
+		Xlocate(win_x, win_y);
+		Xtflush();
 
 		keyflush();
+#if	MSDOS
+		if (x == 3) mask = S_ISVTX;
+		else
+#endif
+		mask = (1 << (8 - x));
 		switch (ch = Xgetkey(1, 0)) {
 			case K_UP:
 				if (y > ymin) y--;
@@ -754,21 +792,20 @@ int flag;
 				break;
 			case 'a':
 			case 'A':
-				if (ymin <= 0) x = y = 0;
+				if (excl && excl != ATR_MODEONLY) break;
+				x = y = 0;
 				break;
 			case 'd':
 			case 'D':
-				if (ymax >= WMODELINE) {
-					x = 0;
-					y = WMODELINE;
-				}
+				if (excl && excl != ATR_TIMEONLY) break;
+				x = 0;
+				y = WMODELINE;
 				break;
 			case 't':
 			case 'T':
-				if (ymax >= WMODELINE + 1) {
-					x = 0;
-					y = WMODELINE + 1;
-				}
+				if (excl && excl != ATR_TIMEONLY) break;
+				x = 0;
+				y = WMODELINE + 1;
 				break;
 			case '0':
 			case '1':
@@ -781,8 +818,9 @@ int flag;
 			case '8':
 			case '9':
 				if (y < WMODELINE) break;
-				putch2(ch);
+				Xputch2(ch);
 				attr.timestr[y - WMODELINE][x] = ch;
+				attr.nlink |= TCH_MTIME;
 /*FALLTHRU*/
 			case K_RIGHT:
 #ifdef	HAVEFLAGS
@@ -828,45 +866,41 @@ int flag;
 				else if (x > 0) x--;
 				break;
 			case K_CTRL('L'):
-				yy = LFILETOP;
+				yy = filetop(win);
 				while (yy + WMODELINE + 5 > n_line - 1) yy--;
 				if (yy <= L_TITLE) yy = L_TITLE + 1;
-				xx = n_column / 2 + ((isbestomit()) ? 7 : 10);
+				xx = n_column / 2
+					+ ((isbestomit()) ? 7 : ATTRWIDTH);
 				showattr(namep, &attr, yy);
 				break;
 			case ' ':
 #ifdef	HAVEFLAGS
 				if (y == WMODELINE - 1) {
 					attr.flags ^= fflaglist[x];
-					locate(xx, yy + y + 2);
+					Xlocate(xx, yy + y + 2);
 					putflags(buf, attr.flags);
-					cputs2(buf);
+					Xcputs2(buf);
+					attr.nlink |= TCH_FLAGS;
 					break;
 				}
 #endif
 				if (y) break;
 #if	MSDOS
 				if (x == 2) break;
-				else if (x == 3) attr.mode ^= S_ISVTX;
-				else {
-					tmp = 1;
-					for (i = 8; i > x; i--) tmp <<= 1;
-					attr.mode ^= tmp;
+#else	/* !MSDOS */
+				if (!((x + 1) % 3) && (attr.mode & mask)) {
+					tmp = (1 << (12 - ((x + 1) / 3)));
+					if (attr.mode & tmp) {
+						attr.mode ^= tmp;
+					}
+					else mask = tmp;
 				}
-#else
-				tmp = 1;
-				for (i = 8; i > x; i--) tmp <<= 1;
-				if (!((x + 1) % 3) && (attr.mode & tmp)) {
-					i = (x * 2) / 3 + 4;
-					if (!(attr.mode & (tmp << i)))
-						tmp <<= i;
-					else attr.mode ^= (tmp << i);
-				}
-				attr.mode ^= tmp;
-#endif
-				locate(xx, yy + y + 2);
+#endif	/* !MSDOS */
+				attr.mode ^= mask;
+				Xlocate(xx, yy + y + 2);
 				putmode(buf, attr.mode, 1);
-				cputs2(buf);
+				Xcputs2(buf);
+				attr.nlink |= TCH_MODE;
 				break;
 			default:
 				break;
@@ -900,18 +934,29 @@ int flag;
 	|| tm -> tm_hour > 23 || tm -> tm_min > 59 || tm -> tm_sec > 59)
 		return(-1);
 
-	attrmode = (flag & 1) ? attr.mode : 0xffff;
+	mask = TCH_CHANGE;
+	switch (excl) {
+		case ATR_MODEONLY:
+			mask |= TCH_MODE;
+			break;
+		case ATR_TIMEONLY:
+			mask |= TCH_MTIME;
+			break;
+		default:
+			if (attrmode != namep -> st_mode) mask |= TCH_MODE;
 #ifdef	HAVEFLAGS
-	attrflags = (flag & 1) ? attr.flags : 0xffffffff;
+			if (attrflags != namep -> st_flags) mask |= TCH_FLAGS;
 #endif
-	attrtime = (flag & 2) ? timelocal2(tm) : (time_t)-1;
-	if (flag == 3) {
-		if (attrmode == namep -> st_mode) attrmode = 0xffff;
-#ifdef	HAVEFLAGS
-		if (attrflags == namep -> st_flags) attrflags = 0xffffffff;
-#endif
-		if (attrtime == namep -> st_mtim) attrtime = (time_t)-1;
+			if (attrtime != namep -> st_mtim) mask |= TCH_MTIME;
+			break;
 	}
+	attrnlink = (attr.nlink & mask);
+	attrmode = attr.mode;
+#ifdef	HAVEFLAGS
+	attrflags = attr.flags;
+#endif
+	attrtime = timelocal2(tm);
+
 	return(1);
 }
 
@@ -920,14 +965,13 @@ char *path;
 {
 	struct stat st;
 
-	st.st_mtime = attrtime;
+	st.st_nlink = attrnlink;
+	st.st_mode = attrmode;
 #ifdef	HAVEFLAGS
 	st.st_flags = attrflags;
 #endif
-	st.st_mode = attrmode;
-#ifndef	NOUID
-	st.st_gid = (gid_t)-1;
-#endif
+	st.st_mtime = attrtime;
+
 	return(touchfile(path, &st));
 }
 
@@ -945,7 +989,7 @@ char *endmes;
 
 	if (mark <= 0) {
 		i = (*func)(fnodospath(path, filepos));
-		if (i < 0) warning(-1, filelist[filepos].name);
+		if (i == -1) warning(-1, filelist[filepos].name);
 		else if (!i) ret++;
 		return(ret);
 	}
@@ -959,9 +1003,13 @@ char *endmes;
 		if (!ismark(&(filelist[filepos]))) continue;
 
 		movepos(old, 0);
-		locate(win_x, win_y);
-		tflush();
+		Xlocate(win_x, win_y);
+		Xtflush();
 		i = (*func)(fnodospath(path, filepos));
+		if (i < -1) {
+			endmes = NULL;
+			break;
+		}
 		if (i < 0) warning(-1, filelist[filepos].name);
 		else if (!i && ret == filepos) ret++;
 
@@ -971,8 +1019,9 @@ char *endmes;
 	if (endmes) warning(0, endmes);
 	filepos = dupfilepos;
 	movepos(old, 0);
-	locate(win_x, win_y);
-	tflush();
+	Xlocate(win_x, win_y);
+	Xtflush();
+
 	return(ret);
 }
 
@@ -992,8 +1041,7 @@ int *maxp, depth;
 		if (isdotdir(dp -> d_name)) continue;
 		strcpy(cp, dp -> d_name);
 
-		if (Xlstat(dir, &st) < 0 || (st.st_mode & S_IFMT) != S_IFDIR)
-			continue;
+		if (Xlstat(dir, &st) < 0 || !s_isdir(&st)) continue;
 		dirlist = b_realloc(dirlist, *maxp, char *);
 		dirlist[(*maxp)++] = strdup2(dir);
 		if (!(d = depth)) /*EMPTY*/;
@@ -1003,6 +1051,7 @@ int *maxp, depth;
 	}
 	if (cp > dir) *(cp - 1) = '\0';
 	Xclosedir(dirp);
+
 	return(dirlist);
 }
 
@@ -1020,31 +1069,33 @@ int verbose;
 	struct stat st;
 	time_t dupmtime, dupatime;
 	char *cp, *fname, path[MAXPATHLEN], **dirlist;
-	int ret, ndir, max;
+	int ret, ndir, max, dupnlink;
 
 	if (intrkey()) return(-2);
 
 	if (verbose) {
-		locate(0, L_CMDLINE);
-		putterm(l_clear);
+		Xlocate(0, L_CMDLINE);
+		Xputterm(L_CLEAR);
 		cp = (dir[0] == '.' && dir[1] == _SC_) ? dir + 2 : dir;
-		cprintf2("[%.*k]", n_column - 2, cp);
-		tflush();
+		Xcprintf2("[%.*k]", n_column - 2, cp);
+		Xtflush();
 	}
 #ifdef	FAKEUNINIT
 	else cp = NULL;	/* fake for -Wuninitialized */
 #endif
 
-	if (!funcd1 && !funcd2) order = 0;
+	if (!funcd1) order = (funcd2) ? ORD_NOPREDIR : ORD_NODIR;
 	strcpy(path, dir);
 
 	ret = max = 0;
 	if (!order) dirlist = NULL;
-	else dirlist = getdirtree(path, NULL, &max, (order == 2) ? 0 : 1);
+	else dirlist = getdirtree(path,
+		NULL, &max, (order == ORD_LOWER) ? 0 : 1);
 	fname = strcatdelim(path);
 
-	destmtime = destatime = (time_t)-1;
-	if ((order == 1 || order == 2) && (ret = (*funcd1)(dir)) < 0) {
+	destnlink = 0;
+	if ((order == ORD_NORMAL || order == ORD_LOWER)
+	&& (ret = (*funcd1)(dir)) < 0) {
 		if (ret == -1) warning(-1, dir);
 		if (dirlist) {
 			for (ndir = 0; ndir < max; ndir++) free(dirlist[ndir]);
@@ -1052,18 +1103,20 @@ int verbose;
 		}
 		return(ret);
 	}
+	dupnlink = destnlink;
 	dupmtime = destmtime;
 	dupatime = destatime;
 
-	if (order) for (ndir = 0; ndir < max; ndir++) {
-		if (order != 2) ret = _applydir(dirlist[ndir], funcf,
+	if (order != ORD_NODIR) for (ndir = 0; ndir < max; ndir++) {
+		if (order != ORD_LOWER)
+			ret = _applydir(dirlist[ndir], funcf,
 				funcd1, funcd2, order, NULL, verbose);
 		else if ((ret = (*funcd1)(dirlist[ndir])) < 0) {
 			if (ret == -1) warning(-1, dir);
 			ret = -2;
 		}
 		else ret = _applydir(dirlist[ndir], funcf,
-			funcd1, funcd2, 0, NULL, verbose);
+			funcd1, funcd2, ORD_NODIR, NULL, verbose);
 		if (ret < -1) {
 			for (; ndir < max; ndir++) free(dirlist[ndir]);
 			free(dirlist);
@@ -1072,10 +1125,10 @@ int verbose;
 		free(dirlist[ndir]);
 
 		if (verbose) {
-			locate(0, L_CMDLINE);
-			putterm(l_clear);
-			cprintf2("[%.*k]", n_column - 2, cp);
-			tflush();
+			Xlocate(0, L_CMDLINE);
+			Xputterm(L_CLEAR);
+			Xcprintf2("[%.*k]", n_column - 2, cp);
+			Xtflush();
 		}
 	}
 	if (dirlist) free(dirlist);
@@ -1091,7 +1144,7 @@ int verbose;
 			strcpy(fname, dp -> d_name);
 
 			if (Xlstat(path, &st) < 0) warning(-1, path);
-			else if ((st.st_mode & S_IFMT) == S_IFDIR) continue;
+			else if (s_isdir(&st)) continue;
 			else if ((ret = (*funcf)(path)) < 0) {
 				if (ret < -1) break;
 				warning(-1, path);
@@ -1101,10 +1154,15 @@ int verbose;
 		if (ret < -1) return(ret);
 	}
 
+	destnlink = dupnlink;
 	destmtime = dupmtime;
 	destatime = dupatime;
-	if (funcd2 && (ret = (*funcd2)(dir)) == -1) warning(-1, dir);
+	if (funcd2 && (ret = (*funcd2)(dir)) < 0) {
+		if (ret < -1) return(ret);
+		warning(-1, dir);
+	}
 	if (endmes) warning(0, endmes);
+
 	return(ret);
 }
 
@@ -1124,13 +1182,14 @@ char *endmes;
 		dir = ".";
 		verbose = 0;
 	}
-	else if (dir[0] == '.' && dir[1] == '.' && !dir[2]) {
+	else if (isdotdir(dir) == 1) {
 		realpath2(dir, path, 0);
 		dir = path;
 	}
 #ifdef	_USEDOSEMU
 	else dir = nodospath(path, dir);
 #endif
+
 	return(_applydir(dir, funcf, funcd1, funcd2, order, endmes, verbose));
 }
 
@@ -1139,9 +1198,11 @@ int copyfile(arg, tr)
 char *arg;
 int tr;
 {
+	int order;
+
 	if (!mark && isdotdir(filelist[filepos].name)) {
-		putterm(t_bell);
-		return(0);
+		Xputterm(T_BELL);
+		return(FNC_NONE);
 	}
 #ifdef	_NOTREE
 	destpath = getdestdir(COPYD_K, arg);
@@ -1153,7 +1214,7 @@ int tr;
 # endif
 #endif	/* !_NOTREE */
 
-	if (!destpath) return((tr) ? 2 : 1);
+	if (!destpath) return((tr) ? FNC_UPDATE : FNC_CANCEL);
 	destdir = NULL;
 	copypolicy = (issamedir(destpath, NULL)) ? (FLAG_SAMEDIR | 2) : 0;
 #ifndef	_NODOSDRIVE
@@ -1167,12 +1228,12 @@ int tr;
 		if (copypolicy) {
 			warning(EEXIST, filelist[filepos].name);
 			free(destpath);
-			return((tr) ? 2 : 1);
+			return((tr) ? FNC_UPDATE : FNC_CANCEL);
 		}
 #endif
-		applydir(filelist[filepos].name, safecopy, cpdir, touchdir,
-			(islowerdir(destpath, filelist[filepos].name)) ? 2 : 1,
-			ENDCP_K);
+		order = (islowerdir()) ? ORD_LOWER : ORD_NORMAL;
+		applydir(filelist[filepos].name, safecopy,
+			cpdir, touchdir, order, ENDCP_K);
 	}
 
 	free(destpath);
@@ -1188,7 +1249,8 @@ int tr;
 	if (forwarddrive >= 0) shutdrv(forwarddrive);
 	forwarddrive = -1;
 #endif
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 /*ARGSUSED*/
@@ -1196,11 +1258,17 @@ int movefile(arg, tr)
 char *arg;
 int tr;
 {
+#ifdef	_USEDOSEMU
+	char path[MAXPATHLEN];
+#endif
+#ifndef	_NOEXTRACOPY
+	int order;
+#endif
 	int i;
 
 	if (!mark && isdotdir(filelist[filepos].name)) {
-		putterm(t_bell);
-		return(0);
+		Xputterm(T_BELL);
+		return(FNC_NONE);
 	}
 #ifdef	_NOTREE
 	destpath = getdestdir(MOVED_K, arg);
@@ -1212,33 +1280,31 @@ int tr;
 # endif
 #endif	/* !_NOTREE */
 
-	if (!destpath || issamedir(destpath, NULL)) return((tr) ? 2 : 1);
+	if (!destpath || issamedir(destpath, NULL))
+		return((tr) ? FNC_UPDATE : FNC_CANCEL);
 	destdir = NULL;
 	copypolicy = removepolicy = 0;
 	if (mark > 0) filepos = applyfile(safemove, ENDMV_K);
-	else if (islowerdir(destpath, filelist[filepos].name))
-		warning(EINVAL, filelist[filepos].name);
+	else if (islowerdir()) warning(EINVAL, filelist[filepos].name);
 	else {
-#ifdef	_USEDOSEMU
-		char path[MAXPATHLEN];
-#endif
-
 		i = safemove(fnodospath(path, filepos));
 		if (!i) filepos++;
-		else if (i < 0) {
+		else if (i == -1) {
 #ifdef	_NOEXTRACOPY
 			warning(-1, filelist[filepos].name);
-#else
+#else	/* !_NOEXTRACOPY */
 			if (errno != EXDEV
 			|| !isdir(&(filelist[filepos]))
 			|| islink(&(filelist[filepos])))
 				warning(-1, filelist[filepos].name);
-			else if (applydir(filelist[filepos].name,
-				safemove, cpdir, mvdir,
-				(islowerdir(destpath, filelist[filepos].name))
-				? 2 : 1,
-				ENDMV_K) >= 0) filepos++;
-#endif
+			else {
+				order = (islowerdir())
+					? ORD_LOWER : ORD_NORMAL;
+				i = applydir(filelist[filepos].name, safemove,
+					cpdir, mvdir, order, ENDMV_K);
+				if (i >= 0) filepos++;
+			}
+#endif	/* !_NOEXTRACOPY */
 		}
 	}
 
@@ -1256,7 +1322,8 @@ int tr;
 	if (forwarddrive >= 0) shutdrv(forwarddrive);
 	forwarddrive = -1;
 #endif
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 static int forcecpfile(path)
@@ -1268,8 +1335,10 @@ char *path;
 	if (getdestpath(path, dest, &st) < 0) return(0);
 	if (safecpfile(path, dest, &st, NULL) < 0) return(-1);
 #ifdef	HAVEFLAGS
-	st.st_flags = 0xffffffff;
+	st.st_flags = (u_long)-1;
 #endif
+
+	st.st_nlink = (TCH_MODE | TCH_UID | TCH_GID | TCH_ATIME | TCH_MTIME);
 	return(touchfile(dest, &st));
 }
 
@@ -1278,21 +1347,24 @@ char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
+	int mode;
 
 	if (isdotdir(path)) return(0);
 	if (getdestpath(path, dest, &st) < 0) return(0);
+	mode = (st.st_mode & 0777);
 #if	MSDOS
-	while (Xmkdir(dest, (st.st_mode & 0777) | S_IWRITE) < 0)
-#else
-	while (Xmkdir(dest, st.st_mode & 0777) < 0)
+	mode |= S_IWRITE;
 #endif
-	{
-		if (errno != EEXIST) return(-1);
-		if (Xstat(dest, &st) < 0) return(-1);
-		if ((st.st_mode & S_IFMT) == S_IFDIR) break;
-		if (Xunlink(dest) < 0) return(-1);
+	if (Xmkdir(dest, mode) >= 0) return(0);
+	if (errno != EEXIST) return(-1);
+	if (stat2(dest, &st) >= 0) {
+		if (s_isdir(&st)) return(0);
+		if (Xunlink(dest) >= 0 && Xmkdir(dest, mode) >= 0) return(0);
 	}
-	return(0);
+
+	errno = EEXIST;
+
+	return(-1);
 }
 
 static int forcetouchdir(path)
@@ -1304,8 +1376,10 @@ char *path;
 	if (isdotdir(path)) return(0);
 	if (getdestpath(path, dest, &st) < 0) return(-2);
 #ifdef	HAVEFLAGS
-	st.st_flags = 0xffffffff;
+	st.st_flags = (u_long)-1;
 #endif
+
+	st.st_nlink = (TCH_MODE | TCH_UID | TCH_GID | TCH_ATIME | TCH_MTIME);
 	return(touchfile(dest, &st));
 }
 
@@ -1317,5 +1391,6 @@ char *dest;
 	destpath = dest;
 	ret = applydir(NULL, forcecpfile, forcecpdir, forcetouchdir, 1, NULL);
 	destpath = NULL;
+
 	return(ret);
 }
