@@ -1,24 +1,26 @@
 /*
  *	command.c
  *
- *	Command Execute Module
+ *	command executing module
  */
 
 #include "fd.h"
-#include "term.h"
 #include "func.h"
 #include "funcno.h"
-#include "kctype.h"
 #include "kanji.h"
 
 #ifndef	_NOORIGSHELL
 #include "system.h"
 #endif
 
-#ifndef	_NODOSDRIVE
-extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
+#ifndef	_NOPTY
+#include "termemu.h"
 #endif
 
+#define	MAXATTRSEL	2
+#define	ATTR_X		35
+#define	ATTR_Y		L_INFO
+
 extern int curcolumns;
 extern int mark;
 extern off_t marksize;
@@ -35,13 +37,19 @@ extern char archivedir[];
 extern int win_x;
 extern int win_y;
 extern char *destpath;
-extern char *histfile;
-extern int savehist;
+#ifndef	_NOTRADLAYOUT
+extern int tradlayout;
+#endif
 extern int sizeinfo;
+extern int wheader;
 extern char fullpath[];
+extern int hideclock;
 #ifndef	_NOORIGSHELL
 extern int fdmode;
 #endif
+#ifndef	_NOPTY
+extern int ptymode;
+#endif
 
 static VOID NEAR replacefname __P_((char *));
 static int cur_up __P_((char *));
@@ -76,6 +84,7 @@ static int dotfile_mode __P_((char *));
 static int fileflg_mode __P_((char *));
 static int log_dir __P_((char *));
 static int log_top __P_((char *));
+static VOID NEAR clearscreen __P_((VOID_A));
 #ifndef	PAGER
 static VOID NEAR dump __P_((char *));
 #endif
@@ -92,6 +101,9 @@ static int help_message __P_((char *));
 #ifndef	_NOCUSTOMIZE
 static int edit_config __P_((char *));
 #endif
+#ifndef	_NOPTY
+static int confirmpty __P_((VOID_A));
+#endif
 static int quit_system __P_((char *));
 static int make_dir __P_((char *));
 static int copy_file __P_((char *));
@@ -113,11 +125,12 @@ static int execute_file __P_((char *));
 static int launch_file __P_((char *));
 static int pack_file __P_((char *));
 static int unpack_file __P_((char *));
-#ifndef	_NOTREE
+# ifndef	_NOTREE
 static int unpack_tree __P_((char *));
-#endif
+# endif
 #endif	/* !_NOARCHIVE */
 static int info_filesys __P_((char *));
+static int NEAR selectattr __P_((char *));
 static int attr_file __P_((char *));
 #ifndef	_NOTREE
 static int tree_dir __P_((char *));
@@ -133,6 +146,11 @@ static VOID NEAR movewin __P_((int));
 static int split_window __P_((char *));
 static int next_window __P_((char *));
 #endif
+#ifndef	_NOEXTRAWIN
+static int widen_window __P_((char *));
+static int narrow_window __P_((char *));
+static int kill_window __P_((char *));
+#endif
 static int warning_bell __P_((char *));
 static int no_operation __P_((char *));
 
@@ -142,7 +160,10 @@ reg_t *findregexp = NULL;
 #ifndef	_NOWRITEFS
 int writefs = 0;
 #endif
-int internal_status = -2;
+#if	FD >= 2
+int loopcursor = 0;
+#endif
+int internal_status = FNC_FAIL;
 bindtable bindlist[MAXBINDTABLE] = {
 	{K_UP,		CUR_UP,		255},
 	{K_DOWN,	CUR_DOWN,	255},
@@ -267,6 +288,11 @@ bindtable bindlist[MAXBINDTABLE] = {
 #ifndef	_NOSPLITWIN
 	{'/',		SPLIT_WINDOW,	255},
 	{'^',		NEXT_WINDOW,	255},
+#endif
+#ifndef	_NOEXTRAWIN
+	{'W',		WIDEN_WINDOW,	255},
+	{'N',		NARROW_WINDOW,	255},
+	{'K',		KILL_WINDOW,	255},
 #endif
 	{K_HOME,	MARK_ALL,	255},
 	{K_END,		MARK_REVERSE,	255},
@@ -294,103 +320,183 @@ char *name;
 static int cur_up(arg)
 char *arg;
 {
-	int n;
+	int n, old;
 
-	if (filepos <= 0) return(0);
+	old = filepos;
 	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
-	if ((filepos -= n) < 0) filepos = 0;
-	return(2);
+	filepos -= n;
+#if	FD >= 2
+	if (loopcursor) {
+		int p;
+
+		p = FILEPERPAGE;
+		while (filepos < 0 || filepos / p < old / p)
+			filepos += p;
+		if (filepos >= maxfile) filepos = maxfile - 1;
+	}
+	else
+#endif
+	if (filepos < 0) filepos = 0;
+
+	return((filepos != old) ? FNC_UPDATE : FNC_NONE);
 }
 
 static int cur_down(arg)
 char *arg;
 {
-	int n;
+	int n, old;
 
-	if (filepos >= maxfile - 1) return(0);
+	old = filepos;
 	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
-	if ((filepos += n) > maxfile - 1) filepos = maxfile - 1;
-	return(2);
+	filepos += n;
+#if	FD >= 2
+	if (loopcursor) {
+		int p;
+
+		p = FILEPERPAGE;
+		if (filepos >= maxfile) filepos = (filepos / p) * p;
+		while (filepos / p > old / p) {
+			filepos -= p;
+			if (filepos < 0) break;
+		}
+		if (filepos < 0) filepos = 0;
+	}
+	else
+#endif
+	if (filepos >= maxfile) filepos = maxfile - 1;
+
+	return((filepos != old) ? FNC_UPDATE : FNC_NONE);
 }
 
 static int cur_right(arg)
 char *arg;
 {
-	int n;
+	int n, r, p, old;
 
+	old = filepos;
 	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
-	if (filepos + FILEPERROW * n < maxfile) filepos += FILEPERROW * n;
-	else if (filepos / FILEPERPAGE == (maxfile - 1) / FILEPERPAGE)
-		return(0);
+	r = FILEPERROW;
+	p = FILEPERPAGE;
+	filepos += r * n;
+
+#if	FD >= 2
+	if (loopcursor) {
+		if (filepos >= maxfile)
+			filepos = (filepos / p) * p + (old % r);
+		while (filepos / p > old / p) {
+			filepos -= p;
+			if (filepos < 0) break;
+		}
+		if (filepos < 0) filepos = 0;
+	}
+	else
+#endif
+	if (filepos < maxfile) /*EMPTY*/;
+	else if (old / p == (maxfile - 1) / p) filepos = old;
 	else {
-		filepos = ((maxfile - 1) / FILEPERROW) * FILEPERROW
-			+ filepos % FILEPERROW;
+		filepos = ((maxfile - 1) / r) * r + (old % r);
 		if (filepos >= maxfile) {
-			filepos -= FILEPERROW;
-			if (filepos / FILEPERPAGE
-			< (maxfile - 1) / FILEPERPAGE)
+			filepos -= r;
+			if (filepos / p < (maxfile - 1) / p)
 				filepos = maxfile - 1;
 		}
 	}
-	return(2);
+
+	return((filepos != old) ? FNC_UPDATE : FNC_NONE);
 }
 
 static int cur_left(arg)
 char *arg;
 {
-	int n;
+	int n, r, old;
 
-	if (filepos < FILEPERROW) return(0);
+	old = filepos;
 	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
-	if (filepos - FILEPERROW * n >= 0) filepos -= FILEPERROW * n;
-	else filepos = filepos % FILEPERROW;
-	return(2);
+	r = FILEPERROW;
+	filepos -= r * n;
+
+#if	FD >= 2
+	if (loopcursor) {
+		int p;
+
+		p = FILEPERPAGE;
+		while (filepos < 0 || filepos / p < old / p)
+			filepos += p;
+		if (filepos >= maxfile) {
+			filepos = (maxfile - 1) / r * r + (old % r);
+			if (filepos >= maxfile) filepos -= r;
+		}
+	}
+	else
+#endif
+	if (filepos >= 0) /*EMPTY*/;
+	else filepos = old % r;
+
+	return((filepos != old) ? FNC_UPDATE : FNC_NONE);
 }
 
 static int roll_up(arg)
 char *arg;
 {
-	int i;
+	int n, p, old;
 
-	i = (filepos / FILEPERPAGE + 1) * FILEPERPAGE;
-	if (i >= maxfile) return(warning_bell(arg));
-	filepos = i;
-	return(2);
+	old = filepos;
+	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
+	p = FILEPERPAGE;
+	filepos = (filepos / p + n) * p;
+	if (filepos >= maxfile) {
+		filepos = old;
+		return(warning_bell(arg));
+	}
+
+	return(FNC_UPDATE);
 }
 
 static int roll_down(arg)
 char *arg;
 {
-	if (filepos < FILEPERPAGE) return(warning_bell(arg));
-	filepos = (filepos / FILEPERPAGE - 1) * FILEPERPAGE;
-	return(2);
+	int n, p, old;
+
+	old = filepos;
+	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
+	p = FILEPERPAGE;
+	filepos = (filepos / p - n) * p;
+	if (filepos < 0) {
+		filepos = old;
+		return(warning_bell(arg));
+	}
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
 static int cur_top(arg)
 char *arg;
 {
-	if (filepos == 0) return(0);
+	if (filepos == 0) return(FNC_NONE);
 	filepos = 0;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
 static int cur_bottom(arg)
 char *arg;
 {
-	if (filepos == maxfile - 1) return(0);
+	if (filepos == maxfile - 1) return(FNC_NONE);
 	filepos = maxfile - 1;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
 static int fname_right(arg)
 char *arg;
 {
-	if (fnameofs <= 0) return(0);
+	if (fnameofs <= 0) return(FNC_NONE);
 	fnameofs--;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
@@ -411,7 +517,8 @@ char *arg;
 			if (!(filelist[filepos].linkname)) i = 0;
 			else {
 				i = strlen(filelist[filepos].linkname);
-				if (i > sizeof(tmp) - 1) i = sizeof(tmp) - 1;
+				if (i > (int)sizeof(tmp) - 1)
+					i = (int)sizeof(tmp) - 1;
 				strncpy(tmp, filelist[filepos].linkname, i);
 			}
 		}
@@ -429,12 +536,14 @@ char *arg;
 			m = filelist[filepos].st_mode;
 			if ((m & S_IFMT) == S_IFDIR || (m & S_IFMT) == S_IFLNK
 			|| (m & S_IFMT) == S_IFSOCK || (m & S_IFMT) == S_IFIFO
-			|| (m & (S_IXUSR | S_IXGRP | S_IXOTH))) i--;
+			|| (m & S_IEXEC_ALL))
+				i--;
 		}
 	}
-	if (i >= strlen3(filelist[filepos].name) - fnameofs) return(0);
+	if (i >= strlen3(filelist[filepos].name) - fnameofs) return(FNC_NONE);
 	fnameofs++;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
@@ -442,7 +551,8 @@ static int one_column(arg)
 char *arg;
 {
 	curcolumns = 1;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
@@ -450,7 +560,8 @@ static int two_columns(arg)
 char *arg;
 {
 	curcolumns = 2;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
@@ -458,7 +569,8 @@ static int three_columns(arg)
 char *arg;
 {
 	curcolumns = 3;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
@@ -466,20 +578,27 @@ static int five_columns(arg)
 char *arg;
 {
 	curcolumns = 5;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 static VOID NEAR markcount(VOID_A)
 {
-	char buf[14 + 1];
-	int x;
+#ifndef	_NOTRADLAYOUT
+	if (istradlayout()) {
+		Xlocate(TC_MARK, TL_PATH);
+		Xcprintf2("%<*d", TD_MARK, mark);
+		Xlocate(TC_MARK + TD_MARK + TW_MARK, TL_PATH);
+		Xcprintf2("%<'*qd", TD_SIZE, marksize);
 
-	x = (isleftshift()) ? 1 : 0;
-	locate(C_MARK + 5 - x, L_STATUS);
-	cputs2(ascnumeric(buf, (off_t)mark, 4, 4));
+		return;
+	}
+#endif
+	Xlocate(C_MARK + W_MARK, L_STATUS);
+	Xcprintf2("%<*d", D_MARK, mark);
 	if (sizeinfo) {
-		locate(C_SIZE + 5 - x, L_SIZE);
-		cputs2(ascnumeric(buf, marksize, 3, 14));
+		Xlocate(C_SIZE + W_SIZE, L_SIZE);
+		Xcprintf2("%<'*qd", D_SIZE, marksize);
 	}
 }
 
@@ -487,7 +606,7 @@ static VOID NEAR markcount(VOID_A)
 static int mark_file(arg)
 char *arg;
 {
-	if (isdir(&(filelist[filepos]))) return(0);
+	if (isdir(&(filelist[filepos]))) return(FNC_NONE);
 	filelist[filepos].tmpflags ^= F_ISMRK;
 	if (ismark(&(filelist[filepos]))) {
 		mark++;
@@ -500,7 +619,8 @@ char *arg;
 			marksize -= getblock(filelist[filepos].st_size);
 	}
 	markcount();
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 static int mark_file2(arg)
@@ -508,7 +628,8 @@ char *arg;
 {
 	mark_file(arg);
 	if (filepos < maxfile - 1) filepos++;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 static int mark_file3(arg)
@@ -516,9 +637,11 @@ char *arg;
 {
 	mark_file(arg);
 	if (filepos < maxfile - 1
-	&& filepos / FILEPERPAGE == (filepos + 1) / FILEPERPAGE) filepos++;
+	&& filepos / FILEPERPAGE == (filepos + 1) / FILEPERPAGE)
+		filepos++;
 	else filepos = (filepos / FILEPERPAGE) * FILEPERPAGE;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 static int mark_all(arg)
@@ -542,7 +665,8 @@ char *arg;
 		}
 	}
 	markcount();
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
@@ -560,7 +684,8 @@ char *arg;
 				marksize += getblock(filelist[i].st_size);
 		}
 	markcount();
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 static reg_t *NEAR prepareregexp(mes, arg)
@@ -579,6 +704,7 @@ char *mes, *arg;
 
 	re = regexp_init(wild, -1);
 	free(wild);
+
 	return(re);
 }
 
@@ -588,7 +714,7 @@ char *arg;
 	reg_t *re;
 	int i;
 
-	if (!(re = prepareregexp(FINDF_K, arg))) return(1);
+	if (!(re = prepareregexp(FINDF_K, arg))) return(FNC_CANCEL);
 	for (i = 0; i < maxfile; i++)
 		if (!isdir(&(filelist[i])) && !ismark(&(filelist[i]))
 		&& regexp_exec(re, filelist[i].name, 1)) {
@@ -599,7 +725,8 @@ char *arg;
 		}
 	regexp_free(re);
 	markcount();
-	return(3);
+
+	return(FNC_HELPSPOT);
 }
 
 static int in_dir(arg)
@@ -610,8 +737,10 @@ char *arg;
 	else
 #endif
 	if (!isdir(&(filelist[filepos]))
-	|| !strcmp(filelist[filepos].name, ".")) return(warning_bell(arg));
-	return(5);
+	|| !strcmp(filelist[filepos].name, "."))
+		return(warning_bell(arg));
+
+	return(FNC_CHDIR);
 }
 
 /*ARGSUSED*/
@@ -623,7 +752,8 @@ char *arg;
 	else
 #endif
 	replacefname(NULL);
-	return(5);
+
+	return(FNC_CHDIR);
 }
 
 /*ARGSUSED*/
@@ -633,7 +763,7 @@ char *arg;
 	int i;
 
 	if (stackdepth >= MAXSTACK || isdotdir(filelist[filepos].name))
-		return(0);
+		return(FNC_NONE);
 	memcpy((char *)&(filestack[stackdepth++]),
 		(char *)&(filelist[filepos]), sizeof(namelist));
 	maxfile--;
@@ -641,7 +771,8 @@ char *arg;
 		memcpy((char *)&(filelist[i]),
 			(char *)&(filelist[i + 1]), sizeof(namelist));
 	if (filepos >= maxfile) filepos = maxfile - 1;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
@@ -650,7 +781,7 @@ char *arg;
 {
 	int i;
 
-	if (stackdepth <= 0) return(0);
+	if (stackdepth <= 0) return(FNC_NONE);
 	for (i = maxfile; i > filepos + 1; i--)
 		memcpy((char *)&(filelist[i]),
 			(char *)&(filelist[i - 1]), sizeof(namelist));
@@ -658,7 +789,8 @@ char *arg;
 	memcpy((char *)&(filelist[filepos + 1]),
 		(char *)&(filestack[--stackdepth]), sizeof(namelist));
 	chgorder = 1;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
@@ -666,7 +798,8 @@ static int symlink_mode(arg)
 char *arg;
 {
 	dispmode ^= F_SYMLINK;
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 /*ARGSUSED*/
@@ -674,7 +807,8 @@ static int filetype_mode(arg)
 char *arg;
 {
 	dispmode ^= F_FILETYPE;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 /*ARGSUSED*/
@@ -682,7 +816,8 @@ static int dotfile_mode(arg)
 char *arg;
 {
 	dispmode ^= F_DOTFILE;
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 /*ARGSUSED*/
@@ -690,7 +825,8 @@ static int fileflg_mode(arg)
 char *arg;
 {
 	dispmode ^= F_FILEFLAG;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 static int log_dir(arg)
@@ -702,29 +838,30 @@ char *arg;
 	char *path;
 
 	if (arg && *arg) path = strdup2(arg);
-	else if (!(path = inputstr(LOGD_K, 0, -1, NULL, 1))) return(1);
-	else if (!*(path = evalpath(path, 1))) {
+	else if (!(path = inputstr(LOGD_K, 0, -1, NULL, HST_PATH)))
+		return(FNC_CANCEL);
+	else if (!*(path = evalpath(path, 0))) {
 		free(path);
-		return(1);
+		return(FNC_CANCEL);
 	}
 #ifndef	_NOARCHIVE
 	if (archivefile && *path != '/') {
 		if (!(cp = archchdir(path))) {
 			warning(-1, path);
 			free(path);
-			return(1);
+			return(FNC_CANCEL);
 		}
 		free(path);
 		if (cp != (char *)-1) filelist[filepos].name = cp;
 		else escapearch();
-		return(4);
+		return(FNC_EFFECT);
 	}
 	else
 #endif
-	if (chdir3(path) < 0) {
+	if (chdir3(path, 0) < 0) {
 		warning(-1, path);
 		free(path);
-		return(1);
+		return(FNC_CANCEL);
 	}
 	free(path);
 #ifndef	_NOARCHIVE
@@ -743,7 +880,8 @@ char *arg;
 	}
 #endif
 	replacefname(NULL);
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 /*ARGSUSED*/
@@ -753,11 +891,28 @@ char *arg;
 	char *path;
 
 	path = strdup2(_SS_);
-	if (chdir2(path) < 0) error(path);
-	if (findpattern) free(findpattern);
-	findpattern = NULL;
+	if (chdir3(path, 1) < 0) error(path);
 	replacefname(path);
-	return(4);
+
+	return(FNC_EFFECT);
+}
+
+static VOID NEAR clearscreen(VOID_A)
+{
+#ifndef	_NOPTY
+	int i, yy;
+
+	if (isptymode()) {
+		yy = filetop(win);
+		for (i = 0; i < FILEPERROW; i++) {
+			Xlocate(0, yy + i);
+			Xputterm(L_CLEAR);
+		}
+	}
+	else
+#endif	/* !_NOPTY */
+	Xputterms(T_CLEAR);
+	Xtflush();
 }
 
 #ifndef	PAGER
@@ -766,61 +921,75 @@ char *file;
 {
 	FILE *fp;
 	char *buf, *prompt;
-	int i;
+	int i, len, min, max;
 
 	if (!(fp = Xfopen(file, "r"))) {
 		warning(-1, file);
 		return;
 	}
-	putterms(t_clear);
-	tflush();
+	clearscreen();
+
+	min = 0;
+	max = n_line - 1;
+# ifndef	_NOPTY
+	if (isptymode()) {
+		min = filetop(win);
+		max = min + FILEPERROW;
+	}
+# endif
+
+	Xtflush();
 	buf = NEXT_K;
 	i = strlen3(file);
-	if (i + strlen(buf) > n_lastcolumn)
-		i = n_lastcolumn - (int)strlen(buf);
+	len = strlen2(buf);
+	if (i + len > n_lastcolumn) i = n_lastcolumn - len;
 	prompt = malloc2(i * KANAWID + strlen(buf) + 1);
 	strncpy3(prompt, file, &i, 0);
 	strcat(prompt, buf);
 
-	i = 0;
+	i = min;
 	while ((buf = fgets2(fp, 0))) {
-		locate(0, i);
-		putterm(l_clear);
-		kanjiputs2(buf, n_column, -1);
+		Xlocate(0, i);
+		Xputterm(L_CLEAR);
+		Xcprintf2("%.*k", n_column, buf);
 		free(buf);
-		if (++i >= n_line - 1) {
-			i = 0;
+		if (++i >= max) {
+			i = min;
+			hideclock = 1;
 			if (!yesno(prompt)) break;
 		}
 	}
 
 	if (Xfeof(fp)) {
-		while (i < n_line + 1) {
-			locate(0, i++);
-			putterm(l_clear);
+		while (i < max) {
+			Xlocate(0, i++);
+			Xputterm(L_CLEAR);
 		}
 	}
 	Xfclose(fp);
 	free(prompt);
 }
-#endif
+#endif	/* !PAGER */
 
 static int NEAR execenv(env, arg)
 char *env, *arg;
 {
 	char *command;
 
-	if (!(command = getenv2(env))) return(0);
-	putterms(t_clear);
-	tflush();
-	execmacro(command, arg, 1, 0, 1);
+	if (!(command = getenv2(env)) || !*command) return(0);
+	clearscreen();
+	ptymacro(command, arg, F_NOCONFIRM | F_IGNORELIST);
+
 	return(1);
 }
 
 static int NEAR execshell(VOID_A)
 {
+#ifndef	_NOPTY
+	int x, y, min, max;
+#endif
 	char *sh;
-	int mode, nl, ret;
+	int mode, ret, wastty;
 
 	sh = getenv2("FD_SHELL");
 #if	MSDOS
@@ -844,18 +1013,29 @@ static int NEAR execshell(VOID_A)
 #endif
 
 	sigvecset(0);
-	mode = termmode(0);
-	nl = stdiomode();
-	kanjifputs(SHEXT_K, stderr);
-	fputc('\n', stderr);
+	mode = Xtermmode(0);
+	wastty = isttyiomode;
+	kanjiputs(SHEXT_K);
+#ifndef	_NOPTY
+	if (isptymode()) {
+		min = filetop(win);
+		max = min + FILEPERROW;
+		y = max;
+		if (!isttyiomode) Xttyiomode(0);
+		if (getxy(&x, &y) < 0 || --y >= max) y = max - 1;
+		regionscroll(C_SCROLLFORW, 1, 0, y, min, max - 1);
+	}
+	else
+#endif	/* !_NOPTY */
+	Xcputnl();
+	Xtflush();
+	if (isttyiomode) Xstdiomode();
+	ret = ptysystem(sh);
 #ifdef	_NOORIGSHELL
-	ret = system(sh);
-#else
-	if (sh) ret = dosystem(sh);
-	else ret = shell_loop(1);
+	checkscreen(-1, -1);
 #endif
-	ttyiomode(nl);
-	termmode(mode);
+	if (wastty) Xttyiomode(wastty - 1);
+	Xtermmode(mode);
 	sigvecset(1);
 
 	return(ret);
@@ -878,22 +1058,24 @@ char *arg;
 #ifndef	_NODOSDRIVE
 	drive = 0;
 #endif
-	if (isdir(&(filelist[filepos]))) return(1);
+	if (isdir(&(filelist[filepos]))) return(FNC_CANCEL);
 #ifndef	_NOARCHIVE
 	else if (archivefile) {
-		if (!(dir = tmpunpack(1))) return(1);
+		if (!(dir = tmpunpack(1))) return(FNC_CANCEL);
 	}
 #endif
 #ifndef	_NODOSDRIVE
-	else if ((drive = tmpdosdupl("", &dir, 1)) < 0) return(1);
+	else if ((drive = tmpdosdupl("", &dir, 1)) < 0) return(FNC_CANCEL);
 #endif
 
 	if (!execenv("FD_PAGER", filelist[filepos].name)) {
 #ifdef	PAGER
-		execmacro(PAGER, filelist[filepos].name, 1, 0, 1);
+		ptymacro(PAGER, filelist[filepos].name,
+			F_NOCONFIRM | F_IGNORELIST);
 #else
 		do {
 			dump(filelist[filepos].name);
+			hideclock = 1;
 		} while (!yesno(PEND_K));
 #endif
 	}
@@ -907,7 +1089,8 @@ char *arg;
 	else
 #endif
 	/*EMPTY*/;
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 static int edit_file(arg)
@@ -922,14 +1105,15 @@ char *arg;
 
 	if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 #ifndef	_NOARCHIVE
-	if (archivefile) return(1);
+	if (archivefile) return(FNC_CANCEL);
 #endif
 #ifndef	_NODOSDRIVE
-	if ((drive = tmpdosdupl("", &dir, 1)) < 0) return(1);
+	if ((drive = tmpdosdupl("", &dir, 1)) < 0) return(FNC_CANCEL);
 #endif
 	if (!execenv("FD_EDITOR", filelist[filepos].name)) {
 #ifdef	EDITOR
-		execmacro(EDITOR, filelist[filepos].name, 1, 0, 1);
+		ptymacro(EDITOR, filelist[filepos].name,
+			F_NOCONFIRM | F_IGNORELIST);
 #endif
 	}
 #ifndef	_NODOSDRIVE
@@ -939,7 +1123,8 @@ char *arg;
 		removetmp(dir, filelist[filepos].name);
 	}
 #endif
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 static int sort_dir(arg)
@@ -976,7 +1161,8 @@ char *arg;
 			i = 5;
 			tmp1 = val[0];
 		}
-		if (selectstr(&tmp1, i, 0, str, val) == K_ESC) return(1);
+		if (selectstr(&tmp1, i, 0, str, val) != K_CR)
+			return(FNC_CANCEL);
 	}
 
 	if (!tmp1) {
@@ -988,8 +1174,8 @@ char *arg;
 		str[1] = ODEC_K;
 		val[0] = 0;
 		val[1] = 8;
-		if (selectstr(&tmp2, 2, 56, str, val) == K_ESC)
-			return(1);
+		if (selectstr(&tmp2, 2, 56, str, val) != K_CR)
+			return(FNC_CANCEL);
 		sorton = tmp1 + tmp2;
 		dupl = (int *)malloc2(maxfile * sizeof(int));
 		for (i = 0; i < maxfile; i++) {
@@ -1005,7 +1191,8 @@ char *arg;
 		free(dupl);
 		chgorder = 1;
 	}
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
 #ifndef	_NOWRITEFS
@@ -1017,15 +1204,16 @@ char *arg;
 	if (writefs >= 2 || findpattern) return(warning_bell(arg));
 	if ((fs = writablefs(".")) <= 0) {
 		warning(0, NOWRT_K);
-		return(1);
+		return(FNC_CANCEL);
 	}
-	if (!yesno(WRTOK_K)) return(1);
-	if (underhome(NULL) <= 0 && !yesno(HOMOK_K)) return(1);
+	if (!yesno(WRTOK_K)) return(FNC_CANCEL);
+	if (underhome(NULL) <= 0 && !yesno(HOMOK_K)) return(FNC_CANCEL);
 	arrangedir(fs);
 	chgorder = 0;
-	return(4);
+
+	return(FNC_EFFECT);
 }
-#endif
+#endif	/* !_NOWRITEFS */
 
 /*ARGSUSED*/
 static int reread_dir(arg)
@@ -1039,7 +1227,8 @@ char *arg;
 #ifndef	_NODOSDRIVE
 	if ((drive = dospath3(""))) flushdrv(drive, NULL);
 #endif
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 #ifndef	_NOCUSTOMIZE
@@ -1049,10 +1238,11 @@ char *arg;
 {
 	if (FILEPERROW < WFILEMINCUSTOM) {
 		warning(0, NOROW_K);
-		return(1);
+		return(FNC_CANCEL);
 	}
 	customize();
-	return(4);
+
+	return(FNC_EFFECT);
 }
 #endif
 
@@ -1065,19 +1255,59 @@ char *arg;
 #else
 	help(archivefile != NULL);
 #endif
-	return(2);
+
+	return(FNC_UPDATE);
 }
 
+#ifndef	_NOPTY
+static int confirmpty(VOID_A)
+{
+	int i;
+
+	for (i = 0; i < MAXWINDOWS; i++) if (ptylist[i].pid) break;
+	if (i < MAXWINDOWS && !yesno(KILL_K)) return(1);
+
+	return(0);
+}
+#endif	/* !_NOPTY */
+
 /*ARGSUSED*/
 static int quit_system(arg)
 char *arg;
 {
+	int n;
+
 #ifndef	_NOARCHIVE
-	if (archivefile) return(-1);
+	if (archivefile) return(FNC_QUIT);
 #endif
-	if (!yesno(QUIT_K)) return(1);
-	if (savehist > 0) savehistory(0, histfile);
-	return(-1);
+	n = 0;
+
+#ifndef	_NOORIGSHELL
+	if (!fdmode) {
+		char *str[3];
+		int val[3];
+
+		Xlocate(0, L_MESLINE);
+		Xputterm(L_CLEAR);
+		Xkanjiputs(QUIT_K);
+		val[0] = 0;
+		val[1] = 1;
+		val[2] = 2;
+		str[0] = QYES_K;
+		str[1] = QNO_K;
+		str[2] = QCHG_K;
+		if (selectstr(&n, 3, 20, str, val) != K_CR || n == 1)
+			return(FNC_CANCEL);
+	}
+	else
+#endif	/* _NOORIGSHELL */
+	if (!yesno(QUIT_K)) return(FNC_CANCEL);
+
+#ifndef	_NOPTY
+	if (confirmpty()) return(FNC_CANCEL);
+#endif
+
+	return((n) ? FNC_FAIL : FNC_QUIT);
 }
 
 static int make_dir(arg)
@@ -1086,15 +1316,17 @@ char *arg;
 	char *path;
 
 	if (arg && *arg) path = strdup2(arg);
-	else if (!(path = inputstr(MAKED_K, 1, -1, NULL, 1))) return(1);
-	else if (!*(path = evalpath(path, 1))) {
+	else if (!(path = inputstr(MAKED_K, 1, -1, NULL, HST_PATH)))
+		return(FNC_CANCEL);
+	else if (!*(path = evalpath(path, 0))) {
 		free(path);
-		return(1);
+		return(FNC_CANCEL);
 	}
 
 	if (mkdir2(path, 0777) < 0) warning(-1, path);
 	free(path);
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 static int copy_file(arg)
@@ -1142,12 +1374,12 @@ char *arg;
 		if (Xaccess(file, F_OK) >= 0 || errno != ENOENT) {
 			warning(-1, file);
 			free(file);
-			return(1);
+			return(FNC_CANCEL);
 		}
 	}
 	else for (;;) {
 		file = inputstr(NEWNM_K, 1, 0, filelist[filepos].name, -1);
-		if (!file || !*file) return(1);
+		if (!file || !*file) return(FNC_CANCEL);
 		if (Xaccess(file, F_OK) < 0) {
 			if (errno == ENOENT) break;
 			warning(-1, file);
@@ -1159,40 +1391,55 @@ char *arg;
 	if (Xrename(fnodospath(path, filepos), file) < 0) {
 		warning(-1, file);
 		free(file);
-		return(1);
+		return(FNC_CANCEL);
 	}
 	if (strdelim(file, 0)) {
 		free(file);
 		file = strdup2("..");
 	}
 	replacefname(file);
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 static int delete_file(arg)
 char *arg;
 {
+	char *cp;
+	int len;
+
 	removepolicy = 0;
 	if (mark > 0) {
-		if (!yesno(DELMK_K)) return(1);
+		if (!yesno(DELMK_K)) return(FNC_CANCEL);
 		filepos = applyfile(rmvfile, NULL);
 	}
 	else if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 	else {
-		if (!yesno(DELFL_K, filelist[filepos].name)) return(1);
+		cp = DELFL_K;
+		len = strlen2(cp) - ((int)sizeof("%.*s") - 1);
+		if (!yesno(cp, n_lastcolumn - len, filelist[filepos].name))
+			return(FNC_CANCEL);
 		filepos = applyfile(rmvfile, NULL);
 	}
 	if (filepos >= maxfile && (filepos -= 2) < 0) filepos = 0;
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 static int delete_dir(arg)
 char *arg;
 {
+	char *cp;
+	int len;
+
 	if (!isdir(&(filelist[filepos])) || isdotdir(filelist[filepos].name))
 		return(warning_bell(arg));
-	if (!yesno(DELDR_K, filelist[filepos].name)) return(1);
+	cp = DELDR_K;
+	len = strlen2(cp) - ((int)sizeof("%.*s") - 1);
+	if (!yesno(cp, n_lastcolumn - len, filelist[filepos].name))
+		return(FNC_CANCEL);
 	removepolicy = 0;
+	cp = filelist[filepos].name;
 #ifndef	NOSYMLINK
 	if (islink(&(filelist[filepos]))) {
 # ifndef	_NODOSDRIVE
@@ -1201,15 +1448,16 @@ char *arg;
 		int ret;
 
 		ret = rmvfile(fnodospath(path, filepos));
-		if (ret < 0) warning(-1, filelist[filepos].name);
+		if (ret < 0) warning(-1, cp);
 		else if (!ret) filepos++;
 	}
 	else
 #endif
-	if (!applydir(filelist[filepos].name, rmvfile, NULL, rmvdir, 3, NULL))
+	if (!applydir(cp, rmvfile, NULL, rmvdir, ORD_NOPREDIR, NULL))
 		filepos++;
 	if (filepos >= maxfile && (filepos -= 2) < 0) filepos = 0;
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 static int find_file(arg)
@@ -1218,16 +1466,17 @@ char *arg;
 	char *wild;
 
 	if (arg && *arg) wild = strdup2(arg);
-	else if (!(wild = inputstr(FINDF_K, 0, 0, "*", -1))) return(1);
+	else if (!(wild = inputstr(FINDF_K, 0, 0, "*", -1)))
+		return(FNC_CANCEL);
 
-	if (
 #ifndef	_NOARCHIVE
-	*wild != '/' &&
+	if (*wild == '/') /*EMPTY*/;
+	else
 #endif
-	strdelim(wild, 1)) {
+	if (strdelim(wild, 1)) {
 		warning(ENOENT, wild);
 		free(wild);
-		return(1);
+		return(FNC_CANCEL);
 	}
 	if (findpattern) free(findpattern);
 	if (*wild) findpattern = wild;
@@ -1235,7 +1484,8 @@ char *arg;
 		free(wild);
 		findpattern = NULL;
 	}
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 static int find_dir(arg)
@@ -1243,22 +1493,32 @@ char *arg;
 {
 	char *cp;
 
-	if (!(findregexp = prepareregexp(FINDD_K, arg))) return(1);
+	if (!(findregexp = prepareregexp(FINDD_K, arg))) return(FNC_CANCEL);
 	destpath = NULL;
 	cp = isdir(&(filelist[filepos])) ? filelist[filepos].name : ".";
-	applydir(cp, findfile, finddir, NULL, 1, NOFND_K);
+	applydir(cp, findfile, finddir, NULL, ORD_NORMAL, NOFND_K);
 	regexp_free(findregexp);
-	if (!destpath) return(1);
+	if (!destpath) return(FNC_CANCEL);
 
 	if (!(cp = strrdelim(destpath, 0))) cp = destpath;
 	else {
 		*(cp++) = '\0';
-		chdir2(destpath);
+		chdir3(destpath, 1);
 	}
 
 	replacefname(strdup2(cp));
 	free(destpath);
-	return(4);
+
+	return(FNC_EFFECT);
+}
+
+int evalstatus(n)
+int n;
+{
+	if (n < 0) return(FNC_CANCEL);
+	if (internal_status > FNC_FAIL) return(internal_status);
+
+	return(FNC_EFFECT);
 }
 
 static int execute_sh(arg)
@@ -1269,21 +1529,21 @@ char *arg;
 
 	if (arg) com = strdup2(arg);
 	else {
-		ttyiomode(1);
+		Xttyiomode(1);
 		com = inputshellloop(-1, NULL);
-		ttyiomode(0);
-		if (!com) return(1);
+		Xttyiomode(0);
+		if (!com) return(FNC_CANCEL);
 	}
 	if (*com) {
-		ret = execusercomm(com, filelist[filepos].name, 0, -1, 0);
-		ret = (ret < 0) ? 1 :
-			(internal_status >= -1) ? internal_status: 4;
+		ret = ptyusercomm(com, filelist[filepos].name, F_ARGSET);
+		ret = evalstatus(ret);
 	}
 	else {
 		execshell();
-		ret = 4;
+		ret = FNC_EFFECT;
 	}
 	free(com);
+
 	return(ret);
 }
 
@@ -1316,9 +1576,9 @@ char *arg;
 #ifdef	CWDINPATH
 	len = (!isdir(&(filelist[filepos])) && isexec(&(filelist[filepos])))
 		? strlen(filelist[filepos].name) + 1 : 0;
-	ttyiomode(1);
+	Xttyiomode(1);
 	com = inputshellloop(len, filelist[filepos].name);
-	ttyiomode(0);
+	Xttyiomode(0);
 #else
 	if (isdir(&(filelist[filepos])) || !isexec(&(filelist[filepos]))) {
 		len = 0;
@@ -1331,24 +1591,24 @@ char *arg;
 		tmp[1] = _SC_;
 		strcpy(&(tmp[2]), filelist[filepos].name);
 	}
-	ttyiomode(1);
+	Xttyiomode(1);
 	com = inputshellloop(len, tmp);
-	ttyiomode(0);
+	Xttyiomode(0);
 	if (tmp != filelist[filepos].name) free(tmp);
 #endif
-	if (!com) return(1);
+	if (!com) return(FNC_CANCEL);
 	if (!*com) {
 		execshell();
 		free(com);
-		return(4);
+		return(FNC_EFFECT);
 	}
 
 #ifndef	_NOARCHIVE
 	if (archivefile) {
 		if (!(dir = tmpunpack(1))) ret = -1;
 		else {
-			ret = execusercomm(com,
-				filelist[filepos].name, 0, -1, 1);
+			ret = ptyusercomm(com, filelist[filepos].name,
+				F_ARGSET | F_IGNORELIST);
 			removetmp(dir, NULL);
 		}
 	}
@@ -1357,16 +1617,18 @@ char *arg;
 #ifndef	_NODOSDRIVE
 	if ((drive = tmpdosdupl("", &dir, 1)) < 0) ret = -1;
 	else if (drive) {
-		ret = execusercomm(com, filelist[filepos].name, 0, -1, 1);
+		ret = ptyusercomm(com, filelist[filepos].name,
+			F_ARGSET | F_IGNORELIST);
 		removetmp(dir, filelist[filepos].name);
 	}
 	else
 #endif
 	{
-		ret = execusercomm(com, filelist[filepos].name, 0, -1, 0);
+		ret = ptyusercomm(com, filelist[filepos].name, F_ARGSET);
 	}
-	ret = (ret < 0) ? 1 : (internal_status >= -1) ? internal_status: 4;
+	ret = evalstatus(ret);
 	free(com);
+
 	return(ret);
 }
 
@@ -1375,7 +1637,8 @@ static int launch_file(arg)
 char *arg;
 {
 	if (!launcher()) return(view_file(arg));
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 static int pack_file(arg)
@@ -1385,20 +1648,22 @@ char *arg;
 	int ret;
 
 	if (arg && *arg) file = strdup2(arg);
-	else if (!(file = inputstr(PACK_K, 1, -1, NULL, 1))) return(1);
-	else if (!*(file = evalpath(file, 1))) {
+	else if (!(file = inputstr(PACK_K, 1, -1, NULL, HST_PATH)))
+		return(FNC_CANCEL);
+	else if (!*(file = evalpath(file, 0))) {
 		free(file);
-		return(1);
+		return(FNC_CANCEL);
 	}
 
 	ret = pack(file);
 	free(file);
 	if (ret < 0) {
-		putterm(t_bell);
-		return(1);
+		Xputterm(T_BELL);
+		return(FNC_CANCEL);
 	}
-	if (!ret) return(1);
-	return(4);
+	if (!ret) return(FNC_CANCEL);
+
+	return(FNC_EFFECT);
 }
 
 static int unpack_file(arg)
@@ -1409,12 +1674,15 @@ char *arg;
 #endif
 	int ret;
 
-	if (archivefile) ret = unpack(archivefile, NULL, arg, 0, 0);
+	if (archivefile) ret = unpack(archivefile, NULL, arg,
+		0, F_ARGSET | F_ISARCH | F_NOADDOPT);
 	else if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
-	else ret = unpack(fnodospath(path, filepos), NULL, arg, 0, 1);
+	else ret = unpack(fnodospath(path, filepos), NULL, arg,
+		0, F_ARGSET | F_ISARCH | F_NOADDOPT | F_IGNORELIST);
 	if (ret < 0) return(warning_bell(arg));
-	if (!ret) return(1);
-	return(4);
+	if (!ret) return(FNC_CANCEL);
+
+	return(FNC_EFFECT);
 }
 
 #ifndef	_NOTREE
@@ -1426,14 +1694,17 @@ char *arg;
 #endif
 	int ret;
 
-	if (archivefile) ret = unpack(archivefile, NULL, NULL, 1, 0);
+	if (archivefile) ret = unpack(archivefile, NULL, NULL,
+		1, F_ARGSET | F_ISARCH | F_NOADDOPT);
 	else if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
-	else ret = unpack(fnodospath(path, filepos), NULL, NULL, 1, 1);
+	else ret = unpack(fnodospath(path, filepos), NULL, NULL,
+		1, F_ARGSET | F_ISARCH | F_NOADDOPT | F_IGNORELIST);
 	if (ret <= 0) {
 		if (ret < 0) warning_bell(arg);
-		return(3);
+		return(FNC_HELPSPOT);
 	}
-	return(4);
+
+	return(FNC_EFFECT);
 }
 #endif	/* !_NOTREE */
 #endif	/* !_NOARCHIVE */
@@ -1445,59 +1716,72 @@ char *arg;
 	int ret;
 
 	if (arg && *arg) path = strdup2(arg);
-	else if (!(path = inputstr(FSDIR_K, 0, -1, NULL, 1))) return(1);
-	else if (!*(path = evalpath(path, 1))) {
+	else if (!(path = inputstr(FSDIR_K, 0, -1, NULL, HST_PATH)))
+		return(FNC_CANCEL);
+	else if (!*(path = evalpath(path, 0))) {
 		free(path);
 		path = strdup2(".");
 	}
 
 	ret = infofs(path);
 	free(path);
-	if (!ret) return(1);
-	return(3);
+	if (!ret) return(FNC_CANCEL);
+
+	return(FNC_HELPSPOT);
+}
+
+static int NEAR selectattr(s)
+char *s;
+{
+	char *str[MAXATTRSEL];
+	int n, val[MAXATTRSEL];
+
+	str[0] = CMODE_K;
+	str[1] = CDATE_K;
+	val[0] = ATR_MODEONLY;
+	val[1] = ATR_TIMEONLY;
+
+	n = ATR_MODEONLY;
+	Xlocate(0, ATTR_Y);
+	Xputterm(L_CLEAR);
+	Xkanjiputs(s);
+	if (selectstr(&n, MAXATTRSEL, ATTR_X, str, val) != K_CR) return(-1);
+
+	return(n);
 }
 
 /*ARGSUSED*/
 static int attr_file(arg)
 char *arg;
 {
-	char *str[2];
-	int i, n, flag, val[2];
-
-	str[0] = CMODE_K;
-	str[1] = CDATE_K;
-	val[0] = 1;
-	val[1] = 2;
+	int i, n, flag;
 
 	if (mark > 0) {
 		for (i = 0; i < maxfile; i++)
 			if (ismark(&(filelist[i]))) break;
 		if (i >= maxfile) i = filepos;
-		flag = 1;
-		locate(0, L_INFO);
-		putterm(l_clear);
-		kanjiputs(ATTRM_K);
-		if (selectstr(&flag, 2, 35, str, val) == K_ESC) return(1);
+		if ((flag = selectattr(ATTRM_K)) < 0) return(FNC_CANCEL);
+		flag |= ATR_MULTIPLE;
 	}
 	else {
 #ifndef	NOSYMLINK
 		if (islink(&(filelist[filepos]))) {
 			warning(0, ILLNK_K);
-			return(1);
+			return(FNC_CANCEL);
 		}
 #endif
 		i = filepos;
-		flag = 3;
+		flag = 0;
 	}
 
 	while ((n = inputattr(&(filelist[i]), flag)) < 0) warning(0, ILTMS_K);
 	if (!n) {
-		if (FILEPERROW < WFILEMINATTR) return(4);
-		return(2);
+		if (FILEPERROW < WFILEMINATTR) return(FNC_EFFECT);
+		return(FNC_UPDATE);
 	}
-
 	applyfile(setattr, NULL);
-	return(4);
+
+	return(FNC_EFFECT);
 }
 
 #ifndef	_NOTREE
@@ -1507,17 +1791,18 @@ char *arg;
 {
 	char *path;
 
-	if (!(path = tree(0, NULL))) return(2);
-	if (chdir2(path) < 0) {
+	if (!(path = tree(0, NULL))) return(FNC_UPDATE);
+	if (chdir3(path, 1) < 0) {
 		warning(-1, path);
 		free(path);
-		return(2);
+		return(FNC_UPDATE);
 	}
 	free(path);
 	if (findpattern) free(findpattern);
 	findpattern = NULL;
 	replacefname(NULL);
-	return(4);
+
+	return(FNC_EFFECT);
 }
 #endif	/* !_NOTREE */
 
@@ -1541,24 +1826,26 @@ char *arg;
 #endif
 	if (arg && *arg) dev = strdup2(arg);
 #if	MSDOS
-	else if (!(dev = inputstr(BKUP_K, 1, -1, NULL, 1))) return(1);
+	else if (!(dev = inputstr(BKUP_K, 1, -1, NULL, HST_PATH)))
+		return(FNC_CANCEL);
 #else
-	else if (!(dev = inputstr(BKUP_K, 1, 5, "/dev/", 1))) return(1);
+	else if (!(dev = inputstr(BKUP_K, 1, 5, "/dev/", HST_PATH)))
+		return(FNC_CANCEL);
 #endif
-	else if (!*(dev = evalpath(dev, 1))) {
+	else if (!*(dev = evalpath(dev, 0))) {
 		free(dev);
-		return(1);
+		return(FNC_CANCEL);
 	}
 	else if (archivefile) {
 		if (!(dir = tmpunpack(0))) {
 			free(dev);
-			return(1);
+			return(FNC_CANCEL);
 		}
 	}
 #ifndef	_NODOSDRIVE
 	else if ((drive = tmpdosdupl("", &dir, 0)) < 0) {
 		free(dev);
-		return(1);
+		return(FNC_CANCEL);
 	}
 #endif
 
@@ -1570,8 +1857,9 @@ char *arg;
 	if (archivefile) removetmp(dir, NULL);
 
 	free(dev);
-	if (ret <= 0) return(1);
-	return(4);
+	if (ret <= 0) return(FNC_CANCEL);
+
+	return(FNC_EFFECT);
 }
 #endif	/* !_NOARCHIVE */
 
@@ -1580,13 +1868,14 @@ static int search_forw(arg)
 char *arg;
 {
 	isearch = 1;
-	locate(0, L_HELP);
-	putterm(l_clear);
-	putterm(t_standout);
-	win_x = kanjiputs(SEAF_K);
+	Xlocate(0, L_HELP);
+	Xputterm(L_CLEAR);
+	Xputterm(T_STANDOUT);
+	win_x = Xkanjiputs(SEAF_K);
 	win_y = L_HELP;
-	putterm(end_standout);
-	return(0);
+	Xputterm(END_STANDOUT);
+
+	return(FNC_NONE);
 }
 
 /*ARGSUSED*/
@@ -1594,13 +1883,14 @@ static int search_back(arg)
 char *arg;
 {
 	isearch = -1;
-	locate(0, L_HELP);
-	putterm(l_clear);
-	putterm(t_standout);
-	win_x = kanjiputs(SEAB_K);
+	Xlocate(0, L_HELP);
+	Xputterm(L_CLEAR);
+	Xputterm(T_STANDOUT);
+	win_x = Xkanjiputs(SEAB_K);
 	win_y = L_HELP;
-	putterm(end_standout);
-	return(0);
+	Xputterm(END_STANDOUT);
+
+	return(FNC_NONE);
 }
 
 #ifndef	_NOSPLITWIN
@@ -1630,71 +1920,205 @@ int oldwin;
 	if (winvar[oldwin].v_fullpath) free(winvar[oldwin].v_fullpath);
 	winvar[oldwin].v_fullpath = strdup2(fullpath);
 	if (winvar[win].v_fullpath) strcpy(fullpath, winvar[win].v_fullpath);
-#ifndef	_NOARCHIVE
+# ifndef	_NOARCHIVE
 	if (winvar[oldwin].v_archivedir) free(winvar[oldwin].v_archivedir);
 	winvar[oldwin].v_archivedir = strdup2(archivedir);
 	if (winvar[win].v_archivedir)
 		strcpy(archivedir, winvar[win].v_archivedir);
-#endif
+# endif
+}
+
+int nextwin(VOID_A)
+{
+	int oldwin;
+
+	if (windows <= 1) return(-1);
+	oldwin = win;
+	if (++win >= windows) win = 0;
+	duplwin(oldwin);
+	movewin(oldwin);
+	if (chdir3(fullpath, 1) < 0) lostcwd(fullpath);
+
+	return(0);
 }
 
 /*ARGSUSED*/
 static int split_window(arg)
 char *arg;
 {
+	winvartable tmp;
 	int i, oldwin;
 
 	oldwin = win;
-	if (windows < MAXWINDOWS) {
-		win = windows++;
-		if (FILEPERROW < WFILEMIN) {
-			windows--;
-			win = oldwin;
-			warning(0, NOROW_K);
-			return(1);
-		}
-		duplwin(oldwin);
-		movewin(oldwin);
-	}
-	else {
+	if (windows >= MAXWINDOWS) {
+# ifdef	_NOEXTRAWIN
+#  ifndef	_NOPTY
+		if (confirmpty()) return(FNC_CANCEL);
+#  endif
 		if (win > 0) {
 			shutwin(0);
 			win = 0;
 			duplwin(oldwin);
 		}
 		for (i = 1; i < windows; i++) shutwin(i);
+#  ifndef	_NOPTY
+		killallpty();
+#  endif
 		windows = 1;
+		calcwin();
+# else	/* !_NOEXTRAWIN */
+		warning(0, NOWIN_K);
+		return(FNC_CANCEL);
+# endif	/* !_NOEXTRAWIN */
 	}
-	return(4);
+# ifdef	_NOEXTRAWIN
+	else if (fileperrow(windows + 1) < WFILEMIN)
+# else
+	else if (winvar[win].v_fileperrow < WFILEMIN * 2 + 1)
+# endif
+	{
+		warning(0, NOROW_K);
+		return(FNC_CANCEL);
+	}
+	else {
+		win++;
+		memcpy((char *)&tmp, (char *)&(winvar[windows]),
+			sizeof(winvartable));
+		memmove((char *)&(winvar[win + 1]), (char *)&(winvar[win]),
+			(windows++ - win) * sizeof(winvartable));
+		memcpy((char *)&(winvar[win]), (char *)&tmp,
+			sizeof(winvartable));
+
+# ifdef	_NOEXTRAWIN
+		calcwin();
+# else
+		i = (winvar[oldwin].v_fileperrow - 1) / 2;
+		winvar[win].v_fileperrow =
+			(winvar[oldwin].v_fileperrow - 1) - i;
+		winvar[oldwin].v_fileperrow = i;
+# endif
+
+# ifndef	_NOPTY
+		insertwin(win, windows);
+		changewsize(wheader, windows);
+# endif
+		duplwin(oldwin);
+		movewin(oldwin);
+	}
+
+	return(FNC_EFFECT);
 }
 
 /*ARGSUSED*/
 static int next_window(arg)
 char *arg;
 {
-	int oldwin;
+	if (nextwin() < 0) return(FNC_NONE);
+# ifndef	_NOPTY
+	if (ptylist[win].pid && ptylist[win].status < 0) {
+		rewritefile(0);
+		VOID_C frontend();
+		if (internal_status > FNC_FAIL) return(internal_status);
+	}
+# endif
 
-	if (windows <= 1) return(0);
-	oldwin = win;
-	if (++win >= windows) win = 0;
-	duplwin(oldwin);
-	movewin(oldwin);
-	if (chdir2(fullpath) < 0) lostcwd(fullpath);
-	return(4);
+	return(FNC_EFFECT);
 }
 #endif	/* !_NOSPLITWIN */
 
+#ifndef	_NOEXTRAWIN
+static int widen_window(arg)
+char *arg;
+{
+	int n, next;
+
+	if (windows <= 1) return(FNC_NONE);
+	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
+	if ((next = win + 1) >= windows) next = 0;
+	if (winvar[next].v_fileperrow - n < WFILEMIN)
+		n = winvar[next].v_fileperrow - WFILEMIN;
+	if (n <= 0) return(warning_bell(arg));
+
+	winvar[next].v_fileperrow -= n;
+	winvar[win].v_fileperrow += n;
+# ifndef	_NOPTY
+	changewsize(wheader, windows);
+# endif
+
+	return(FNC_EFFECT);
+}
+
+static int narrow_window(arg)
+char *arg;
+{
+	int n, next;
+
+	if (windows <= 1) return(FNC_NONE);
+	if (!arg || (n = atoi2(arg)) <= 0) n = 1;
+	if ((next = win + 1) >= windows) next = 0;
+	if (FILEPERROW - n < WFILEMIN) n = FILEPERROW - WFILEMIN;
+	if (n <= 0) return(warning_bell(arg));
+
+	winvar[next].v_fileperrow += n;
+	winvar[win].v_fileperrow -= n;
+# ifndef	_NOPTY
+	changewsize(wheader, windows);
+# endif
+
+	return(FNC_EFFECT);
+}
+
+/*ARGSUSED*/
+static int kill_window(arg)
+char *arg;
+{
+	winvartable tmp;
+	int prev;
+
+	if (windows <= 1) return(FNC_NONE);
+	memcpy((char *)&tmp, (char *)&(winvar[win]), sizeof(winvartable));
+	memmove((char *)&(winvar[win]), (char *)&(winvar[win + 1]),
+		(--windows - win) * sizeof(winvartable));
+	memcpy((char *)&(winvar[windows]), (char *)&tmp, sizeof(winvartable));
+
+	if ((prev = win - 1) < 0) prev = windows - 1;
+	winvar[prev].v_fileperrow += winvar[windows].v_fileperrow + 1;
+
+# ifndef	_NOPTY
+	deletewin(win, windows);
+	changewsize(wheader, windows);
+# endif
+
+	win = prev;
+	duplwin(windows);
+	movewin(windows);
+	shutwin(windows);
+	if (chdir3(fullpath, 1) < 0) lostcwd(fullpath);
+
+# ifndef	_NOPTY
+	if (ptylist[win].pid && ptylist[win].status < 0) {
+		rewritefile(0);
+		VOID_C frontend();
+		if (internal_status > FNC_FAIL) return(internal_status);
+	}
+# endif
+
+	return(FNC_EFFECT);
+}
+#endif	/* !_NOEXTRAWIN */
+
 /*ARGSUSED*/
 static int warning_bell(arg)
 char *arg;
 {
-	putterm(t_bell);
-	return(0);
+	Xputterm(T_BELL);
+
+	return(FNC_NONE);
 }
 
 /*ARGSUSED*/
 static int no_operation(arg)
 char *arg;
 {
-	return(0);
+	return(FNC_NONE);
 }
