@@ -1,7 +1,7 @@
 /*
  *	doscom.c
  *
- *	Builtin Command for DOS
+ *	builtin commands for DOS
  */
 
 #include <fcntl.h>
@@ -30,43 +30,10 @@
 #endif
 
 #if	MSDOS && !defined (FD)
-#include <dos.h>
 #define	VOL_FAT32	"FAT32"
-# ifdef	DJGPP
-# include <dpmi.h>
-# include <go32.h>
-# include <sys/farptr.h>
-#  if	(DJGPP >= 2)
-#  include <libc/dosio.h>
-#  else
-#  define	__dpmi_regs	_go32_dpmi_registers
-#  define	__tb	_go32_info_block.linear_address_of_transfer_buffer
-#  define	__tb_offset	(__tb & 15)
-#  define	__tb_segment	(__tb / 16)
-#  endif
-# define	PTR_SEG(ptr)		(__tb_segment)
-# define	PTR_OFF(ptr, ofs)	(__tb_offset + (ofs))
-# else	/* !DJGPP */
-#  ifdef	__TURBOC__	/* Oops!! Borland C++ has not x.bp !! */
-typedef union DPMI_REGS {
-	struct XREGS {
-		u_short ax, bx, cx, dx, si, di, bp, flags;
-	} x;
-	struct HREGS {
-		u_char al, ah, bl, bh, cl, ch, dl, dh;
-	} h;
-} __dpmi_regs;
-#  else
-typedef union REGS	__dpmi_regs;
-#  endif
-# define	__attribute__(x)
-# define	PTR_SEG(ptr)		FP_SEG(ptr)
-# define	PTR_OFF(ptr, ofs)	FP_OFF(ptr)
-# endif	/* !DJGPP */
 #endif	/* MSDOS && !FD */
 
 #if	MSDOS
-#include <io.h>
 # ifdef	__TURBOC__
 # include <dir.h>
 # endif
@@ -90,6 +57,8 @@ typedef union REGS	__dpmi_regs;
 #define	C_EOF		K_CTRL('D')
 #endif	/* !MSDOS */
 
+#include "system.h"
+
 #ifndef	FD
 # if	MSDOS
 typedef struct _statfs_t {
@@ -174,17 +143,13 @@ extern int errno;
 #ifdef	DEBUG
 extern char *_mtrace_file;
 #endif
-#ifdef	LSI_C
-extern u_char _openfile[];
-#endif
-
-#include "system.h"
 
 #if	defined (DOSCOMMAND) \
 && (!defined (FD) || (FD >= 2 && !defined (_NOORIGSHELL)))
 
 #ifdef	FD
 #include "term.h"
+#define	c_left		termstr[C_LEFT]
 #else	/* !FD */
 # if	MSDOS
 # define	gettext		dummy_gettext	/* fake for DJGPP gcc-3.3 */
@@ -211,25 +176,20 @@ extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
 extern int Xlstat __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
-extern int Xsymlink __P_((char *, char *));
-extern int Xreadlink __P_((char *, char *, int));
 extern int Xchmod __P_((char *, int));
 extern int Xunlink __P_((char *));
 extern int Xrename __P_((char *, char *));
 extern int Xopen __P_((char *, int, int));
 # ifdef	_NODOSDRIVE
-# define	Xclose		close
-# define	Xread		read
-# define	Xwrite		write
+# define	Xclose(f)	((close(f)) ? -1 : 0)
 # define	Xlseek		lseek
 # else	/* !_NODOSDRIVE */
 extern int Xclose __P_((int));
-extern int Xread __P_((int, char *, int));
-extern int Xwrite __P_((int, char *, int));
 extern off_t Xlseek __P_((int, off_t, int));
 # endif	/* !_NODOSDRIVE */
 extern int Xmkdir __P_((char *, int));
 extern int Xrmdir __P_((char *));
+extern int stat2 __P_((char *, struct stat *));
 extern int chdir2 __P_((char *));
 #else	/* !FD */
 # if	MSDOS
@@ -255,35 +215,40 @@ extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
 # define	Xlstat		Xstat
 # else
-# define	Xstat		stat
-# define	Xlstat		lstat
+# define	Xstat(p, s)	((stat(p, s)) ? -1 : 0)
+# define	Xlstat(p, s)	((lstat(p, s)) ? -1 : 0)
 # endif
-#define	Xaccess(p, m)	(access(p, m) ? -1 : 0)
-#define	Xsymlink(o, n)	(symlink(o, n) ? -1 : 0)
-#define	Xreadlink	readlink
-#define	Xchmod		chmod
-#define	Xunlink(p)	(unlink(p) ? -1 : 0)
-#define	Xrename(o, n)	(rename(o, n) ? -1 : 0)
+#define	Xaccess(p, m)	((access(p, m)) ? -1 : 0)
+# ifndef	NOSYMLINK
+# define	Xsymlink(o, n)	((symlink(o, n)) ? -1 : 0)
+# define	Xreadlink	readlink
+# endif
+#define	Xchmod(p, m)	((chmod(p, m)) ? -1 : 0)
+#define	Xunlink(p)	((unlink(p)) ? -1 : 0)
+#define	Xrename(f, t)	((rename(f, t)) ? -1 : 0)
 #define	Xopen		open
-#define	Xclose		close
-#define	Xread		read
-#define	Xwrite		write
+#define	Xclose(f)	((close(f)) ? -1 : 0)
 #define	Xlseek		lseek
 # if	MSDOS
 #  ifdef	DJGPP
-#  define	Xmkdir(p, m)	(mkdir(p, m) ? -1 : 0)
+#  define	Xmkdir(p, m)	((mkdir(p, m)) ? -1 : 0)
 #  else
 extern int Xmkdir __P_((char *, int));
 #  endif
 # else
-# define	Xmkdir		mkdir
+# define	Xmkdir(p, m)	((mkdir(p, m)) ? -1 : 0)
 # endif
-#define	Xrmdir(p)	(rmdir(p) ? -1 : 0)
+#define	Xrmdir(p)	((rmdir(p)) ? -1 : 0)
 # if	MSDOS
 extern int intcall __P_((int, __dpmi_regs *, struct SREGS *));
 extern int chdir2 __P_((char *));
 # else
-# define	chdir2(p)	(chdir(p) ? -1 : 0)
+# define	chdir2(p)	((chdir(p)) ? -1 : 0)
+# endif
+# ifdef	NOSYMLINK
+# define	stat2		Xstat
+# else
+extern int stat2 __P_((char *, struct stat *));
 # endif
 #endif	/* !FD */
 
@@ -346,7 +311,12 @@ extern char *strdup2 __P_((char *));
 extern int getinfofs __P_((char *, off_t *, off_t *, off_t *));
 extern char *realpath2 __P_((char *, char *, int));
 extern int touchfile __P_((char *, struct stat *));
-extern int safewrite __P_((int, char *, int));
+#ifndef	NODIRLOOP
+extern int issamebody __P_((char *, char *));
+#endif
+#ifndef	NOSYMLINK
+extern int cpsymlink __P_((char *, char *));
+#endif
 extern char *inputstr __P_((char *, int, int, char *, int));
 #else	/* !FD */
 static int NEAR getinfofs __P_((char *, off_t *, off_t *, off_t *));
@@ -372,7 +342,12 @@ static VOID NEAR ttymode __P_((int));
 static int NEAR getkey2 __P_((int));
 # endif	/* !MSDOS */
 static int NEAR touchfile __P_((char *, struct stat *));
-static int NEAR safewrite __P_((int, char *, int));
+#ifndef	NODIRLOOP
+static int NEAR issamebody __P_((char *, char *));
+#endif
+#ifndef	NOSYMLINK
+static int NEAR cpsymlink __P_((char *, char *));
+#endif
 static char *NEAR inputstr __P_((char *, int, int, char *, int));
 #endif	/* !FD */
 
@@ -577,6 +552,7 @@ int rdlink;
 #   ifdef	DJGPP
 	dosmemget(__tb + i, MAXPATHLEN, resolved);
 #   endif
+
 	return(resolved);
 }
 
@@ -626,6 +602,7 @@ struct timeval tvp[2];
 	putdostime(&(reg.x.dx), &(reg.x.cx), t);
 	i = intcall(0x21, &reg, &sreg);
 	close(fd);
+
  	return(i);
 }
 # else	/* !MSDOS */
@@ -666,7 +643,8 @@ int sig;
 	int i;
 
 	while ((i = read(ttyio, &ch, sizeof(u_char))) < 0 && errno == EINTR);
-	if (i < sizeof(u_char)) return(EOF);
+	if (i < (int)sizeof(u_char)) return(EOF);
+
 	return((int)ch);
 }
 # endif	/* !MSDOS */
@@ -680,6 +658,7 @@ struct stat *stp;
 
 	times.actime = stp -> st_atime;
 	times.modtime = stp -> st_mtime;
+
 	return(Xutime(path, &times));
 # else
 	struct timeval tvp[2];
@@ -688,28 +667,43 @@ struct stat *stp;
 	tvp[0].tv_usec = 0;
 	tvp[1].tv_sec = stp -> st_mtime;
 	tvp[1].tv_usec = 0;
+
 	return(Xutimes(path, tvp));
 # endif
 }
 
-static int NEAR safewrite(fd, buf, size)
-int fd;
-char *buf;
-int size;
+# ifndef	NODIRLOOP
+static int NEAR issamebody(src, dest)
+char *src, *dest;
 {
-	int n;
+	struct stat st1, st2;
 
-	n = Xwrite(fd, buf, size);
-# if	MSDOS
-	if (n >= 0 && n < size) {
-		n = -1;
-		errno = ENOSPC;
-	}
-# else
-	if (n < 0 && errno == EINTR) n = size;
-# endif
-	return(n);
+	if (Xstat(src, &st1) < 0 || Xstat(dest, &st2) < 0) return(0);
+
+	return(st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino);
 }
+# endif	/* !NODIRLOOP */
+
+# ifndef	NOSYMLINK
+static int NEAR cpsymlink(src, dest)
+char *src, *dest;
+{
+	struct stat st;
+	char path[MAXPATHLEN];
+	int len;
+
+	if ((len = Xreadlink(src, path, sizeof(path) - 1)) < 0) return(-1);
+	if (Xlstat(dest, &st) >= 0) {
+#  ifndef	NODIRLOOP
+		if (issamebody(src, dest)) return(0);
+#  endif
+		if (Xunlink(dest) < 0) return(-1);
+	}
+	path[len] = '\0';
+
+	return(Xsymlink(path, dest));
+}
+# endif	/* !NOSYMLINK */
 
 /*ARGSUSED*/
 static char *NEAR inputstr(prompt, delsp, ptr, def, h)
@@ -731,7 +725,8 @@ int h;
 			free(cp);
 			return(NULL);
 		}
-		if (c != K_BS) {
+		if (c > (int)MAXUTYPE(u_char)) i--;
+		else if (c != K_BS) {
 			cp = c_realloc(cp, i, &size);
 			cp[i] = c;
 			if (!iscntrl2(c)) fputc(c, stdout);
@@ -748,6 +743,7 @@ int h;
 	}
 	cp[i++] = '\0';
 	fputnl(stdout);
+
 	return(realloc2(cp, i));
 }
 #endif	/* !FD */
@@ -811,6 +807,7 @@ static int NEAR inputkey(VOID_A)
 		fputnl(stdout);
 		c = -1;
 	}
+
 	return(c);
 }
 
@@ -869,6 +866,7 @@ CONST VOID_P vp2;
 	}
 	if (cp1) *cp1 = '.';
 	if (cp2) *cp2 = '.';
+
 	return(ret);
 }
 
@@ -1025,6 +1023,7 @@ char *argv[];
 			return(-1);
 		}
 	}
+
 	return(i);
 }
 
@@ -1053,6 +1052,7 @@ int len, lower;
 	if (lower) for (i = 0; s[i]; i++) s[i] = tolower2(s[i]);
 	else for (i = 0; s[i]; i++) s[i] = toupper2(s[i]);
 	kanjifputs(s, stdout);
+
 	return(olen);
 }
 
@@ -1237,6 +1237,7 @@ int n_incline;
 		fprintf2(stdout, "\n(continuing %k)\n", dirwd);
 		dirline = n_incline;
 	}
+
 	return(0);
 }
 
@@ -1363,7 +1364,7 @@ off_t *sump, *bsump;
 	dirlist = NULL;
 	while ((dp = Xreaddir(dirp))) {
 		strcpy(file, dp -> d_name);
-		if (Xstat(dirwd, &st) < 0) continue;
+		if (stat2(dirwd, &st) < 0) continue;
 
 		dirlist = b_realloc(dirlist, n, struct filestat_t);
 		dirlist[n].nam = strdup2(dp -> d_name);
@@ -1428,10 +1429,12 @@ off_t *sump, *bsump;
 					break;
 				case 'S':
 					if ((dirlist[n].mod & S_IFMT)
-					!= S_IFSOCK) f ^= 1;
+					!= S_IFSOCK)
+						f ^= 1;
 					break;
 				case 'A':
-					if (!(dirlist[n].mod & S_ISVTX)) f ^= 1;
+					if (!(dirlist[n].mod & S_ISVTX))
+						f ^= 1;
 					break;
 				default:
 					break;
@@ -1508,6 +1511,7 @@ off_t *sump, *bsump;
 #endif
 	}
 	if (dirlist) free(dirlist);
+
 	return(max);
 }
 
@@ -1609,6 +1613,7 @@ char *argv[];
 #endif
 
 	fflush(stdout);
+
 	return(RET_SUCCESS);
 }
 
@@ -1624,6 +1629,7 @@ char *argv[];
 		doserror(argv[2], ER_TOOMANYPARAM);
 		return(-1);
 	}
+
 	return(0);
 }
 
@@ -1637,6 +1643,7 @@ char *argv[];
 		dosperror(argv[1]);
 		return(RET_FAIL);
 	}
+
 	return(RET_SUCCESS);
 }
 
@@ -1650,6 +1657,7 @@ char *argv[];
 		dosperror(argv[1]);
 		return(RET_FAIL);
 	}
+
 	return(RET_SUCCESS);
 }
 
@@ -1701,7 +1709,7 @@ char *argv[];
 		} while (!strchr("ynYN", key));
 		if (key == 'n' || key == 'N') return(RET_SUCCESS);
 	}
-	if (!(wild = evalwild(argv[n]))) {
+	if (!(wild = evalwild(argv[n], 0))) {
 		doserror(argv[n], ER_FILENOTFOUND);
 		return(RET_FAIL);
 	}
@@ -1716,18 +1724,21 @@ char *argv[];
 					freevar(wild);
 					return(ret);
 				}
-				fputc(key, stdout);
+				if (key <= (int)MAXUTYPE(u_char)
+				&& isprint2(key))
+					fputc(key, stdout);
 				fputnl(stdout);
 			} while (!strchr("ynYN", key));
 			if (key == 'n' || key == 'N') continue;
 		}
-		if (Xunlink(wild[i]) != 0) {
+		if (Xunlink(wild[i]) < 0) {
 			dosperror(wild[i]);
 			ret = RET_FAIL;
 			ERRBREAK;
 		}
 	}
 	freevar(wild);
+
 	return(ret);
 }
 
@@ -1753,6 +1764,7 @@ char *dest, *src, *wild, *swild;
 		if (src[j]) j++;
 	}
 	dest[n] = '\0';
+
 	return(dest);
 }
 
@@ -1779,7 +1791,7 @@ char *argv[];
 		return(RET_FAIL);
 	}
 
-	if (!(wild = evalwild(argv[1]))) {
+	if (!(wild = evalwild(argv[1], 0))) {
 		doserror(argv[1], ER_FILENOTFOUND);
 		return(RET_FAIL);
 	}
@@ -1801,6 +1813,7 @@ char *argv[];
 		}
 	}
 	freevar(wild);
+
 	return(ret);
 }
 
@@ -1848,6 +1861,7 @@ char *argv[];
 				break;
 		}
 	}
+
 	return(i);
 }
 
@@ -1867,6 +1881,7 @@ int bin;
 		if (toupper2(*cp) == 'B') bin = CF_BINARY;
 		else if (toupper2(*cp) == 'A') bin = CF_TEXT;
 	}
+
 	return(bin);
 }
 
@@ -1915,6 +1930,14 @@ char *file, *src;
 
 	flags = (copyflag & CF_VERIFY) ? O_RDWR : O_WRONLY;
 	flags |= (O_BINARY | O_CREAT | O_TRUNC);
+
+#ifndef	NODIRLOOP
+	if (issamebody(src, file)) {
+		flags |= O_EXCL;
+		if (Xunlink(file) < 0) return(-1);
+	}
+#endif
+
 	return(Xopen(file, flags, 0666));
 }
 
@@ -1927,26 +1950,23 @@ int size, bin;
 	int i, n;
 
 	if (!(bin & (CF_BINARY | CF_TEXT))) bin = copyflag;
-	if (!(bin & CF_TEXT))
-		while ((n = Xread(fd, (char *)buf, size)) < 0
-		&& errno == EINTR);
+	if (!(bin & CF_TEXT)) n = sureread(fd, buf, size);
 	else for (n = 0; n < size; n++) {
-		while ((i = Xread(fd, (char *)&ch, sizeof(u_char))) < 0
-		&& errno == EINTR);
+		i = sureread(fd, &ch, sizeof(u_char));
 		if (i < 0) return(-1);
 #if	MSDOS
-		if (i >= sizeof(u_char) && ch != C_EOF) buf[n] = ch;
+		if (i >= (int)sizeof(u_char) && ch != C_EOF) buf[n] = ch;
 #else
-		if (i < sizeof(u_char)) break;
+		if (i < (int)sizeof(u_char)) break;
 		else if (ch != C_EOF) buf[n] = ch;
 #endif
 		else for (;;) {
-			while ((i = Xread(fd, (char *)&ch, sizeof(u_char))) < 0
-			&& errno == EINTR);
+			i = sureread(fd, &ch, sizeof(u_char));
 			if (i < 0) return(-1);
-			if (i < sizeof(u_char) || ch == '\n') return(n);
+			if (i < (int)sizeof(u_char) || ch == '\n') return(n);
 		}
 	}
+
 	return(n);
 }
 
@@ -1960,9 +1980,10 @@ int fd, bin;
 	if (!(bin & (CF_BINARY | CF_TEXT))) bin = copyflag;
 	if (bin & CF_TEXT) {
 		ch = C_EOF;
-		if (safewrite(fd, (char *)&ch, sizeof(u_char)) < 0) ret = -1;
+		if (surewrite(fd, &ch, sizeof(u_char)) < 0) ret = -1;
 	}
-	Xclose(fd);
+	VOID_C Xclose(fd);
+
 	return(ret);
 }
 
@@ -1976,12 +1997,8 @@ int sbin, dbin, dfd;
 	int i, n, fd1, fd2, tty, retry, duperrno;
 
 #ifndef	NOSYMLINK
-	if (dfd < 0 && (stp -> st_mode & S_IFMT) == S_IFLNK) {
-		if ((i = Xreadlink(src, (char *)buf, BUFSIZ - 1)) < 0)
-			return(-1);
-		buf[i] = '\0';
-		return(Xsymlink((char *)buf, dest) >= 0);
-	}
+	if (dfd < 0 && (stp -> st_mode & S_IFMT) == S_IFLNK)
+		return(cpsymlink(src, dest));
 #endif
 
 	if ((fd1 = Xopen(src, O_TEXT | O_RDONLY, stp -> st_mode)) < 0)
@@ -1991,7 +2008,7 @@ int sbin, dbin, dfd;
 		ofs = Xlseek(fd2, (off_t)0, L_INCR);
 	}
 	else if ((fd2 = writeopen(dest, src)) <= 0) {
-		Xclose(fd1);
+		VOID_C Xclose(fd1);
 		return(fd2);
 	}
 	else ofs = (off_t)0;
@@ -2018,7 +2035,8 @@ int sbin, dbin, dfd;
 				duperrno = errno;
 				break;
 			}
-			if ((i = safewrite(fd2, (char *)buf, i)) < 0) {
+			if (surewrite(fd2, buf, i) < 0) {
+				i = -1;
 				duperrno = errno;
 				break;
 			}
@@ -2054,7 +2072,7 @@ int sbin, dbin, dfd;
 	}
 
 	if (dfd < 0) textclose(fd2, dbin);
-	Xclose(fd1);
+	VOID_C Xclose(fd1);
 
 	if (i < 0) {
 		VOID_C Xunlink(dest);
@@ -2062,10 +2080,11 @@ int sbin, dbin, dfd;
 		return(dfd < 0 ? -1 : -2);
 	}
 
-#if	!MSDOS && defined (UF_SETTABLE) && defined (SF_SETTABLE)
-	stp -> st_flags = 0xffffffff;
+#ifdef	FD
+	stp -> st_nlink = (TCH_ATIME | TCH_MTIME);
 #endif
 	if (!tty && touchfile(dest, stp) < 0) return(-1);
+
 	return(1);
 }
 
@@ -2190,7 +2209,7 @@ char *argv[];
 		copyflag |= CF_VERBOSE;
 	}
 	for (n = 0; n < nf; n++) {
-		if (!(wild = evalwild(arg[n]))) {
+		if (!(wild = evalwild(arg[n], 0))) {
 			if (nf > 1) continue;
 			doserror(arg[n], ER_FILENOTFOUND);
 			ret = RET_FAIL;
@@ -2231,6 +2250,7 @@ char *argv[];
 		fprintf2(stdout, "%9d file(s) copied", nc);
 		fputnl(stdout);
 	}
+
 	return(ret);
 }
 
@@ -2240,13 +2260,14 @@ int argc;
 char *argv[];
 {
 #ifdef	FD
-	putterms(t_clear);
+	putterms(T_CLEAR);
 	tflush();
 #else
 	fputs(t_clear, stdout);
 	fflush(stdout);
 #endif
 	fputnl(stderr);
+
 	return(RET_SUCCESS);
 }
 
@@ -2283,13 +2304,13 @@ char *argv[];
 	i = 0;
 	for (;;) {
 		if ((n = textread(fd, &ch, sizeof(ch), CF_TEXT)) < 0) break;
-		if (n < sizeof(ch)) {
-			n = safewrite(STDOUT_FILENO, cp, i);
+		if (n < (int)sizeof(ch)) {
+			n = surewrite(STDOUT_FILENO, cp, i);
 			break;
 		}
 		else if (ch == '\n') {
 			cp[i++] = '\n';
-			if ((n = safewrite(STDOUT_FILENO, cp, i)) < 0) break;
+			if ((n = surewrite(STDOUT_FILENO, cp, i)) < 0) break;
 			i = 0;
 		}
 		else {
@@ -2298,7 +2319,7 @@ char *argv[];
 		}
 	}
 	free(cp);
-	Xclose(fd);
+	VOID_C Xclose(fd);
 #if	MSDOS && !defined (LSI_C)
 	if (omode >= 0) setmode(STDOUT_FILENO, omode);
 #endif
@@ -2307,6 +2328,7 @@ char *argv[];
 		dosperror(argv[1]);
 		return(RET_FAIL);
 	}
+
 	return(RET_SUCCESS);
 }
 #endif	/* DOSCOMMAND && (!FD || (FD >= 2 && !_NOORIGSHELL)) */
