@@ -1,13 +1,13 @@
 /*
  *	file.c
  *
- *	File Access Module
+ *	file accessing module
  */
 
 #include <fcntl.h>
 #include "fd.h"
+#include "termio.h"
 #include "func.h"
-#include "kctype.h"
 #include "kanji.h"
 
 #ifndef	ENOSPC
@@ -26,18 +26,13 @@ extern char *preparefile __P_((char *, char *));
 #endif
 
 #ifndef	_NODOSDRIVE
-extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
-extern int shutdrv __P_((int));
 extern int lastdrv;
-# if	MSDOS
+# if	MSDOS && !defined (_NOUSELFN)
 extern int checkdrive __P_((int));
 # endif
 #endif
 
 extern int mark;
-extern off_t marksize;
-extern off_t totalsize;
-extern off_t blocksize;
 extern char fullpath[];
 extern char *tmpfilename;
 extern int physical_path;
@@ -81,6 +76,7 @@ static int NEAR cpfile __P_((char *, char *, struct stat *, struct stat *));
 static VOID changemes __P_((VOID_A));
 static int NEAR genrand __P_((int));
 static int dounlink __P_((char *));
+static int dormdir __P_((char *));
 #ifndef	_NOWRITEFS
 static int NEAR isexist __P_((char *));
 static int NEAR realdirsiz __P_((char *, int, int, int, int));
@@ -113,6 +109,7 @@ char *path, *file;
 	path[0] = '.';
 	path[1] = _SC_;
 	strcpy(&(path[2]), file);
+
 	return(path);
 }
 #endif	/* _USEDOSEMU */
@@ -138,11 +135,12 @@ gid_t gid;
 	else if (isgroupmember(gid)) mode >>= 3;
 #endif	/* !NOUID */
 	if (dir && !(mode & F_ISEXE)) mode &= ~(F_ISRED | F_ISWRI);
+
 	return(mode & 007);
 }
 
-int getstatus(list)
-namelist *list;
+int getstatus(namep)
+namelist *namep;
 {
 #ifdef	_USEDOSEMU
 	char path[MAXPATHLEN];
@@ -150,46 +148,43 @@ namelist *list;
 	struct stat st, lst;
 	char *cp;
 
-	cp = nodospath(path, list -> name);
+	cp = nodospath(path, namep -> name);
 	if (Xlstat(cp, &lst) < 0 || stat2(cp, &st) < 0) return(-1);
-	list -> flags = 0;
-	if ((st.st_mode & S_IFMT) == S_IFDIR) list -> flags |= F_ISDIR;
-	if ((lst.st_mode & S_IFMT) == S_IFLNK) list -> flags |= F_ISLNK;
+	namep -> flags = 0;
+	if (s_isdir(&st)) namep -> flags |= F_ISDIR;
+	if (s_islnk(&lst)) namep -> flags |= F_ISLNK;
 
 	if (isdisplnk(dispmode))
 		memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
 #if	!MSDOS
 	if ((lst.st_mode & S_IFMT) == S_IFCHR
-	|| (lst.st_mode & S_IFMT) == S_IFBLK) list -> flags |= F_ISDEV;
+	|| (lst.st_mode & S_IFMT) == S_IFBLK)
+		namep -> flags |= F_ISDEV;
 #endif
 
-	list -> st_mode = lst.st_mode;
-	list -> st_nlink = lst.st_nlink;
+	namep -> st_mode = lst.st_mode;
+	namep -> st_nlink = lst.st_nlink;
 #ifndef	NOUID
-	list -> st_uid = lst.st_uid;
-	list -> st_gid = lst.st_gid;
+	namep -> st_uid = lst.st_uid;
+	namep -> st_gid = lst.st_gid;
 #endif
 #if	MSDOS
-	list -> st_size = lst.st_size;
+	namep -> st_size = lst.st_size;
 #else
-	list -> st_size = isdev(list) ? (off_t)(lst.st_rdev) : lst.st_size;
+	namep -> st_size = isdev(namep) ? (off_t)(lst.st_rdev) : lst.st_size;
 #endif
 #ifdef	HAVEFLAGS
-	list -> st_flags = lst.st_flags;
+	namep -> st_flags = lst.st_flags;
 #endif
-	list -> st_mtim = lst.st_mtime;
-	list -> flags |=
+	namep -> st_mtim = lst.st_mtime;
+	namep -> flags |=
 #ifdef	NOUID
 		logical_access((u_int)(st.st_mode));
 #else
 		logical_access((u_int)(st.st_mode), st.st_uid, st.st_gid);
 #endif
+	namep -> tmpflags |= F_STAT;
 
-	if (isfile(list) && list -> st_size) {
-		totalsize += getblock(list -> st_size);
-		if (ismark(list)) marksize += getblock(list -> st_size);
-	}
-	list -> tmpflags |= F_STAT;
 	return(0);
 }
 
@@ -197,50 +192,45 @@ int cmplist(vp1, vp2)
 CONST VOID_P vp1;
 CONST VOID_P vp2;
 {
-	namelist *list1, *list2;
+	namelist *namep1, *namep2;
 	char *cp1, *cp2;
 	int tmp;
 
-	list1 = (namelist *)vp1;
-	list2 = (namelist *)vp2;
-	if (!sorton) return(list1 -> ent - list2 -> ent);
+	namep1 = (namelist *)vp1;
+	namep2 = (namelist *)vp2;
+	if (!sorton) return(namep1 -> ent - namep2 -> ent);
 
 #ifndef	_NOPRECEDE
-	if (!havestat(list1) && getstatus(list1) < 0) return(1);
-	if (!havestat(list2) && getstatus(list2) < 0) return(-1);
+	if (!havestat(namep1) && getstatus(namep1) < 0) return(1);
+	if (!havestat(namep2) && getstatus(namep2) < 0) return(-1);
 #endif
 
-	if (!isdir(list1) && isdir(list2)) return(1);
-	if (isdir(list1) && !isdir(list2)) return(-1);
-
-	if (list1 -> name[0] == '.' && list1 -> name[1] == '\0') return(-1);
-	if (list2 -> name[0] == '.' && list2 -> name[1] == '\0') return(1);
-	if (list1 -> name[0] == '.' && list1 -> name[1] == '.'
-	&& list1 -> name[2] == '\0') return(-1);
-	if (list2 -> name[0] == '.' && list2 -> name[1] == '.'
-	&& list2 -> name[2] == '\0') return(1);
+	if (!isdir(namep1) && isdir(namep2)) return(1);
+	if (isdir(namep1) && !isdir(namep2)) return(-1);
+	if ((tmp = isdotdir(namep2 -> name) - isdotdir(namep1 -> name)))
+		return(tmp);
 
 	switch (sorton & 7) {
 		case 5:
-			tmp = (int)strlen(list1 -> name)
-				- (int)strlen(list2 -> name);
+			tmp = (int)strlen(namep1 -> name)
+				- (int)strlen(namep2 -> name);
 			if (tmp) break;
 /*FALLTHRU*/
 		case 1:
-			tmp = strpathcmp2(list1 -> name, list2 -> name);
+			tmp = strpathcmp2(namep1 -> name, namep2 -> name);
 			break;
 		case 2:
-			if (isdir(list1)) {
-				tmp = strpathcmp2(list1 -> name,
-					list2 -> name);
+			if (isdir(namep1)) {
+				tmp = strpathcmp2(namep1 -> name,
+					namep2 -> name);
 				break;
 			}
-			cp1 = list1 -> name + strlen(list1 -> name);
-			cp2 = list2 -> name + strlen(list2 -> name);
+			cp1 = namep1 -> name + strlen(namep1 -> name);
+			cp2 = namep2 -> name + strlen(namep2 -> name);
 			for (;;) {
-				while (cp1 > list1 -> name)
+				while (cp1 > namep1 -> name)
 					if (*(--cp1) == '.') break;
-				while (cp2 > list2 -> name)
+				while (cp2 > namep2 -> name)
 					if (*(--cp2) == '.') break;
 				if (*cp2 != '.') {
 					tmp = (*cp1 == '.');
@@ -255,18 +245,18 @@ CONST VOID_P vp2;
 			}
 			break;
 		case 3:
-			if (isdir(list1))
-				tmp = strpathcmp2(list1 -> name,
-					list2 -> name);
-			else if (list1 -> st_size == list2 -> st_size)
+			if (isdir(namep1))
+				tmp = strpathcmp2(namep1 -> name,
+					namep2 -> name);
+			else if (namep1 -> st_size == namep2 -> st_size)
 				tmp = 0;
-			else tmp = (list1 -> st_size > list2 -> st_size)
+			else tmp = (namep1 -> st_size > namep2 -> st_size)
 				? 1 : -1;
 			break;
 		case 4:
-			if (list1 -> st_mtim == list2 -> st_mtim)
+			if (namep1 -> st_mtim == namep2 -> st_mtim)
 				tmp = 0;
-			else tmp = (list1 -> st_mtim > list2 -> st_mtim)
+			else tmp = (namep1 -> st_mtim > namep2 -> st_mtim)
 				? 1 : -1;
 			break;
 		default:
@@ -275,7 +265,7 @@ CONST VOID_P vp2;
 	}
 
 	if (sorton > 7) tmp = -tmp;
-	if (!tmp) tmp = list1 -> ent - list2 -> ent;
+	if (!tmp) tmp = namep1 -> ent - namep2 -> ent;
 
 	return(tmp);
 }
@@ -310,8 +300,8 @@ CONST VOID_P vp2;
 			tmp = 0;
 			break;
 	}
-
 	if (sorton > 7) tmp = -tmp;
+
 	return(tmp);
 }
 #endif
@@ -334,8 +324,7 @@ char *arcregstr;
 	}
 #ifndef	_NOARCHIVE
 	else if (arcregstr) while ((dp = Xreaddir(dirp))) {
-		if (stat2(dp -> d_name, &st) < 0
-		|| (st.st_mode & S_IFMT) == S_IFDIR) continue;
+		if (stat2(dp -> d_name, &st) < 0 || s_isdir(&st)) continue;
 
 		tmp.name = dp -> d_name;
 		tmp.flags = 0;
@@ -346,6 +335,7 @@ char *arcregstr;
 	}
 #endif
 	else dp = Xreaddir(dirp);
+
 	return(dp);
 }
 
@@ -365,11 +355,10 @@ char *buf;
 	}
 	if (!homedir) {
 		if (!(cp = gethomedir())) return(-1);
-		if (
 #ifndef	_NODOSDRIVE
-		dospath2(cp) ||
+		if (dospath2(cp)) return(-1);
 #endif
-		_chdir2(cp) < 0) return(-1);
+		if (_chdir2(cp) < 0) return(-1);
 		homedir = getwd2();
 		if (_chdir2(cwd) < 0) lostcwd(cwd);
 	}
@@ -377,22 +366,19 @@ char *buf;
 	if (buf && !physical_path) strcpy(cwd, fullpath);
 	len = strlen(homedir);
 #if	MSDOS
-	if (len <= 3 || strnpathcmp(cwd, homedir, len)) {
+	if (len <= 3) cp = NULL;
 #else
-	if (len <= 1 || strnpathcmp(cwd, homedir, len)) {
+	if (len <= 1) cp = NULL;
 #endif
-		cp = NULL;
-		if (buf) *buf = '\0';
-	}
-	else {
-		cp = cwd + len;
-		if (buf) strcpy(buf, cp);
-	}
+	else cp = underpath(cwd, homedir, len);
+
+	if (buf) strcpy(buf, (cp) ? cp : "");
 #ifdef	DEBUG
 	free(homedir);
 	homedir = NULL;
 #endif
-	return(cp ? 1 : 0);
+
+	return((cp) ? 1 : 0);
 }
 
 int preparedir(dir)
@@ -423,25 +409,65 @@ char *dir;
 			}
 			else if (stat2(dir, &st) < 0) return(-1);
 		}
-		if ((st.st_mode & S_IFMT) != S_IFDIR) {
+		if (!s_isdir(&st)) {
 			errno = ENOTDIR;
 			return(-1);
 		}
 	}
-
 	entryhist(1, realpath2(dir, tmp, 0), 1);
+
 	return(0);
 }
 
+#ifndef	NOFLOCK
+int lockfile(fd, mode)
+int fd, mode;
+{
+	static int lockmode[] = {
+# ifdef	USEFCNTLOCK
+		F_RDLCK, F_WRLCK, F_UNLCK,
+# else	/* !USEFCNTLOCK */
+#  ifdef	USELOCKF
+		F_LOCK, F_LOCK, F_ULOCK,
+#  else
+		LOCK_SH, LOCK_EX, LOCK_UN,
+#  endif
+# endif	/* !USEFCNTLOCK */
+	};
+# ifdef	USEFCNTLOCK
+	struct flock lock;
+# endif
+
+# ifndef	_NODOSDRIVE
+	if (fd >= DOSFDOFFSET) return(0);
+# endif
+
+# ifdef	USEFCNTLOCK
+	lock.l_type = lockmode[mode];
+	lock.l_start = lock.l_len = (off_t)0;
+	lock.l_whence = SEEK_SET;
+	return(fcntl(fd, F_SETLKW, (int)&lock));
+# else	/* !USEFCNTLOCK */
+#  ifdef	USELOCKF
+	return(lockf(fd, lockmode[mode], (off_t)0));
+#  else
+	return(flock(fd, lockmode[mode]));
+#  endif
+# endif	/* !USEFCNTLOCK */
+}
+#endif	/* !NOFLOCK */
+
 int touchfile(path, stp)
 char *path;
 struct stat *stp;
 {
-	struct stat st, dummy;
+	struct stat st;
+	u_int mode;
 	int ret, duperrno;
 
+	if (!(stp -> st_nlink)) return(0);
 	if (Xlstat(path, &st) < 0) return(-1);
-	if ((st.st_mode & S_IFMT) == S_IFLNK) return(1);
+	if (s_islnk(&st)) return(1);
 
 	ret = 0;
 #ifdef	FAKEUNINIT
@@ -450,10 +476,19 @@ struct stat *stp;
 #if	MSDOS
 	if (!(st.st_mode & S_IWRITE)) Xchmod(path, (st.st_mode | S_IWRITE));
 #endif
-	if (stp -> st_atime != (time_t)-1 && stp -> st_mtime != (time_t)-1) {
+
+	if (stp -> st_nlink & (TCH_ATIME | TCH_MTIME)) {
 #ifdef	USEUTIME
 		struct utimbuf times;
+#else
+		struct timeval tvp[2];
+#endif
 
+		if (!(stp -> st_nlink & TCH_ATIME))
+			stp -> st_atime = st.st_atime;
+		if (!(stp -> st_nlink & TCH_MTIME))
+			stp -> st_mtime = st.st_mtime;
+#ifdef	USEUTIME
 		times.actime = stp -> st_atime;
 		times.modtime = stp -> st_mtime;
 		if (Xutime(path, &times) < 0) {
@@ -461,8 +496,6 @@ struct stat *stp;
 			ret = -1;
 		}
 #else
-		struct timeval tvp[2];
-
 		tvp[0].tv_sec = stp -> st_atime;
 		tvp[0].tv_usec = 0;
 		tvp[1].tv_sec = stp -> st_mtime;
@@ -473,26 +506,21 @@ struct stat *stp;
 		}
 #endif
 	}
+
 #ifdef	HAVEFLAGS
-	if (stp -> st_flags != 0xffffffff) {
-# ifndef	_NODOSDRIVE
-		if (dospath(path, NULL)) {
-			duperrno = EACCES;
-			ret = -1;
-		}
-		else
-# endif
-		if (chflags(path, stp -> st_flags) < 0) {
+	if (stp -> st_nlink & TCH_FLAGS) {
+		if (Xchflags(path, stp -> st_flags) < 0) {
 			duperrno = errno;
 			ret = -1;
 		}
 	}
 #endif
-	dummy.st_mode = 0xffff;
-	if (stp -> st_mode != dummy.st_mode) {
-		stp -> st_mode &= ~S_IFMT;
-		stp -> st_mode |= (st.st_mode & S_IFMT);
-		if (Xchmod(path, stp -> st_mode) < 0) {
+
+	if (stp -> st_nlink & TCH_MODE) {
+		mode = stp -> st_mode;
+		mode &= ~S_IFMT;
+		mode |= (st.st_mode & S_IFMT);
+		if (Xchmod(path, mode) < 0) {
 			duperrno = errno;
 			ret = -1;
 		}
@@ -500,17 +528,22 @@ struct stat *stp;
 #if	MSDOS
 	else if (!(st.st_mode & S_IWRITE)) Xchmod(path, st.st_mode);
 #endif
+
 #ifndef	NOUID
-	if (stp -> st_gid != (gid_t)-1) {
-# ifndef	_NODOSDRIVE
-		if (dospath(path, NULL)) /*EMPTY*/;
-		else
-# endif
-		if (chown(path, stp -> st_uid, stp -> st_gid) < 0)
-			chown(path, (uid_t)-1, stp -> st_gid);
+	if (stp -> st_nlink & (TCH_UID | TCH_GID)) {
+		if (!(stp -> st_nlink & TCH_UID)) stp -> st_uid = (uid_t)-1;
+		if (!(stp -> st_nlink & TCH_GID)) stp -> st_gid = (gid_t)-1;
+		if (Xchown(path, stp -> st_uid, stp -> st_gid) >= 0) /*EMPTY*/;
+		else if (!(stp -> st_nlink & TCH_CHANGE))
+			Xchown(path, (uid_t)-1, stp -> st_gid);
+		else {
+			duperrno = errno;
+			ret = -1;
+		}
 	}
 #endif	/* !NOUID */
 	if (ret < 0) errno = duperrno;
+
 	return(ret);
 }
 
@@ -524,6 +557,7 @@ char *path;
 	dosdrive = 0;
 	n = Xchdir(path);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -537,6 +571,7 @@ char *path;
 	dosdrive = 0;
 	cp = Xgetwd(path);
 	dosdrive = dupdosdrive;
+
 	return(cp);
 }
 
@@ -550,6 +585,7 @@ struct stat *stp;
 	dosdrive = 0;
 	n = Xlstat(path, stp);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -563,6 +599,7 @@ int mode;
 	dosdrive = 0;
 	n = Xmkdir(path, mode);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -575,6 +612,7 @@ char *path;
 	dosdrive = 0;
 	n = Xrmdir(path);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 #endif	/* !_NODOSDRIVE */
@@ -597,27 +635,45 @@ char *path;
 
 	if (path != fullpath) strncpy2(fullpath, path, MAXPATHLEN - 1);
 	warning(0, cp);
+#ifndef	_NOUSEHASH
+	searchhash(NULL, "", "");
+#endif
 	errno = duperrno;
 }
 
-int safewrite(fd, buf, size)
-int fd;
-char *buf;
-int size;
+#ifndef	NODIRLOOP
+int issamebody(src, dest)
+char *src, *dest;
 {
-	int n;
+	struct stat st1, st2;
 
-	n = Xwrite(fd, buf, size);
-#if	MSDOS
-	if (n >= 0 && n < size) {
-		n = -1;
-		errno = ENOSPC;
-	}
-#else
-	if (n < 0 && errno == EINTR) n = size;
-#endif
-	return(n);
+	if (Xstat(src, &st1) < 0 || Xstat(dest, &st2) < 0) return(0);
+
+	return(st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino);
 }
+#endif	/* !NODIRLOOP */
+
+#ifndef	NOSYMLINK
+int cpsymlink(src, dest)
+char *src, *dest;
+{
+	struct stat st;
+	char path[MAXPATHLEN];
+	int len;
+
+	if ((len = Xreadlink(src, path, sizeof(path) - 1)) < 0) return(-1);
+	if (Xlstat(dest, &st) >= 0) {
+# ifndef	NODIRLOOP
+		if (issamebody(src, dest)) return(0);
+# endif
+		if (Xunlink(dest) < 0) return(-1);
+	}
+
+	path[len] = '\0';
+
+	return(Xsymlink(path, dest));
+}
+#endif	/* !NOSYMLINK */
 
 /*ARGSUSED*/
 static int NEAR cpfile(src, dest, stp1, stp2)
@@ -628,55 +684,64 @@ struct stat *stp1, *stp2;
 	struct stat st;
 #endif
 	char buf[BUFSIZ];
-	int i, fd1, fd2, duperrno;
+	int i, fd1, fd2, flags, mode, duperrno;
 
 #if	MSDOS
 	if (!stp2 && Xlstat(dest, &st) >= 0) stp2 = &st;
 	if (stp2 && !(stp2 -> st_mode & S_IWRITE))
 		Xchmod(src, stp2 -> st_mode | S_IWRITE);
 #endif
-	if ((stp1 -> st_mode & S_IFMT) == S_IFLNK) {
-		if ((i = Xreadlink(src, buf, BUFSIZ - 1)) < 0) return(-1);
-		buf[i] = '\0';
-		return(Xsymlink(buf, dest));
-	}
-	if ((fd1 = Xopen(src, O_BINARY | O_RDONLY, stp1 -> st_mode)) < 0)
-		return(-1);
-#if	MSDOS
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC,
-		stp1 -> st_mode | S_IWRITE);
-#else
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC, stp1 -> st_mode);
+#ifndef	NOSYMLINK
+	if (s_islnk(stp1)) return(cpsymlink(src, dest));
 #endif
-	if (fd2 < 0) {
-		Xclose(fd1);
+
+	flags = (O_BINARY | O_RDONLY);
+	mode = stp1 -> st_mode;
+	if ((fd1 = Xopen(src, flags, mode)) < 0) return(-1);
+
+	flags = (O_BINARY | O_WRONLY | O_CREAT | O_TRUNC);
+#if	MSDOS
+	mode |= S_IWRITE;
+#endif
+#ifndef	NODIRLOOP
+	if (issamebody(src, dest)) {
+		flags |= O_EXCL;
+		if (Xunlink(dest) < 0) {
+			VOID_C Xclose(fd1);
+			return(-1);
+		}
+	}
+#endif
+	if ((fd2 = Xopen(dest, flags, mode)) < 0) {
+		VOID_C Xclose(fd1);
 		return(-1);
 	}
 
+#ifdef	FAKEUNINIT
+	duperrno = errno;	/* fake for -Wuninitialized */
+#endif
 	for (;;) {
-		while ((i = Xread(fd1, buf, BUFSIZ)) < 0 && errno == EINTR);
-		duperrno = errno;
-		if (i < BUFSIZ) break;
-		if ((i = safewrite(fd2, buf, BUFSIZ)) < 0) {
+		if ((i = sureread(fd1, buf, BUFSIZ)) <= 0) {
 			duperrno = errno;
 			break;
 		}
+		if (surewrite(fd2, buf, i) < 0) {
+			i = -1;
+			duperrno = errno;
+			break;
+		}
+		if (i < BUFSIZ) break;
 	}
-	if (i > 0 && (i = safewrite(fd2, buf, i)) < 0) duperrno = errno;
 
-	Xclose(fd2);
-	Xclose(fd1);
+	VOID_C Xclose(fd2);
+	VOID_C Xclose(fd1);
 
 	if (i < 0) {
 		Xunlink(dest);
 		errno = duperrno;
 		return(-1);
 	}
-#ifdef	HAVEFLAGS
-	stp1 -> st_flags = 0xffffffff;
-#endif
+	stp1 -> st_nlink = (TCH_ATIME | TCH_MTIME);
 #ifdef	_USEDOSCOPY
 	if (touchfile(dest, stp1) < 0) return(-1);
 #else
@@ -684,6 +749,7 @@ struct stat *stp1, *stp2;
 	if (inheritcopy && touchfile(dest, stp1) < 0) return(-1);
 # endif
 #endif
+
 	return(0);
 }
 
@@ -727,13 +793,12 @@ char *src, *dest;
 struct stat *stp1, *stp2;
 {
 	if (Xrename(src, dest) >= 0) return(0);
-	if (errno != EXDEV || (stp1 -> st_mode & S_IFMT) == S_IFDIR)
-		return(-1);
+	if (errno != EXDEV || s_isdir(stp1)) return(-1);
 	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0)
 		return(-1);
-#ifdef	HAVEFLAGS
-	stp1 -> st_flags = 0xffffffff;
-#endif
+
+	stp1 -> st_nlink = (TCH_MODE | TCH_UID | TCH_GID
+		| TCH_ATIME | TCH_MTIME);
 	return (touchfile(dest, stp1));
 }
 
@@ -775,7 +840,7 @@ int len;
 	int i, j, c;
 
 	if (!buf) {
-		for (i = 0; i < sizeof(seq) / sizeof(char); i++) {
+		for (i = 0; i < (int)sizeof(seq) / sizeof(char); i++) {
 			j = genrand(sizeof(seq) / sizeof(char));
 			c = seq[i];
 			seq[i] = seq[j];
@@ -789,6 +854,7 @@ int len;
 		}
 		buf[i] = '\0';
 	}
+
 	return(buf);
 }
 
@@ -804,7 +870,7 @@ char *dir;
 	}
 	realpath2(deftmpdir, path, 1);
 	free(deftmpdir);
-#if	MSDOS && !defined (_NODOSDRIVE)
+#if	MSDOS && !defined (_NOUSELFN) && !defined (_NODOSDRIVE)
 	if (checkdrive(toupper2(path[0]) - 'A') && !nodosgetwd(path)) {
 		lostcwd(path);
 		deftmpdir = NULL;
@@ -814,7 +880,7 @@ char *dir;
 	deftmpdir = strdup2(path);
 #if	MSDOS
 	n = getcurdrv();
-	*path = (n >= 'A' && n <= 'Z') ? toupper2(*path) : tolower2(*path);
+	*path = (isupper2(n)) ? toupper2(*path) : tolower2(*path);
 #endif
 
 	mask = 0777 & ~tmpumask;
@@ -827,9 +893,9 @@ char *dir;
 		}
 	}
 	if (!tmpfilename) {
-		n = sizeof(TMPPREFIX) - 1;
+		n = (int)sizeof(TMPPREFIX) - 1;
 		strncpy(cp, TMPPREFIX, n);
-		len = sizeof(path) - 1 - (cp - path);
+		len = (int)sizeof(path) - 1 - (cp - path);
 		if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 		len -= n;
 		genrandname(NULL, 0);
@@ -848,7 +914,7 @@ char *dir;
 	}
 
 	strncpy((cp = strcatdelim(path)), dir, n);
-	len = sizeof(path) - 1 - (cp - path);
+	len = (int)sizeof(path) - 1 - (cp - path);
 	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 	len -= n;
 	genrandname(NULL, 0);
@@ -868,6 +934,7 @@ char *dir;
 		nodosrmdir(path);
 		errno = n;
 	}
+
 	return(-1);
 }
 
@@ -888,6 +955,7 @@ char *dir;
 	}
 	else if (errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
 		return(-1);
+
 	return(0);
 }
 
@@ -900,7 +968,7 @@ char *file;
 	path[0] = '\0';
 	if (mktmpdir(path) < 0) return(-1);
 	cp = strcatdelim(path);
-	len = sizeof(path) - 1 - (cp - path);
+	len = (int)sizeof(path) - 1 - (cp - path);
 	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 	genrandname(NULL, 0);
 
@@ -917,6 +985,7 @@ char *file;
 	duperrno = errno;
 	rmtmpdir(NULL);
 	errno = duperrno;
+
 	return(-1);
 }
 
@@ -925,6 +994,7 @@ char *file;
 {
 	if ((Xunlink(file) < 0 && errno != ENOENT) || rmtmpdir(NULL) < 0)
 		return(-1);
+
 	return(0);
 }
 
@@ -938,6 +1008,7 @@ static int dormdir(path)
 char *path;
 {
 	if (isdotdir(path)) return(0);
+
 	return(Xrmdir(path));
 }
 
@@ -994,12 +1065,13 @@ char *dir, *file;
 #if	MSDOS
 	spawnlpe(P_WAIT, "DELTREE.EXE", "DELTREE", "/Y", buf, NULL, environ);
 #else
-	if ((pid = fork()) < 0L) return(-1);
+	if ((pid = fork()) < (p_id_t)0) return(-1);
 	else if (!pid) {
 		execle("/bin/rm", "rm", "-rf", buf, NULL, environ);
 		_exit(1);
 	}
 #endif
+
 	return(0);
 }
 
@@ -1013,6 +1085,7 @@ int drive;
 	path[1] = drive;
 	path[2] = '\0';
 	if (mktmpdir(path) < 0) return(NULL);
+
 	return(strdup2(path));
 }
 
@@ -1067,6 +1140,7 @@ int single;
 		return(-1);
 	}
 	*dirp = tmpdir;
+
 	return(drive);
 }
 
@@ -1101,6 +1175,7 @@ char *file;
 	struct stat st;
 
 	if (Xlstat(file, &st) < 0 && errno == ENOENT) return(0);
+
 	return(1);
 }
 
@@ -1138,6 +1213,7 @@ int dos, boundary, dirsize, ofs;
 		len += ofs;
 		if (lfn) return((len / LFNENTSIZ + 1) * DOSDIRENT + DOSDIRENT);
 	}
+
 	return(DOSDIRENT);
 }
 
@@ -1153,6 +1229,7 @@ int size, dos, boundary, dirsize, ofs;
 		size -= DOSDIRENT;
 		return((size / DOSDIRENT) * LFNENTSIZ - 1);
 	}
+
 	return(DOSBODYLEN);
 }
 
@@ -1168,6 +1245,7 @@ char *from, *to;
 	from = nodospath(fpath, from);
 	to = nodospath(tpath, to);
 #endif
+
 	return(Xrename(from, to));
 }
 
@@ -1214,9 +1292,10 @@ char *tmpdir, *old;
 					else break;
 				}
 				else {
-					if ((fd = Xopen(fname,
-					O_CREAT | O_EXCL, 0600))) {
-						Xclose(fd);
+					fd = Xopen(fname,
+						O_CREAT | O_EXCL, 0600);
+					if (fd >= 0) {
+						VOID_C Xclose(fd);
 						return(fname);
 					}
 					if (errno != EEXIST) break;
@@ -1229,6 +1308,7 @@ char *tmpdir, *old;
 		}
 	}
 	free(fname);
+
 	return(NULL);
 }
 
@@ -1239,6 +1319,7 @@ char *dir;
 	struct stat st;
 
 	if (Xlstat(dir, &st) < 0) return(getblocksize(dir));
+
 	return((off_t)st.st_size);
 }
 
@@ -1248,7 +1329,7 @@ off_t bsiz;
 {
 	struct stat st;
 	u_char ch, *tmp;
-	int i, c, n, fd;
+	int i, n, fd;
 
 	if (Xlstat(dir, &st) < 0
 	|| (fd = Xopen(dir, O_BINARY | O_RDONLY, 0666)) < 0)
@@ -1261,16 +1342,14 @@ off_t bsiz;
 			free(tmp);
 			return(NULL);
 		}
-		while ((c = Xread(fd, (char *)&ch, sizeof(ch))) < 0
-		&& errno == EINTR);
-		if (c < 0) {
+		if (sureread(fd, &ch, sizeof(ch)) < 0) {
 			free(tmp);
 			return(NULL);
 		}
 		tmp[i] = ch + 1;
 	}
 	tmp[i] = 0;
-	Xclose(fd);
+	VOID_C Xclose(fd);
 
 	return(tmp);
 }
@@ -1306,31 +1385,31 @@ int fs;
 	off_t persec, totalent, dirblocksize;
 	u_char *entnum;
 	char **tmpfiles;
-	int n, tmpno, block, ptr, totalptr;
+	int n, tmpno, block, ptr, totalptr, headbyte;
 #endif
 	DIR *dirp;
 	struct dirent *dp;
-	int dos, headbyte, boundary, dirsize, namofs;
+	int dos, boundary, dirsize, namofs;
 	int i, top, size, len, fnamp, ent;
 	char *cp, *tmpdir, **fnamelist, path[MAXPATHLEN];
 
 	switch (fs) {
 #if	!MSDOS
-		case 2:	/* IRIX File System */
+		case FSID_EFS:		/* IRIX File System */
 			dos = 0;
 			headbyte = 4;
 			boundary = 2;
 			dirsize = sizeof(u_long);
 			namofs = 0;
 			break;
-		case 3:	/* SystemV R3 File System */
+		case FSID_SYSV:		/* SystemV R3 File System */
 			dos = 0;
 			headbyte = 0;
 			boundary = 8;
 			dirsize = sizeof(u_short);
 			namofs = 0;
 			break;
-		case 6:	/* Linux File System */
+		case FSID_LINUX:	/* Linux File System */
 			dos = 0;
 			headbyte = 0;
 			boundary = 4;
@@ -1338,7 +1417,7 @@ int fs;
 			namofs = 3;
 			break;
 # ifndef	_NODOSDRIVE
-		case 7:	/* Windows95 File System on DOSDRIVE */
+		case FSID_DOSDRIVE:	/* Windows95 File System on DOSDRIVE */
 			dos = 3;
 			headbyte = -1;
 			boundary = LFNENTSIZ;
@@ -1347,23 +1426,29 @@ int fs;
 			break;
 # endif
 #endif	/* !MSDOS */
-		case 4:	/* MS-DOS File System */
+		case FSID_FAT:		/* MS-DOS File System */
 			dos = 1;
+#if	!MSDOS
 			headbyte = -1;
+#endif
 			boundary = 1;
 			dirsize = DOSDIRENT;
 			namofs = 0;
 			break;
-		case 5:	/* Windows95 File System */
+		case FSID_LFN:		/* Windows95 File System */
 			dos = 2;
+#if	!MSDOS
 			headbyte = -1;
+#endif
 			boundary = LFNENTSIZ;
 			dirsize = DOSDIRENT;
 			namofs = 0;
 			break;
 		default:
 			dos = 0;
+#if	!MSDOS
 			headbyte = 0;
+#endif
 			boundary = 4;
 			dirsize = 8;	/* long + short + short */
 			namofs = 0;
@@ -1454,7 +1539,7 @@ int fs;
 	}
 
 #if	!MSDOS
-	if (fs != 2) entnum = NULL;	/* except IRIX File System */
+	if (fs != FSID_EFS) entnum = NULL;	/* except IRIX File System */
 	else if (!(entnum = getentnum(".", persec))) {
 		warning(-1, ".");
 		restorefile(tmpdir, path, fnamp);
@@ -1480,13 +1565,13 @@ int fs;
 		size = realdirsiz(fnamelist[i],
 			dos, boundary, dirsize, namofs);
 		switch (fs) {
-			case 2:	/* IRIX File System */
+			case FSID_EFS:	/* IRIX File System */
 				if (totalptr > ptr + 1) ent -= totalptr;
 				else ent -= ptr + 1;
 				break;
-			case 3:	/* SystemV R3 File System */
-			case 4:	/* MS-DOS File System */
-			case 5:	/* Windows95 File System */
+			case FSID_SYSV:	/* SystemV R3 File System */
+			case FSID_FAT:	/* MS-DOS File System */
+			case FSID_LFN:	/* Windows95 File System */
 				ent = size;
 				break;
 			default:
