@@ -1,7 +1,7 @@
 /*
  *	func.h
  *
- *	Function Prototype Declaration
+ *	function prototype declarations
  */
 
 #ifdef	NOERRNO
@@ -77,15 +77,38 @@ extern char *sys_errlist[];
 #define	strerror2(n)		(char *)sys_errlist[n]
 #endif
 
+#ifdef	_NOPTY
+#define	isptymode()		0
+#else	/* !_NOPTY */
+#define	isptymode()		ptymode
+#endif	/* !_NOPTY */
+
 /* main.c */
 extern VOID error __P_((char *));
+extern VOID setlinecol __P_((VOID_A));
 extern VOID checkscreen __P_((int, int));
+#ifdef	SIGWINCH
+extern VOID pollscreen __P_((int));
+#endif
 extern int sigvecset __P_((int));
 extern VOID title __P_((VOID_A));
 #ifndef	_NOCUSTOMIZE
 extern VOID saveorigenviron __P_((VOID_A));
 #endif
 extern int loadruncom __P_((char *, int));
+extern VOID initfd __P_((char **));
+extern VOID prepareexitfd __P_((int));
+
+/* system.c */
+#ifdef	_NOORIGSHELL
+#define	dosystem		system
+#define	dopopen(c)		Xpopen(c, "r")
+#define	dopclose		Xpclose
+#else
+extern int dosystem __P_((char *));
+extern FILE *dopopen __P_((char *));
+extern int dopclose __P_((FILE *));
+#endif
 
 /* dosemu.c or unixemu.c */
 #ifdef	_USEDOSPATH
@@ -108,7 +131,7 @@ extern VOID Xrewinddir __P_((DIR *));
 #if	MSDOS
 extern int rawchdir __P_((char *));
 #else
-#define	rawchdir	chdir
+#define	rawchdir(p)	((chdir(p)) ? -1 : 0)
 #endif
 extern int Xchdir __P_((char *));
 extern char *Xgetwd __P_((char *));
@@ -123,11 +146,17 @@ extern int Xutime __P_((char *, struct utimbuf *));
 #else
 extern int Xutimes __P_((char *, struct timeval []));
 #endif
+#ifdef	HAVEFLAGS
+extern int Xchflags __P_((char *, u_long));
+#endif
+#ifndef	NOUID
+extern int Xchown __P_((char *, uid_t, gid_t));
+#endif
 extern int Xunlink __P_((char *));
 extern int Xrename __P_((char *, char *));
 extern int Xopen __P_((char *, int, int));
 #ifdef	_NODOSDRIVE
-#define	Xclose		close
+#define	Xclose(f)	((close(f)) ? -1 : 0)
 #define	Xread		read
 #define	Xwrite		write
 #define	Xlseek		lseek
@@ -137,9 +166,9 @@ extern int Xread __P_((int, char *, int));
 extern int Xwrite __P_((int, char *, int));
 extern off_t Xlseek __P_((int, off_t, int));
 #endif
-#if	!defined (LSI_C) && defined (_NODOSDRIVE)
-#define	Xdup		dup
-#define	Xdup2		dup2
+#ifdef	_NODOSDRIVE
+#define	Xdup		safe_dup
+#define	Xdup2		safe_dup2
 #else
 extern int Xdup __P_((int));
 extern int Xdup2 __P_((int, int));
@@ -150,6 +179,7 @@ extern FILE *Xfopen __P_((char *, char *));
 #ifdef	_NODOSDRIVE
 #define	Xfdopen		fdopen
 #define	Xfclose		fclose
+#define	Xfileno		fileno
 #define	Xfeof		feof
 #define	Xfread		fread
 #define	Xfwrite		fwrite
@@ -161,6 +191,7 @@ extern FILE *Xfopen __P_((char *, char *));
 #else
 extern FILE *Xfdopen __P_((int, char *));
 extern int Xfclose __P_((FILE *));
+extern int Xfileno __P_((FILE *));
 extern int Xfeof __P_((FILE *));
 extern int Xfread __P_((char *, int, int, FILE *));
 extern int Xfwrite __P_((char *, int, int, FILE *));
@@ -180,12 +211,96 @@ extern int Xpclose __P_((FILE *));
 # endif
 #endif	/* _NOORIGSHELL */
 
+/* pty.c */
+#ifndef	_NOPTY
+extern int Xopenpty __P_((int *, char *, ALLOC_T));
+extern int Xlogin_tty __P_((char *, char *, char *));
+extern p_id_t Xforkpty __P_((int *, char *, char *));
+#endif
+
+/* termemu.c */
+#ifdef	_NOPTY
+#define	ptymacro		execmacro
+#define	ptyusercomm		execusercomm
+#define	ptysystem		dosystem
+#else	/* !_NOPTY */
+extern VOID regionscroll __P_((int, int, int, int, int, int));
+extern int selectpty __P_((int, int [], char [], int));
+extern VOID syncptyout __P_((int, int));
+extern int recvbuf __P_((int, VOID_P, int));
+extern VOID sendbuf __P_((int, VOID_P, int));
+extern int recvword __P_((int, int *));
+extern VOID sendword __P_((int, int));
+extern int recvstring __P_((int, char **));
+extern VOID sendstring __P_((int, char *));
+extern VOID sendparent __P_((int, ...));
+extern int ptymacro __P_((char *, char *, int));
+# ifdef	_NOORIGSHELL
+# define	ptyusercomm(c,a,f)	ptymacro(c, a, f | F_EVALMACRO)
+# else
+# define	ptyusercomm		ptymacro
+# endif
+#define	ptysystem(c)		ptymacro(c, NULL, F_DOSYSTEM)
+extern VOID killpty __P_((int, int *));
+extern VOID killallpty __P_((VOID_A));
+#endif	/* !_NOPTY */
+
+/* frontend.c */
+#ifdef	_NOPTY
+#define	Xttyiomode		ttyiomode
+#define	Xstdiomode		stdiomode
+#define	Xtermmode		termmode
+#define	Xputch2			putch2
+#define	Xcputs2			cputs2
+#define	Xputterm		putterm
+#define	Xputterms		putterms
+#define	Xsetscroll		setscroll
+#define	Xlocate			locate
+#define	Xtflush			tflush
+#define	Xcprintf2		cprintf2
+#define	Xcputnl			cputnl
+#define	Xkanjiputs		kanjiputs
+#define	Xchgcolor		chgcolor
+#define	Xmovecursor		movecursor
+#else	/* !_NOPTY */
+extern int waitstatus __P_((p_id_t, int, int *));
+extern VOID Xttyiomode __P_((int));
+extern VOID Xstdiomode __P_((VOID_A));
+extern int Xtermmode __P_((int));
+extern VOID Xputch2 __P_((int));
+extern VOID Xcputs2 __P_((char *));
+extern VOID Xputterm __P_((int));
+extern VOID Xputterms __P_((int));
+extern int Xsetscroll __P_((int, int));
+extern VOID Xlocate __P_((int, int));
+extern VOID Xtflush __P_((VOID_A));
+extern int Xcprintf2 __P_((CONST char *, ...));
+extern VOID Xcputnl __P_((VOID_A));
+extern int Xkanjiputs __P_((char *));
+extern VOID Xchgcolor __P_((int, int));
+extern VOID Xmovecursor __P_((int, int, int));
+extern VOID changewin __P_((int, p_id_t));
+extern VOID changewsize __P_((int, int));
+extern VOID insertwin __P_((int, int));
+extern VOID deletewin __P_((int, int));
+# ifndef	_NOKANJICONV
+extern VOID changekcode __P_((VOID_A));
+# endif
+extern int frontend __P_((VOID_A));
+#endif	/* !_NOPTY */
+
+/* backend.c */
+#ifndef	_NOPTY
+extern VOID resetptyterm __P_((int, int));
+extern int backend __P_((VOID_A));
+#endif
+
 /* libc.c */
 extern int stat2 __P_((char *, struct stat *));
 extern char *realpath2 __P_((char *, char *, int));
 extern int _chdir2 __P_((char *));
 extern int chdir2 __P_((char *));
-extern int chdir3 __P_((char *));
+extern int chdir3 __P_((char *, int));
 extern int mkdir2 __P_((char *, int));
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
@@ -200,8 +315,7 @@ extern int strncpy3 __P_((char *, char *, int *, int));
 extern int strlen2 __P_((char *));
 extern int strlen3 __P_((char *));
 extern int atoi2 __P_((char *));
-extern int fprintf2 __P_((FILE *, CONST char *, ...));
-extern int snprintf2 __P_((char *, int, CONST char *, ...));
+extern char *asprintf3 __P_((CONST char *, ...));
 extern VOID perror2 __P_((char *));
 extern int setenv2 __P_((char *, char *, int));
 extern char *getenv2 __P_((char *));
@@ -211,7 +325,7 @@ extern sigcst_t signal2 __P_((int, sigcst_t));
 #define	signal2	signal
 #endif
 extern int system2 __P_((char *, int));
-extern FILE *popen2 __P_((char *, char *));
+extern FILE *popen2 __P_((char *));
 extern int pclose2 __P_((FILE *));
 extern char *getwd2 __P_((VOID_A));
 extern time_t time2 __P_((VOID_A));
@@ -238,6 +352,11 @@ extern int cmptree __P_((CONST VOID_P, CONST VOID_P));
 extern struct dirent *searchdir __P_((DIR *, reg_t *, char *));
 extern int underhome __P_((char *));
 extern int preparedir __P_((char *));
+#ifdef	NOFLOCK
+#define	lockfile(f, m)
+#else
+extern int lockfile __P_((int, int));
+#endif
 extern int touchfile __P_((char *, struct stat *));
 #ifdef	_NODOSDRIVE
 #define	nodoslstat	Xlstat
@@ -245,7 +364,12 @@ extern int touchfile __P_((char *, struct stat *));
 extern int nodoslstat __P_((char *, struct stat *));
 #endif
 extern VOID lostcwd __P_((char *));
-extern int safewrite __P_((int, char *, int));
+#ifndef	NODIRLOOP
+extern int issamebody __P_((char *, char *));
+#endif
+#ifndef	NOSYMLINK
+extern int cpsymlink __P_((char *, char *));
+#endif
 extern int safecpfile __P_((char *, char *, struct stat *, struct stat *));
 extern int safemvfile __P_((char *, char *, struct stat *, struct stat *));
 extern char *genrandname __P_((char *, int));
@@ -280,8 +404,7 @@ extern int forcemovefile __P_((char *));
 
 /* parse.c */
 extern char *skipspace __P_((char *));
-extern char *evalnumeric __P_((char *, long *, int));
-extern char *ascnumeric __P_((char *, off_t, int, int));
+extern char *sscanf2 __P_((char *, CONST char *, ...));
 #ifdef	_NOORIGSHELL
 extern char *strtkchr __P_((char *, int, int));
 extern int getargs __P_((char *, char ***));
@@ -300,52 +423,73 @@ extern VOID adjustpath __P_((VOID_A));
 #endif
 extern char *includepath __P_((char *, char *));
 #if	(FD < 2) && !defined (_NOARCHIVE)
-extern char *getrange __P_((char *, u_char *, u_char *, u_char *));
+extern char *getrange __P_((char *, int, u_char *, u_char *, u_char *));
 #endif
 extern int evalprompt __P_((char **, char *));
-#if	FD >= 2
-extern char **duplvar __P_((char **, int));
-#endif
 #ifndef	_NOARCHIVE
 extern char *getext __P_((char *, u_char *));
 extern int extcmp __P_((char *, int, char *, int, int));
 #endif
 extern int getkeycode __P_((char *, int));
 extern char *getkeysym __P_((int, int));
-extern char *decodestr __P_((char *, int *, int));
+extern char *decodestr __P_((char *, u_char *, int));
 #ifndef	_NOKEYMAP
 extern char *encodestr __P_((char *, int));
 #endif
 
 /* builtin.c */
 #ifndef	_NOARCHIVE
-extern VOID printlaunchcomm __P_((int, int, FILE *));
-extern VOID printarchcomm __P_((int, FILE *));
+extern VOID freelaunch __P_((launchtable *));
+extern int searchlaunch __P_((launchtable *, int, launchtable *));
+extern int parselaunch __P_((int, char *[], launchtable *));
+extern VOID addlaunch __P_((int, launchtable *));
+extern VOID deletelaunch __P_((int));
+extern VOID freearch __P_((archivetable *));
+extern int searcharch __P_((archivetable *, int, archivetable *));
+extern int parsearch __P_((int, char *[], archivetable *));
+extern VOID addarch __P_((int, archivetable *));
+extern VOID deletearch __P_((int));
+extern VOID printlaunchcomm __P_((launchtable *, int, int, int, FILE *));
+extern VOID printarchcomm __P_((archivetable *, int, int, FILE *));
 # ifndef	_NOBROWSE
 extern VOID freebrowse __P_((launchtable *));
 # endif
 #endif
+extern int ismacro __P_((int));
 extern int freemacro __P_((int));
-extern VOID printmacro __P_((int, FILE *));
+extern int searchkeybind __P_((bindtable *, bindtable *));
+extern int parsekeybind __P_((int, char *[], bindtable *));
+extern int addkeybind __P_((int, bindtable *, char *, char *, char *));
+extern VOID deletekeybind __P_((int));
+extern char *gethelp __P_((bindtable *));
+extern VOID printmacro __P_((bindtable *, int, int, FILE *));
 #ifdef	_USEDOSEMU
-extern int searchdrv __P_((devinfo *, int, char *, int, int, int, int));
+extern int searchdrv __P_((devinfo *, devinfo *, int));
 extern int deletedrv __P_((int));
-extern int insertdrv __P_((int, int, char *, int, int, int));
-extern VOID printsetdrv __P_((int, int, FILE *));
+extern int insertdrv __P_((int, devinfo *));
+extern VOID printsetdrv __P_((devinfo [], int, int, int, FILE *));
+extern int parsesetdrv __P_((int, char *[], devinfo *));
 #endif
 #ifndef	_NOKEYMAP
-extern VOID printkeymap __P_((int, char *, int, FILE *));
+extern int parsekeymap __P_((int, char *[], keyseq_t *));
+extern VOID printkeymap __P_((keyseq_t *, int, FILE *));
 #endif
 #ifdef	_NOORIGSHELL
+extern int searchalias __P_((char *, int));
+extern int addalias __P_((char *, char *));
+extern int deletealias __P_((char *));
 extern VOID printalias __P_((int, FILE *));
+extern int searchfunction __P_((char *));
+extern int addfunction __P_((char *, char **));
+extern int deletefunction __P_((char *));
 extern VOID printfunction __P_((int, int, FILE *));
-#endif
+#endif	/* _NOORIGSHELL */
 extern int checkbuiltin __P_((char *));
 extern int checkinternal __P_((char *));
 extern int execbuiltin __P_((int, int, char *[]));
 extern int execinternal __P_((int, int, char *[]));
 #ifdef	_NOORIGSHELL
-extern int execpseudoshell __P_((char *, int, int));
+extern int execpseudoshell __P_((char *, int));
 #endif
 #ifndef	_NOCOMPLETE
 extern int completebuiltin __P_((char *, int, int, char ***));
@@ -387,24 +531,25 @@ extern VOID freedefine __P_((VOID_A));
 #define	BL_DOT		"."
 
 /* shell.c */
-extern char *evalcommand __P_((char *, char *, macrostat *, int));
+extern char *restorearg __P_((char *));
+extern char *evalcommand __P_((char *, char *, macrostat *));
 extern int replaceargs __P_((int *, char ***, char **, int));
 extern int replacearg __P_((char **));
-extern int argfputs __P_((char *, FILE *));
 extern VOID demacroarg __P_((char **));
 extern char *inputshellstr __P_((char *, int, char *));
 extern char *inputshellloop __P_((int, char *));
 extern int isinternalcomm __P_((char *));
-extern int execmacro __P_((char *, char *, int, int, int));
+extern int execmacro __P_((char *, char *, int));
 extern FILE *popenmacro __P_((char *, char *, int));
 #ifdef	_NOORIGSHELL
-extern int execusercomm __P_((char *, char *, int, int, int));
+extern int execusercomm __P_((char *, char *, int));
 #else
 #define	execusercomm	execmacro
 #endif
 extern int entryhist __P_((int, char *, int));
 extern char *removehist __P_((int));
 extern int loadhistory __P_((int, char *));
+extern VOID convhistory __P_((char *, FILE *));
 extern int savehistory __P_((int, char *));
 extern int parsehist __P_((char *, int *));
 extern char *evalhistory __P_((char *));
@@ -428,21 +573,16 @@ extern int sjis2ujis __P_((char *, u_char *, int));
 extern int ujis2sjis __P_((char *, u_char *, int));
 #endif
 #if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
-extern VOID readunitable __P_((VOID_A));
+extern VOID readunitable __P_((int));
 extern VOID discardunitable __P_((VOID_A));
 extern u_int unifysjis __P_((u_int, int));
 extern u_int cnvunicode __P_((u_int, int));
 #endif
 #ifndef	_NOKANJICONV
 extern int kanjiconv __P_((char *, char *, int, int, int, int));
-# ifndef	_NOKANJIFCONV
-extern char *kanjiconv2 __P_((char *, char *, int, int, int));
-# endif
+extern char *kanjiconv2 __P_((char *, char *, int, int, int, int));
+extern char *newkanjiconv __P_((char *, int, int, int));
 #endif
-extern int kanjifputs __P_((char *, FILE *));
-extern int kanjiputs __P_((char *));
-extern int kanjiprintf __P_((CONST char *, ...));
-extern int kanjiputs2 __P_((char *, int, int));
 #ifndef	_NOKANJIFCONV
 extern int getkcode __P_((char *));
 #endif
@@ -452,6 +592,7 @@ extern char *convput __P_((char *, char *, int, int, char *, int *));
 /* input.c */
 extern int intrkey __P_((VOID_A));
 extern int Xgetkey __P_((int, int));
+extern int kanjiputs2 __P_((char *, int, int));
 extern char *inputstr __P_((char *, int, int, char *, int));
 extern int yesno __P_((CONST char *, ...));
 extern VOID warning __P_((int, char *));
@@ -507,10 +648,10 @@ extern VOID freestrarray __P_((char **, int));
 extern char **copystrarray __P_((char **, char **, int *, int));
 extern bindtable *copybind __P_((bindtable *, bindtable *));
 # ifndef	_NOARCHIVE
-extern VOID freelaunch __P_((launchtable *, int));
+extern VOID freelaunchlist __P_((launchtable *, int));
 extern launchtable *copylaunch __P_((launchtable *, launchtable *,
 	int *, int));
-extern VOID freearch __P_((archivetable *, int));
+extern VOID freearchlist __P_((archivetable *, int));
 extern archivetable *copyarch __P_((archivetable *, archivetable *,
 	int *, int));
 # endif
@@ -518,26 +659,36 @@ extern archivetable *copyarch __P_((archivetable *, archivetable *,
 extern VOID freedosdrive __P_((devinfo *));
 extern devinfo *copydosdrive __P_((devinfo *, devinfo *));
 # endif
-extern VOID rewritecust __P_((int));
+extern VOID rewritecust __P_((VOID_A));
 extern int customize __P_((VOID_A));
 #endif	/* !_NOCUSTOMIZE */
 
 /* command.c */
+extern int evalstatus __P_((int));
+#ifndef	_NOSPLITWIN
+extern int nextwin __P_((VOID_A));
+#endif
 
 /* browse.c */
 extern VOID helpbar __P_((VOID_A));
-extern int putmode __P_((char *, u_int));
+extern int putmode __P_((char *, u_int, int));
 #ifdef	HAVEFLAGS
 extern int putflags __P_((char *, u_long));
 #endif
+#ifndef	NOUID
+extern int putowner __P_((char *, uid_t));
+extern int putgroup __P_((char *, gid_t));
+#endif
 extern VOID waitmes __P_((VOID_A));
+extern int filetop __P_((int));
 extern int calcwidth __P_((VOID_A));
 extern int putname __P_((namelist *, int, int));
-extern int listupfile __P_((namelist *, int, char *));
+extern int listupfile __P_((namelist *, int, char *, int));
 #ifndef	_NOSPLITWIN
 extern int shutwin __P_((int));
 #endif
+extern VOID calcwin __P_((VOID_A));
 extern VOID movepos __P_((int, int));
 extern VOID rewritefile __P_((int));
 extern VOID addlist __P_((VOID_A));
-extern VOID main_fd __P_((char *));
+extern VOID main_fd __P_((char **, int));
