@@ -1,14 +1,12 @@
 /*
  *	info.c
  *
- *	Information Module
+ *	information module
  */
 
 #include <errno.h>
 #include "fd.h"
-#include "term.h"
 #include "funcno.h"
-#include "kctype.h"
 #include "kanji.h"
 
 #ifdef	NOERRNO
@@ -31,18 +29,20 @@ extern int errno;
 #ifdef	USEMNTENTH
 #include <mntent.h>
 typedef struct mntent	mnt_t;
-#define	getmntent2(fp, mntp)	getmntent(fp)
-#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
+#define	setmntent2	setmntent
+#define	getmntent2(f,m)	getmntent(f)
+#define	hasmntopt2(m,o)	strmntopt((m) -> mnt_opts, o)
+#define	endmntent2	endmntent
 #endif	/* USEMNTENTH */
 
 #ifdef	USEMNTTABH
 #include <sys/mnttab.h>
 #define	MOUNTED		MNTTAB
 typedef struct mnttab	mnt_t;
-#define	setmntent		fopen
-#define	getmntent2(fp, mntp)	(getmntent(fp, mntp) ? NULL : mntp)
-#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_mntopts, opt)
-#define	endmntent		fclose
+#define	setmntent2	fopen
+#define	getmntent2(f,m)	(getmntent(f, m) ? NULL : m)
+#define	hasmntopt2(m,o)	strmntopt((m) -> mnt_mntopts, o)
+#define	endmntent2	fclose
 #define	mnt_dir		mnt_mountp
 #define	mnt_fsname	mnt_special
 #define	mnt_type	mnt_fstype
@@ -75,13 +75,13 @@ typedef struct _mnt_t {
 	char *mnt_type;
 	char *mnt_opts;
 } mnt_t;
-static FILE *NEAR setmntent __P_((char *, char *));
+static FILE *NEAR setmntent2 __P_((char *, char *));
 static mnt_t *NEAR getmntent2 __P_((FILE *, mnt_t *));
 #define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
 #if	defined (USEMNTINFO) || defined (USEGETMNT)
-#define	endmntent(fp)
+#define	endmntent2(f)
 #else
-#define	endmntent(fp)		{ if (fp) free(fp); }
+#define	endmntent2		free
 #endif
 static int mnt_ptr = 0;
 static int mnt_size = 0;
@@ -90,10 +90,10 @@ static int mnt_size = 0;
 #ifdef	USEGETFSENT
 #include <fstab.h>
 typedef struct fstab	mnt_t;
-#define	setmntent(file, mode)	(FILE *)(setfsent(), NULL)
+#define	setmntent2(file, mode)	(FILE *)(setfsent(), NULL)
 #define	getmntent2(fp, mntp)	getfsent()
 #define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> fs_mntops, opt)
-#define	endmntent(fp)		endfsent()
+#define	endmntent2(fp)		endfsent()
 #define	mnt_dir		fs_file
 #define	mnt_fsname	fs_spec
 #define	mnt_type	fs_vfstype
@@ -182,11 +182,11 @@ typedef struct fs_data	statfs_t;
 
 #if	MSDOS
 typedef struct _statfs_t {
-	long	f_bsize;
-	long	f_blocks;
-	long	f_bfree;
-	long	f_bavail;
-	long	f_files;
+	long f_bsize;
+	long f_blocks;
+	long f_bfree;
+	long f_bavail;
+	long f_files;
 } statfs_t;
 extern int unixstatfs __P_((char *, statfs_t *));
 #define	statfs2		unixstatfs
@@ -196,19 +196,12 @@ extern int unixstatfs __P_((char *, statfs_t *));
 extern VOID error __P_((char *));
 extern int _chdir2 __P_((char *));
 extern char *strcpy2 __P_((char *, char *));
-extern int snprintf2 __P_((char *, int, CONST char *, ...));
 extern char *getwd2 __P_((VOID_A));
 extern VOID warning __P_((int, char *));
 #ifdef	_USEDOSPATH
 extern int dospath __P_((char *, char *));
 #endif
 #if	MSDOS && !defined (_NOUSELFN)
-# ifdef	LSI_C
-# define	toupper2	toupper
-# else	/* !LSI_C */
-# define	toupper2(c)	(uppercase[(u_char)(c)])
-extern u_char uppercase[256];
-# endif	/* !LSI_C */
 extern int supportLFN __P_((char *));
 # ifndef	_NODOSDRIVE
 extern int checkdrive __P_((int));
@@ -219,16 +212,28 @@ extern int dosstatfs __P_((int, char *));
 #endif
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
-extern char *ascnumeric __P_((char *, off_t, int, int));
-extern int kanjiputs __P_((char *));
-extern int kanjiputs2 __P_((char *, int, int));
+#if	!MSDOS
+extern int Xstat __P_((char *, struct stat *));
+#endif
 extern int Xaccess __P_((char *, int));
-extern int kanjiprintf __P_((CONST char *, ...));
+extern int filetop __P_((int));
+#ifdef	_NOPTY
+#define	Xlocate			locate
+#define	Xputterm		putterm
+#define	Xputch2			putch2
+#define	Xcputs2			cputs2
+#define	Xcprintf2		cprintf2
+#else
+extern VOID Xlocate __P_((int, int));
+extern VOID Xputterm __P_((int));
+extern VOID Xputch2 __P_((int));
+extern VOID Xcputs2 __P_((char *));
+extern VOID Xcprintf2 __P_((CONST char *, ...));
+#endif
 
 extern bindtable bindlist[];
 extern functable funclist[];
 extern char fullpath[];
-extern int sizeinfo;
 extern char *distributor;
 #ifndef	_NODOSDRIVE
 extern int needbavail;
@@ -264,6 +269,9 @@ extern int needbavail;
 #ifndef	MNTTYPE_EXT2
 #define	MNTTYPE_EXT2	"ext2"	/* Linux */
 #endif
+#ifndef	MNTTYPE_EXT3
+#define	MNTTYPE_EXT3	"ext3"	/* Linux */
+#endif
 #ifndef	MNTTYPE_JFS
 #define	MNTTYPE_JFS	"jfs"	/* AIX */
 #endif
@@ -311,22 +319,59 @@ static int NEAR getfsinfo __P_((char *, statfs_t *, mnt_t *));
 static char *NEAR strmntopt __P_((char *, char *));
 int writablefs __P_((char *));
 off_t getblocksize __P_((char *));
-static int NEAR info1line __P_((int, char *, off_t, char *, char *));
+static int NEAR info1line __P_((int, int, char *, off_t, char *, char *));
 off_t calcKB __P_((off_t, off_t));
 int getinfofs __P_((char *, off_t *, off_t *, off_t *));
 int infofs __P_((char *));
 
-static int keycodelist[] = {
+static CONST int keycodelist[] = {
 	K_HOME, K_END, K_DL, K_IL, K_DC, K_IC,
 	K_BEG, K_EOL, K_NPAGE, K_PPAGE, K_CLR, K_ENTER, K_HELP,
 	K_BS, '\t', K_CR, K_ESC
 };
 #define	KEYCODESIZ	((int)(sizeof(keycodelist) / sizeof(int)))
-static char *keystrlist[] = {
+static CONST char *keystrlist[] = {
 	"Home", "End", "DelLin", "InsLin", "Del", "Ins",
 	"Beg", "Eol", "PageDn", "PageUp", "Clr", "Enter", "Help",
 	"Bs", "Tab", "Ret", "Esc"
 };
+static CONST strtable mntlist[] = {
+	{FSID_FAT, MNTTYPE_PC},
+	{FSID_LFN, MNTTYPE_DOS7},
+	{FSID_LFN, MNTTYPE_FAT32},
+#if	MSDOS
+# ifndef	_NODOSDRIVE
+	{FSID_FAT, MNTTYPE_FAT12},
+	{FSID_FAT, MNTTYPE_FAT16},
+	{FSID_FAT, MNTTYPE_FAT16X},
+# endif
+#else	/* !MSDOS */
+	{FSID_UFS, MNTTYPE_43},
+	{FSID_UFS, MNTTYPE_42},
+	{FSID_UFS, MNTTYPE_UFS},
+	{FSID_UFS, MNTTYPE_FFS},
+	{FSID_UFS, MNTTYPE_JFS},
+	{FSID_EFS, MNTTYPE_EFS},
+	{FSID_SYSV, MNTTYPE_SYSV},
+	{FSID_SYSV, MNTTYPE_DGUX},
+	{FSID_LINUX, MNTTYPE_EXT2},
+	{FSID_LINUX, MNTTYPE_EXT3},
+# ifdef	LINUX
+	{FSID_FAT, MNTTYPE_MSDOS},
+# else
+	{FSID_LFN, MNTTYPE_MSDOS},
+# endif
+	{FSID_LFN, MNTTYPE_UMSDOS},
+	{FSID_LFN, MNTTYPE_VFAT},
+	{0, MNTTYPE_ADVFS},
+	{0, MNTTYPE_VXFS},
+#endif	/* !MSDOS */
+#ifdef	DARWIN
+	/* Macintosh HFS+ is pseudo file system covered with skin */
+	{0, MNTTYPE_HFS},
+#endif
+};
+#define	MNTLISTSIZ	((int)(sizeof(mntlist) / sizeof(strtable)))
 
 
 static int NEAR code2str(buf, code)
@@ -336,103 +381,111 @@ int code;
 	buf = buf + strlen(buf);
 	if (code >= K_F(1) && code <= K_F(20))
 		snprintf2(buf, KEYWID + 1, "F%-6d", code - K_F0);
-	else if ((code & ~0x7f) == 0x80 && isalpha(code & 0x7f))
+	else if ((code & ~0x7f) == 0x80 && isalpha2(code & 0x7f))
 		snprintf2(buf, KEYWID + 1, "Alt-%c  ", code & 0x7f);
 	else if (code == K_UP)
-		snprintf2(buf, KEYWID + 1, "%-7.7s", UPAR_K);
+		snprintf2(buf, KEYWID + 1, "%-*.*s", KEYWID, KEYWID, UPAR_K);
 	else if (code == K_DOWN)
-		snprintf2(buf, KEYWID + 1, "%-7.7s", DWNAR_K);
+		snprintf2(buf, KEYWID + 1, "%-*.*s", KEYWID, KEYWID, DWNAR_K);
 	else if (code == K_RIGHT)
-		snprintf2(buf, KEYWID + 1, "%-7.7s", RIGAR_K);
+		snprintf2(buf, KEYWID + 1, "%-*.*s", KEYWID, KEYWID, RIGAR_K);
 	else if (code == K_LEFT)
-		snprintf2(buf, KEYWID + 1, "%-7.7s", LEFAR_K);
+		snprintf2(buf, KEYWID + 1, "%-*.*s", KEYWID, KEYWID, LEFAR_K);
 	else {
 		int i;
 
 		for (i = 0; i < KEYCODESIZ; i++)
 			if (code == keycodelist[i]) break;
 		if (i < KEYCODESIZ)
-			snprintf2(buf, KEYWID + 1, "%-7.7s", keystrlist[i]);
+			snprintf2(buf, KEYWID + 1, "%-*.*s",
+				KEYWID, KEYWID, keystrlist[i]);
+#ifdef	CODEEUC
+		else if (isekana2(code))
+			snprintf2(buf, KEYWID + 1 + 1,
+				"'%c%c'    ", C_EKANA, code & 0xff);
+#endif
+		else if (code > MAXUTYPE(u_char)) return(0);
 #ifndef	CODEEUC
-		else if (iskna(code))
+		else if (iskana2(code))
 			snprintf2(buf, KEYWID + 1, "'%c'    ", code);
 #endif
-		else if (isctl(code))
+		else if (iscntrl2(code))
 			snprintf2(buf, KEYWID + 1,
 				"Ctrl-%c ", (code + '@') & 0x7f);
-		else if (code < K_MIN && !ismsb(code))
+		else if (!ismsb(code))
 			snprintf2(buf, KEYWID + 1, "'%c'    ", code);
 		else return(0);
 	}
+
 	return(1);
 }
 
 static int NEAR checkline(y)
 int y;
 {
-	if (y >= LFILEBOTTOM) {
+	if (y >= FILEPERROW) {
 		warning(0, HITKY_K);
-		y = LFILETOP + 1;
+		y = 1;
 	}
+
 	return(y);
 }
 
 VOID help(mode)
 int mode;
 {
-	char buf[20];
-	int i, j, c, x, y;
+	char buf[(KEYWID + 1 + 1) * 2 + 1];
+	int i, j, c, x, y, yy;
 
 	if (distributor && *distributor) {
 		i = n_column - (int)strlen(distributor) - 24;
-		locate(i, L_HELP);
-		putch2('[');
-		putterm(t_standout);
-		cprintf2(" Distributed by: %s ", distributor);
-		putterm(end_standout);
-		putch2(']');
+		Xlocate(i, L_HELP);
+		Xputch2('[');
+		Xputterm(T_STANDOUT);
+		Xcprintf2(" Distributed by: %s ", distributor);
+		Xputterm(END_STANDOUT);
+		Xputch2(']');
 	}
 
-	x = 0;
-	y = LFILETOP;
-	locate(0, y++);
-	putterm(l_clear);
+	yy = filetop(win);
+	x = y = 0;
+	Xlocate(0, yy + y++);
+	Xputterm(L_CLEAR);
 
 	for (i = 0; i < FUNCLISTSIZ ; i++) {
 		if (i == NO_OPERATION || i == WARNING_BELL) continue;
-		locate(x * (n_column / 2), y);
-		if (x ^= 1) putterm(l_clear);
+		Xlocate(x * (n_column / 2), yy + y);
+		if (x ^= 1) Xputterm(L_CLEAR);
 		else y++;
-		if (mode && !(funclist[i].status & ARCH)) continue;
+		if (mode && !(funclist[i].status & FN_ARCHIVE)) continue;
 
 		c = 0;
 		buf[0] = '\0';
 		for (j = 0; j < MAXBINDTABLE && bindlist[j].key >= 0; j++)
 			if (i == (int)(bindlist[j].f_func)) {
-				if ((c += code2str(buf,
-					(int)(bindlist[j].key))) >= 2) break;
+				c += code2str(buf, (int)(bindlist[j].key));
+				if (c >= 2) break;
 			}
 		if (c < 2)
 		for (j = 0; j < MAXBINDTABLE && bindlist[j].key >= 0; j++)
 			if (i == (int)(bindlist[j].d_func)) {
-				if ((c += code2str(buf,
-					(int)(bindlist[j].key))) >= 2) break;
+				c += code2str(buf, (int)(bindlist[j].key));
+				if (c >= 2) break;
 			}
 		if (!c) continue;
 
-		cputs2("  ");
-		if (c < 2) cprintf2("%-7.7s", " ");
-		kanjiputs(buf);
-		cputs2(": ");
-		kanjiputs(mesconv(funclist[i].hmes, funclist[i].hmes_eng));
+		Xcputs2("  ");
+		if (c < 2) Xcprintf2("%*s", KEYWID, "");
+		Xcprintf2("%k: %k",
+			buf, mesconv(funclist[i].hmes, funclist[i].hmes_eng));
 
 		y = checkline(y);
 	}
-	if (y > LFILETOP + 1) {
+	if (y > 1) {
 		if (x) y++;
-		for (; y < LFILEBOTTOM; y++) {
-			locate(0, y);
-			putterm(l_clear);
+		for (; y < FILEPERROW; y++) {
+			Xlocate(0, yy + y);
+			Xputterm(L_CLEAR);
 		}
 		warning(0, HITKY_K);
 	}
@@ -440,7 +493,7 @@ int mode;
 
 #ifdef	USEMNTCTL
 /*ARGSUSED*/
-static FILE *NEAR setmntent(file, mode)
+static FILE *NEAR setmntent2(file, mode)
 char *file, *mode;
 {
 	char *buf;
@@ -449,6 +502,7 @@ char *file, *mode;
 	buf = malloc2(mnt_size);
 	mntctl(MCTL_QUERY, mnt_size, (struct vmount *)buf);
 	mnt_ptr = 0;
+
 	return((FILE *)buf);
 }
 
@@ -462,7 +516,7 @@ mnt_t *mntp;
 	struct vfs_ent *entp;
 	struct vmount *vmntp;
 	char *cp, *buf, *host;
-	int len;
+	ALLOC_T len;
 
 	if (mnt_ptr >= mnt_size) return(NULL);
 	buf = (char *)fp;
@@ -472,7 +526,7 @@ mnt_t *mntp;
 	len = strlen(cp) + 1;
 	if (!(vmntp -> vmt_flags & MNT_REMOTE)) {
 		fsname = realloc2(fsname, len);
-		strcpy(fsname, cp);
+		memcpy(fsname, cp, len);
 	}
 	else {
 		host = &(buf[mnt_ptr
@@ -485,14 +539,14 @@ mnt_t *mntp;
 	cp = &(buf[mnt_ptr + vmntp -> vmt_data[VMT_STUB].vmt_off]);
 	len = strlen(cp) + 1;
 	dir = realloc2(dir, len);
-	strcpy(dir, cp);
+	memcpy(dir, cp, len);
 
 	entp = getvfsbytype(vmntp -> vmt_gfstype);
 	if (entp) {
 		cp = entp -> vfsent_name;
 		len = strlen(cp) + 1;
 		type = realloc2(type, len);
-		strcpy(type, cp);
+		memcpy(type, cp, len);
 	}
 	else if (type) {
 		free(type);
@@ -503,15 +557,15 @@ mnt_t *mntp;
 	mntp -> mnt_dir = dir;
 	mntp -> mnt_type = (type) ? type : "???";
 	mntp -> mnt_opts = (vmntp -> vmt_flags & MNT_READONLY) ? "ro" : "";
-
 	mnt_ptr += vmntp -> vmt_length;
+
 	return(mntp);
 }
 #endif	/* USEMNTCTL */
 
 #if	defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETFSSTAT)
 /*ARGSUSED*/
-static FILE *NEAR setmntent(file, mode)
+static FILE *NEAR setmntent2(file, mode)
 char *file, *mode;
 {
 #ifndef	USEMNTINFO
@@ -535,6 +589,7 @@ char *file, *mode;
 	}
 # endif
 #endif
+
 	return((FILE *)buf);
 }
 
@@ -546,10 +601,6 @@ static mnt_t *NEAR getmntent2(fp, mntp)
 FILE *fp;
 mnt_t *mntp;
 {
-	static char *fsname = NULL;
-	static char *dir = NULL;
-	static char *type = NULL;
-	struct statfs *buf;
 #ifdef	USEMNTINFO
 # ifdef	USEVFCNAME
 	struct vfsconf *conf;
@@ -574,25 +625,29 @@ mnt_t *mntp;
 					? mnt_names[n] : NULL)
 # endif
 #endif	/* !USEMNTINFO */
+	static char *fsname = NULL;
+	static char *dir = NULL;
+	static char *type = NULL;
+	struct statfs *buf;
 	char *cp;
-	int len;
+	ALLOC_T len;
 
 	if (mnt_ptr >= mnt_size) return(NULL);
 	buf = (struct statfs *)fp;
 
 	len = strlen(buf[mnt_ptr].f_mntfromname) + 1;
 	fsname = realloc2(fsname, len);
-	strcpy(fsname, buf[mnt_ptr].f_mntfromname);
+	memcpy(fsname, buf[mnt_ptr].f_mntfromname, len);
 
 	len = strlen(buf[mnt_ptr].f_mntonname) + 1;
 	dir = realloc2(dir, len);
-	strcpy(dir, buf[mnt_ptr].f_mntonname);
+	memcpy(dir, buf[mnt_ptr].f_mntonname, len);
 
 	cp = (char *)getvfsbynumber(buf[mnt_ptr].f_type);
 	if (cp) {
 		len = strlen(cp) + 1;
 		type = realloc2(type, len);
-		strcpy(type, cp);
+		memcpy(type, cp, len);
 	}
 	else if (type) {
 		free(type);
@@ -603,19 +658,20 @@ mnt_t *mntp;
 	mntp -> mnt_dir = dir;
 	mntp -> mnt_type = (type) ? type : "???";
 	mntp -> mnt_opts = (buf[mnt_ptr].f_flags & MNT_RDONLY) ? "ro" : "";
-
 	mnt_ptr++;
+
 	return(mntp);
 }
 #endif	/* USEMNTINFOR || USEMNTINFO || USEGETFSSTAT */
 
 #ifdef	USEGETMNT
 /*ARGSUSED*/
-static FILE *NEAR setmntent(file, mode)
+static FILE *NEAR setmntent2(file, mode)
 char *file, *mode;
 {
 	mnt_ptr = 0;
-	return(NULL);
+
+	return((FILE *)1);
 }
 
 /*ARGSUSED*/
@@ -627,22 +683,22 @@ mnt_t *mntp;
 	static char *dir = NULL;
 	static char *type = NULL;
 	struct fs_data buf;
-	int len;
+	ALLOC_T len;
 
-	if (getmnt(&mnt_ptr, &buf, sizeof(struct fs_data),
-		NOSTAT_MANY, NULL) <= 0) return(NULL);
+	if (getmnt(&mnt_ptr, &buf, sizeof(buf), NOSTAT_MANY, NULL) <= 0)
+		return(NULL);
 
 	len = strlen(buf.fd_req.devname) + 1;
 	fsname = realloc2(fsname, len);
-	strcpy(fsname, buf.fd_req.devname);
+	memcpy(fsname, buf.fd_req.devname, len);
 
 	len = strlen(buf.fd_req.path) + 1;
 	dir = realloc2(dir, len);
-	strcpy(dir, buf.fd_req.path);
+	memcpy(dir, buf.fd_req.path, len);
 
 	len = strlen(gt_names[buf.fd_req.fstype]) + 1;
 	type = realloc2(type, len);
-	strcpy(type, gt_names[buf.fd_req.fstype]);
+	memcpy(type, gt_names[buf.fd_req.fstype], len);
 
 	mntp -> mnt_fsname = fsname;
 	mntp -> mnt_dir = dir;
@@ -658,11 +714,16 @@ char *path;
 statfs_t *fsbuf;
 mnt_t *mntbuf;
 {
+	statfs_t fs;
+	mnt_t mnt;
 #if	MSDOS
 # ifndef	_NODOSDRIVE
 	int i;
 # endif
 
+	if (!fsbuf) fsbuf = &fs;
+	if (!mntbuf) mntbuf = &mnt;
+
 	mntbuf -> mnt_fsname = "MSDOS";
 	mntbuf -> mnt_dir[0] = dospath(path, NULL);
 	mntbuf -> mnt_dir[1] = ':';
@@ -706,27 +767,24 @@ mnt_t *mntbuf;
 # endif	/* !_NOUSELFN */
 	mntbuf -> mnt_opts = "";
 
-	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0) {
-		if (errno == ENOENT) return(-1);
-		memset((char *)fsbuf, 0xff, sizeof(statfs_t));
-	}
+	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0) return(-1);
 #else	/* !MSDOS */
-# if	!defined (USEMNTENTH) && !defined (USEGETFSENT)
-	mnt_t mnt;
-# endif
 	mnt_t *mntp;
 	FILE *fp;
 	char *dir, fsname[MAXPATHLEN];
-	int len, match;
-#ifndef	_NODOSDRIVE
+	ALLOC_T len, match;
+# ifndef	_NODOSDRIVE
 	int drv;
-#endif
+# endif
+
+	if (!fsbuf) fsbuf = &fs;
+	if (!mntbuf) mntbuf = &mnt;
 
 	dir = NULL;
-	if (!strncmp(path, "/dev/", 4)) {
+	if (!strncmp(path, "/dev/", sizeof("/dev/") - 1)) {
 		if (_chdir2(path) < 0) dir = path;
 	}
-#ifndef	_NODOSDRIVE
+# ifndef	_NODOSDRIVE
 	else if ((drv = dospath(path, NULL))) {
 		static char dosmntdir[4];
 		char buf[sizeof(long) * 3 + 1];
@@ -735,41 +793,42 @@ mnt_t *mntbuf;
 		mntbuf -> mnt_dir = dosmntdir;
 		dosmntdir[0] = drv;
 		strcpy(&(dosmntdir[1]), ":\\");
-		mntbuf -> mnt_type = (drv >= 'a' && drv <= 'z')
-			? MNTTYPE_DOS7 : MNTTYPE_PC;
+		mntbuf -> mnt_type =
+			(islower2(drv)) ? MNTTYPE_DOS7 : MNTTYPE_PC;
 		mntbuf -> mnt_opts = "";
 		if (dosstatfs(drv, buf) < 0) return(-1);
 		if (buf[sizeof(long) * 3] & 001)
 			mntbuf -> mnt_type = MNTTYPE_FAT32;
 		fsbuf -> f_bsize = *((long *)&(buf[sizeof(long) * 0]));
-# ifdef	USEFSDATA
+#  ifdef	USEFSDATA
 		fsbuf -> fd_req.btot = calcKB((off_t)(fsbuf -> f_bsize),
 			(off_t)(*((long *)&(buf[sizeof(long) * 1]))));
 		fsbuf -> fd_req.bfree =
 		fsbuf -> fd_req.bfreen = calcKB((off_t)(fsbuf -> f_bsize),
 			(off_t)(*((long *)&(buf[sizeof(long) * 2]))));
-# else	/* !USEFSDATA */
-#  ifdef	USESTATVFSH
-		fsbuf -> f_frsize = 0;
-#  endif
+#  else	/* !USEFSDATA */
+#   ifdef	USESTATVFSH
+		fsbuf -> f_frsize = 0L;
+#   endif
 		fsbuf -> f_blocks = *((long *)&(buf[sizeof(long) * 1]));
 		fsbuf -> f_bfree =
 		fsbuf -> f_bavail = *((long *)&(buf[sizeof(long) * 2]));
-# endif	/* !USEFSDATA */
-		fsbuf -> f_files = -1;
+#  endif	/* !USEFSDATA */
+		fsbuf -> f_files = -1L;
+
 		return(0);
 	}
-#endif	/* !_NODOSDRIVE */
+# endif	/* !_NODOSDRIVE */
 	else if (_chdir2(path) < 0) return(-1);
 
 	if (!dir) {
 		dir = getwd2();
 		if (_chdir2(fullpath) < 0) error(fullpath);
-		match = 0;
+		match = (ALLOC_T)0;
 
-		fp = setmntent(MOUNTED, "r");
+		if (!(fp = setmntent2(MOUNTED, "rb"))) return(-1);
 		for (;;) {
-#ifdef	DEBUG
+# ifdef	DEBUG
 			_mtrace_file = "getmntent(start)";
 			mntp = getmntent2(fp, &mnt);
 			if (_mtrace_file) _mtrace_file = NULL;
@@ -778,17 +837,18 @@ mnt_t *mntbuf;
 				malloc(0);	/* dummy malloc */
 			}
 			if (!mntp) break;
-#else
+# else
 			if (!(mntp = getmntent2(fp, &mnt))) break;
-#endif
-			if ((len = strlen(mntp -> mnt_dir)) < match
-			|| strncmp(mntp -> mnt_dir, dir, len)
+# endif
+			len = strlen(mntp -> mnt_dir);
+			if (len < match || strncmp(mntp -> mnt_dir, dir, len)
 			|| (mntp -> mnt_dir[len - 1] != _SC_
-			&& dir[len] && dir[len] != _SC_)) continue;
+			&& dir[len] && dir[len] != _SC_))
+				continue;
 			match = len;
 			strcpy(fsname, mntp -> mnt_fsname);
 		}
-		endmntent(fp);
+		endmntent2(fp);
 
 		free(dir);
 		if (!match) {
@@ -797,10 +857,11 @@ mnt_t *mntbuf;
 		}
 		dir = fsname;
 	}
-	fp = setmntent(MOUNTED, "r");
+
+	if (!(fp = setmntent2(MOUNTED, "rb"))) return(-1);
 	while ((mntp = getmntent2(fp, &mnt)))
 		if (!strcmp(dir, mntp -> mnt_fsname)) break;
-	endmntent(fp);
+	endmntent2(fp);
 	if (!mntp) {
 		errno = ENOENT;
 		return(-1);
@@ -809,8 +870,9 @@ mnt_t *mntbuf;
 
 	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0
 	&& (path == dir || statfs2(path, fsbuf) < 0))
-		memset((char *)fsbuf, 0xff, sizeof(statfs_t));
+		return(-1);
 #endif	/* !MSDOS */
+
 	return(0);
 }
 
@@ -818,7 +880,7 @@ static char *NEAR strmntopt(s1, s2)
 char *s1, *s2;
 {
 	char *cp;
-	int len;
+	ALLOC_T len;
 
 	len = strlen(s2);
 	for (cp = s1; cp && *cp;) {
@@ -826,60 +888,32 @@ char *s1, *s2;
 			return(cp);
 		if ((cp = strchr(cp, ','))) cp++;
 	}
+
 	return(NULL);
 }
 
 int writablefs(path)
 char *path;
 {
-	statfs_t fsbuf;
-	mnt_t mntbuf;
 #ifdef	_USEDOSEMU
 	int drv;
 #endif
+	mnt_t mntbuf;
+	int i;
 
 	if (Xaccess(path, R_OK | W_OK | X_OK) < 0) return(-1);
 #ifdef	_USEDOSEMU
 	if ((drv = dospath(path, NULL)))
-		return((drv >= 'A' && drv <= 'Z') ? 4 : 7);
+		return((isupper2(drv)) ? FSID_FAT : FSID_DOSDRIVE);
 #endif
-	if (getfsinfo(path, &fsbuf, &mntbuf) < 0
-	|| hasmntopt2(&mntbuf, "ro")) return(0);
+	if (getfsinfo(path, NULL, &mntbuf) < 0
+	|| hasmntopt2(&mntbuf, "ro"))
+		return(0);
+
+	for (i = 0; i < MNTLISTSIZ; i++)
+		if (!strcmp(mntbuf.mnt_type, mntlist[i].str))
+			return(mntlist[i].no);
 
-	if (!strcmp(mntbuf.mnt_type, MNTTYPE_PC)) return(4);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_DOS7)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FAT32)) return(5);
-#if	MSDOS
-# ifndef	_NODOSDRIVE
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_FAT12)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FAT16)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FAT16X)) return(4);
-# endif
-#else	/* !MSDOS */
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_43)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_42)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_UFS)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FFS)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_JFS)) return(1);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_EFS)) return(2);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_SYSV)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_DGUX)) return(3);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_EXT2)) return(6);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_MSDOS))
-# ifdef	LINUX
-		return(4);
-# else
-		return(5);
-# endif
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_UMSDOS)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_VFAT)) return(5);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_ADVFS)) return(0);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_VXFS)) return(0);
-#endif	/* !MSDOS */
-#ifdef	DARWIN
-	/* Macintosh HFS+ is pseudo file system covered with skin */
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_HFS)) return(0);
-#endif
 	return(0);
 }
 
@@ -891,47 +925,53 @@ char *dir;
 	statfs_t fsbuf;
 
 	if (statfs2(dir, &fsbuf) < 0) return((off_t)1024);
+
 	return((off_t)blocksize(fsbuf));
 #else	/* !MSDOS */
 	statfs_t fsbuf;
-	mnt_t mntbuf;
-# ifndef	DEV_BSIZE
 	struct stat st;
-# endif
 
-	if (!strcmp(dir, ".") && getfsinfo(dir, &fsbuf, &mntbuf) >= 0)
-		return((off_t)blocksize(fsbuf));
+	if (strcmp(dir, ".")) {
+		/* In case of the newborn directory */
+		if (Xstat(dir, &st) >= 0) return((off_t)(st.st_size));
+	}
+	else {
+# ifdef	CYGWIN
+		if (statfs2(dir, &fsbuf) >= 0)
+# else
+		if (getfsinfo(dir, &fsbuf, NULL) >= 0)
+# endif
+			return((off_t)blocksize(fsbuf));
+		if (Xstat(dir, &st) >= 0) return((off_t)(st.st_blksize));
+	}
+
 # ifdef	DEV_BSIZE
 	return(DEV_BSIZE);
 # else
-	if (Xstat(dir, &st) < 0) error(dir);
-	return((off_t)(st.st_size));
+	return((off_t)512);
 # endif
 #endif	/* !MSDOS */
 }
 
-static int NEAR info1line(y, ind, n, s, unit)
-int y;
+static int NEAR info1line(yy, y, ind, n, s, unit)
+int yy, y;
 char *ind;
 off_t n;
 char *s, *unit;
 {
-	char buf[MAXCOLSCOMMA(3) + 1];
 	int width;
 
-	locate(0, y);
-	putterm(l_clear);
-	locate(n_column / 2 - 20, y);
-	kanjiprintf("%-20.20s", ind);
-	locate(n_column / 2 + 2, y);
+	Xlocate(0, yy + y);
+	Xputterm(L_CLEAR);
+	Xlocate(n_column / 2 - 20, yy + y);
+	Xcprintf2("%-20.20k", ind);
+	Xlocate(n_column / 2 + 2, yy + y);
 	if (s) {
 		width = n_column - (n_column / 2 + 2);
-		kanjiputs2(s, width, 0);
-	}
-	else {
-		cputs2(ascnumeric(buf, n, 3, MAXCOLSCOMMA(3)));
-		kanjiprintf(" %s", unit);
+		Xcprintf2("%-*.*k", width, width, s);
 	}
+	else Xcprintf2("%<'*qd %k", MAXCOLSCOMMA(3), n, unit);
+
 	return(checkline(++y));
 }
 
@@ -941,7 +981,7 @@ off_t block, byte;
 	if (block < (off_t)0 || byte <= (off_t)0) return((off_t)-1);
 	if (byte == (off_t)1024) return(block);
 	else if (byte > 1024) {
-		byte = (byte + (off_t)512L) / (off_t)1024;
+		byte = (byte + (off_t)512) / (off_t)1024;
 		return(block * byte);
 	}
 	else {
@@ -955,23 +995,21 @@ char *path;
 off_t *totalp, *freep, *bsizep;
 {
 	statfs_t fsbuf;
-	mnt_t mntbuf;
+	int n;
 
 #ifndef	_NODOSDRIVE
 	needbavail++;
 #endif
-	if (getfsinfo(path, &fsbuf, &mntbuf) < 0) {
-#ifndef	_NODOSDRIVE
-		needbavail--;
-#endif
-		return(-1);
-	}
-	*totalp = fsbuf.f_blocks;
-	*freep = fsbuf.f_bavail;
-	*bsizep = blocksize(fsbuf);
+	n = getfsinfo(path, &fsbuf, NULL);
 #ifndef	_NODOSDRIVE
 	needbavail--;
 #endif
+	if (n < 0) return(-1);
+
+	*totalp = fsbuf.f_blocks;
+	*freep = fsbuf.f_bavail;
+	*bsizep = blocksize(fsbuf);
+
 	return(0);
 }
 
@@ -980,7 +1018,7 @@ char *path;
 {
 	statfs_t fsbuf;
 	mnt_t mntbuf;
-	int y;
+	int y, yy;
 
 #ifndef	_NODOSDRIVE
 	needbavail++;
@@ -992,34 +1030,36 @@ char *path;
 #endif
 		return(0);
 	}
-	y = LFILETOP;
-	locate(0, y++);
-	putterm(l_clear);
+	yy = filetop(win);
+	y = 0;
+	Xlocate(0, yy + y++);
+	Xputterm(L_CLEAR);
 
-	y = info1line(y, FSNAM_K, (off_t)0, mntbuf.mnt_fsname, NULL);
-	y = info1line(y, FSMNT_K, (off_t)0, mntbuf.mnt_dir, NULL);
-	y = info1line(y, FSTYP_K, (off_t)0, mntbuf.mnt_type, NULL);
-	y = info1line(y, FSTTL_K,
+	y = info1line(yy, y, FSNAM_K, (off_t)0, mntbuf.mnt_fsname, NULL);
+	y = info1line(yy, y, FSMNT_K, (off_t)0, mntbuf.mnt_dir, NULL);
+	y = info1line(yy, y, FSTYP_K, (off_t)0, mntbuf.mnt_type, NULL);
+	y = info1line(yy, y, FSTTL_K,
 		calcKB((off_t)(fsbuf.f_blocks), (off_t)blocksize(fsbuf)),
 		NULL, "Kbytes");
-	y = info1line(y, FSUSE_K,
+	y = info1line(yy, y, FSUSE_K,
 		calcKB((off_t)(fsbuf.f_blocks - fsbuf.f_bfree),
 			(off_t)blocksize(fsbuf)),
 		NULL, "Kbytes");
-	y = info1line(y, FSAVL_K,
+	y = info1line(yy, y, FSAVL_K,
 		calcKB((off_t)(fsbuf.f_bavail), (off_t)blocksize(fsbuf)),
 		NULL, "Kbytes");
-	y = info1line(y, FSBSZ_K, (off_t)(fsbuf.f_bsize), NULL, "bytes");
-	y = info1line(y, FSINO_K, (off_t)(fsbuf.f_files), NULL, UNIT_K);
-	if (y > LFILETOP + 1) {
-		for (; y < LFILEBOTTOM; y++) {
-			locate(0, y);
-			putterm(l_clear);
+	y = info1line(yy, y, FSBSZ_K, (off_t)(fsbuf.f_bsize), NULL, "bytes");
+	y = info1line(yy, y, FSINO_K, (off_t)(fsbuf.f_files), NULL, UNIT_K);
+	if (y > 1) {
+		for (; y < FILEPERROW; y++) {
+			Xlocate(0, yy + y);
+			Xputterm(L_CLEAR);
 		}
 		warning(0, HITKY_K);
 	}
 #ifndef	_NODOSDRIVE
 	needbavail--;
 #endif
+
 	return(1);
 }
