@@ -1,7 +1,7 @@
 /*
  *	machine.h
  *
- *	Definition of Machine Depended Identifier
+ *	definitions of machine depended identifier
  */
 
 #ifndef	__MACHINE_H_
@@ -26,6 +26,7 @@
 #define	NOTZFILEH
 #define	USETIMEH
 #define	USEUTIME
+#define	NOFLOCK
 #define	USEMKTIME
 #define	USESTRERROR
 #define	SENSEPERSEC	20
@@ -80,6 +81,7 @@ typedef long	off_t;
 #define	CODEEUC
 #define	CPP7BIT
 #define	USELEAPCNT
+#define	SOCKETLIB	"-lsocket -lnsl"
 #include <sys/param.h>
 # if	!defined (MAXHOSTNAMELEN) \
 || defined (USGr4) || defined (__svr4__) || defined (__SVR4)
@@ -193,14 +195,35 @@ typedef long	off_t;
 || defined (__H3050) || defined (__H3050R) || defined (__H3050RX)
 #define	SYSV
 #define	OSTYPE			"HPUX"
-#define	EXTENDCCOPT		"-D_FILE_OFFSET_BITS=64"
-#define	USETERMINFO
-#define	TERMCAPLIB		"-lcurses"
-#define	STRICTSTDC
+# ifdef	__CLASSIC_C__
+# define	NOSTDC
+# endif
+# ifdef	__STDC_EXT__
+# define	EXTENDCCOPT		"-D_FILE_OFFSET_BITS=64"
+# endif
+/*
+ *	This is a fake '#if' for some buggy HP-UX terminfo library.
+ *	Will you please define 'BUGGY_HPUX' manually on such environment.
+ *
+ *	Some HP-UX has buggy libcurses/terminfo, to use libtermcap/termcap.
+ *	Another HP-UX has buggy libtermlib/termcap, to use libcurses/terminfo.
+ *	Or another HP-UX does not support termcap at all.
+ *	If you use HP-UX and the terminal trouble occured,
+ *	you should try to define 'BUGGY_HPUX' here and to re-compile.
+ *
+#define	BUGGY_HPUX
+ *
+ */
+# ifndef	__HP_cc
+# define	BUGGY_HPUX		/* Maybe HP-UX 10.20 */
+# endif
+# if	!defined (BUGGY_HPUX)
+# define	USETERMINFO
+# define	TERMCAPLIB		"-lcurses"
+# endif
 #define	USEPID_T
 #define	NOSIGLIST
 #define	NOTZFILEH
-#define	USEMKNODH
 #define	NOTMGMTOFF
 #define	USEREGCOMP
 #define	NOTERMVAR
@@ -215,6 +238,9 @@ typedef long	off_t;
 #define	USESETVBUF
 #define	SIGFNCINT
 #define	USESTRERROR
+#define	USESETSID
+#define	USESETRESUID
+#define	USESETRESGID
 #endif
 
 #if	defined (nec_ews) || defined (_nec_ews)
@@ -302,6 +328,7 @@ typedef long	off_t;
 && !defined (__FreeBSD__) && !defined (__NetBSD__) && !defined (__OpenBSD__)
 #define	CODEEUC
 #define	TARUSESPACE
+#define	NOFUNCMACRO
 #define	EXTENDLIB	"-lc_r"
 #define	USEMNTINFOR
 # if	defined (SYSTYPE_BSD)
@@ -447,9 +474,10 @@ typedef long	off_t;
 #define	USEMANLANG
 #define	BSDINSTALL
 #define	TARUSESPACE
-#define	TERMCAPLIB		"-lncurses"
+#define	TERMCAPLIB		"-ltermcap"
 #define	NOSIGLIST
 #define	DECLERRLIST
+#define	HAVECLINE
 #define	SYSVDIRENT
 #define	NODNAMLEN
 #define	NODRECLEN
@@ -460,6 +488,7 @@ typedef long	off_t;
 #define	USEMKTIME
 #define	DEFFDSETSIZE
 #define	SIGFNCINT
+#define	WAITKEYPAD		36
 #endif
 
 #if	defined (linux)
@@ -496,6 +525,7 @@ typedef long	off_t;
 #define	DECLERRLIST
 #define	NOTZFILEH
 #define	USETIMEH
+#define	HAVECLINE
 #define	SYSVDIRENT
 #define	NODNAMLEN
 #define	HAVETIMEZONE
@@ -507,7 +537,8 @@ typedef long	off_t;
 && !defined (ia64) && !defined (__ia64) && !defined (__ia64__) \
 && !defined (x86_64) && !defined (__x86_64) && !defined (__x86_64__) \
 && !defined (s390x) && !defined (__s390x) && !defined (__s390x__) \
-&& !defined (CONFIG_ARCH_S390X)
+&& !defined (CONFIG_ARCH_S390X) \
+&& (!defined (PPC) || !defined (__GNUC__) || __GNUC__ >= 3)	/* for bug */
 # define	USELLSEEK
 # endif
 #define	SIGFNCINT
@@ -765,13 +796,21 @@ typedef long	off_t;
 /* #define USETERMINFO	;use terminfo library instead of termcap */
 /* #define TERMCAPLIB	;library needed for termcap */
 /* #define REGEXPLIB	;library needed for regular expression */
+/* #define SOCKETLIB	;library needed for socket */
 /* #define EXTENDLIB	;library needed for the other extended */
 
 /* #define UNKNOWNFS	;use unsupported type FileSystem */
 
+/* #define NOFILEMACRO	;cc(1) cannot evaluate __FILE__ macro */
+/* #define NOFUNCMACRO	;cc(1) cannot evaluate __FUNCTION__ macro */
+/* #define NOLINEMACRO	;cc(1) cannot evaluate __LINE__ macro */
 /* #define FORCEDSTDC	;not defined __STDC__, but expect standard C */
 /* #define STRICTSTDC	;cannot allow K&R type function declaration */
+/* #define NOSTDC	;defined __STDC__, but expect traditional C */
+/* #define NOCONST	;defined __STDC__, but cannot use 'const' qualifier */
 /* #define NOVOID	;cannot use type 'void' */
+/* #define NOLONGLONG	;cannot use type 'long long' */
+/* #define HAVELONGLONG	;have type 'long long' */
 /* #define NOUID_T	;uid_t, gid_t is not defined in <sys/types.h> */
 /* #define USEPID_T	;pid_t is defined in <sys/types.h> as process ID */
 /* #define DECLSIGLIST	;'sys_siglist[]' declared in <signal.h> */
@@ -790,9 +829,12 @@ typedef long	off_t;
 /* #define USESTDARGH	;use <stdarg.h> for va_list */
 /* #define USEMKDEVH	;use <sys/mkdev.h> for major()/minor() */
 /* #define USEMKNODH	;use <sys/mknod.h> for major()/minor() */
+/* #define USELOCKFH	;use <sys/lockf.h> for lockf() */
 /* #define USESGTTY	;use sgtty interface */
 /* #define USETERMIO	;use termio interface */
 /* #define USETERMIOS	;use termios interface */
+/* #define HAVECLINE	;struct termios has c_line */
+/* #define USEDEVPTMX	;use /dev/ptmx and /dev/pts/? for pseudo terminal */
 /* #define USEDIRECT	;use 'struct direct' instead of dirent */
 /* #define SYSVDIRENT	;dirent interface behaves as System V */
 /* #define NODNAMLEN	;struct dirent hasn't d_namlen */
@@ -840,6 +882,9 @@ typedef long	off_t;
 /* #define NOTERMVAR	;have not termcap variables such as PC, ospeed, etc. */
 /* #define USEUTIME	;use utime() instead of utimes() */
 /* #define USEGETWD	;use getwd() instead of getcwd() */
+/* #define USEFCNTLOCK	;use fcntl() lock instead of flock() */
+/* #define USELOCKF	;use lockf() instead of flock() */
+/* #define NOFLOCK	;have neither flock() nor lockf() */
 /* #define USETIMELOCAL	;have timelocal() as inverse of localtime() */
 /* #define USEMKTIME	;use mktime() instead of timelocal() */
 /* #define USESYSCONF	;use sysconf() for getting system configuration */
@@ -859,13 +904,17 @@ typedef long	off_t;
 /* #define SIGFNCINT	;the 2nd argument function of signal() needs int */
 /* #define USESTRERROR	;use strerror() instead of sys_errlist[] */
 /* #define GETTODARGS	;the number of arguments in gettimeofday() */
+/* #define USESETSID	;use setsid() to set session ID */
+/* #define USESOCKLEN	;use socklen_t for bind()/connect()/accept() */
+/* #define USESETREUID	;use setreuid() to set effective user ID */
+/* #define USESETRESUID	;use setresuid() to set effective user ID */
+/* #define USESETREGID	;use setregid() to set effective group ID */
+/* #define USESETRESGID	;use setresgid() to set effective group ID */
 
 /* #define SENSEPERSEC	;ratio of key sense per 1 second */
 /* #define WAITKEYPAD	;interval to wait after getting input of ESC [ms] */
 /* #define WAITMETA	;interval to wait after getting input of META [ms] */
 
-#include "config.h"
-
 
 /*                             */
 /* DO NOT DELETE or EDIT BELOW */
@@ -902,6 +951,7 @@ typedef long	off_t;
 #define	USESETPGID
 #define	USESETVBUF
 #define	USESTRERROR
+#define	USESETSID
 #endif
 
 #ifdef	SVR4
@@ -922,6 +972,21 @@ typedef long	off_t;
 # endif
 #endif
 
+#ifndef	BSD4
+#define	USEDEVPTMX
+#endif
+
+#if	defined (POSIX) || defined (SYSV)
+#define	USEFCNTLOCK
+#endif
+
+/*                                 */
+/* Eval configurations by Configur */
+/*                                 */
+
+#include "config.h"
+
+
 #ifdef	USETERMIOS
 # ifdef	USETERMIO
 # undef	USETERMIO
@@ -936,23 +1001,55 @@ typedef long	off_t;
 #define	USESGTTY
 #endif
 
-#if	defined (__STDC__) || defined (FORCEDSTDC)
+#if	defined (__STDC__) && !defined (NOSTDC) && !defined (FORCEDSTDC)
+#define	FORCEDSTDC
+#endif
+
+#if	defined (__STRICT_ANSI__) && !defined (STRICTSTDC)
+#define	STRICTSTDC
+#endif
+
+#if	defined (FORCEDSTDC) && !defined (USESTDARGH)
 #define	USESTDARGH
 #endif
 
-#if	(defined (__STDC__) || defined (FORCEDSTDC)) \
-&& !defined (__STRICT_ANSI__) && !defined (STRICTSTDC)
+#if	defined (__STDC__) && !defined (NOSTDC) \
+&& !defined (NOLONGLONG) && !defined (HAVELONGLONG)
+#define	HAVELONGLONG
+#endif
+
+#if	defined (FORCEDSTDC) && !defined (STRICTSTDC)
 #define	__P_(args)	args
+#else
+#define	__P_(args)	()
+#endif
+
+#if	defined (FORCEDSTDC) && !defined (NOCONST)
 #define	CONST		const
+#else
+#define	CONST
+#endif
+
+#ifdef	FORCEDSTDC
 #define	ALLOC_T		size_t
 #define	VOID_A		void
 #else
-#define	__P_(args)	()
-#define	CONST
 #define	ALLOC_T		unsigned
 #define	VOID_A
 #endif
 
+#ifdef	NOVOID
+#define	VOID
+#define	VOID_T	int
+#define	VOID_P	char *
+#define	VOID_C
+#else
+#define	VOID	void
+#define	VOID_T	void
+#define	VOID_P	void *
+#define	VOID_C	(void)
+#endif
+
 #if	defined (USEREGCMP) && !defined (REGEXPLIB)
 #define	REGEXPLIB		"-lPW"
 #endif
@@ -980,6 +1077,9 @@ typedef long	off_t;
 #ifndef	REGEXPLIB
 #define	REGEXPLIB		""
 #endif
+#ifndef	SOCKETLIB
+#define	SOCKETLIB		""
+#endif
 #ifndef	EXTENDLIB
 #define	EXTENDLIB		""
 #endif
@@ -1079,4 +1179,4 @@ typedef long	off_t;
 #define	GETTODARGS	2
 #endif
 
-#endif	/* __MACHINE_H_ */
+#endif	/* !__MACHINE_H_ */
